(* BD security for Nondeterministic IO automata *)
theory BD_Security_NIO 
imports BD_Security Nondet_IO_Automaton Filtermap
begin

declare Let_def[simp]

no_notation relcomp (infixr "O" 75)


locale BD_Security_NIO_Aut = Nondet_IO_Automaton istate step
 for istate :: 'state and step :: "'state \<Rightarrow> 'act \<Rightarrow> 'out \<Rightarrow> 'state \<Rightarrow> bool"
+
fixes (* value filtering and production:  *)
   \<phi> :: "('state,'act,'out) trans => bool" and f :: "('state,'act,'out) trans \<Rightarrow> 'value"
 and (* observation filtering and production: *)
   \<gamma> :: "('state,'act,'out) trans => bool" and g :: "('state,'act,'out) trans \<Rightarrow> 'obs"
 and (* declassification trigger:  *)
   T :: "('state,'act,'out) trans \<Rightarrow> bool"
 and (* declassification bound: *)
   B :: "'value list \<Rightarrow> 'value list \<Rightarrow> bool"
begin

(* The value function: *)
definition V :: "('state,'act,'out) trace \<Rightarrow> 'value list" where "V \<equiv> filtermap \<phi> f"
(* The observation function: *)
definition O :: "('state,'act,'out) trace \<Rightarrow> 'obs list" where "O \<equiv> filtermap \<gamma> g"

lemma V_simps[simp]:
"V [] = []"  "\<not> \<phi> trn \<Longrightarrow> V (trn # tr) = V tr"  "\<phi> trn \<Longrightarrow> V (trn # tr) = f trn # V tr"
unfolding V_def by auto

lemma V_Cons_unfold: "V (trn # tr) = (if \<phi> trn then f trn # V tr else V tr)" 
by auto

lemma O_simps[simp]:
"O [] = []"  "\<not> \<gamma> trn \<Longrightarrow> O (trn # tr) = O tr"  "\<gamma> trn \<Longrightarrow> O (trn # tr) = g trn # O tr"
unfolding O_def by auto

lemma O_Cons_unfold: "O (trn # tr) = (if \<gamma> trn then g trn # O tr else O tr)" 
by auto

lemma V_append: "V (tr @ tr1) = V tr @ V tr1" 
unfolding V_def using filtermap_append by auto

lemma V_snoc:
"\<not> \<phi> trn \<Longrightarrow> V (tr ## trn) = V tr"  "\<phi> trn \<Longrightarrow> V (tr ## trn) = V tr ## f trn"
unfolding V_def by auto

lemma O_snoc:
"\<not> \<gamma> trn \<Longrightarrow> O (tr ## trn) = O tr"  "\<gamma> trn \<Longrightarrow> O (tr ## trn) = O tr ## g trn"
unfolding O_def by auto

lemma V_Nil_list_ex: "V tr = [] \<longleftrightarrow> \<not> list_ex \<phi> tr"
unfolding V_def using filtermap_Nil_list_ex by auto

lemma V_Nil_never: "V tr = [] \<longleftrightarrow> never \<phi> tr"
unfolding V_def using filtermap_Nil_never by auto

lemma length_V: "length (V tr) \<le> length tr"
by (auto simp: V_def length_filtermap) 

lemma V_list_all: "V tr = map f tr \<longleftrightarrow> list_all \<phi> tr"
by (auto simp: V_def length_filtermap)

lemma V_eq_Cons: 
assumes "V tr = v # vl1" 
shows "\<exists> trn tr2 tr1. tr = tr2 @ [trn] @ tr1 \<and> never \<phi> tr2 \<and> \<phi> trn \<and> f trn = v \<and> V tr1 = vl1" 
using assms filtermap_eq_Cons unfolding V_def by auto 

lemma V_eq_append: 
assumes "V tr = vl1 @ vl2" 
shows "\<exists> tr1 tr2. tr = tr1 @ tr2 \<and> V tr1 = vl1 \<and> V tr2 = vl2" 
using assms filtermap_eq_append[of \<phi> f] unfolding V_def by auto

lemma V_eq_RCons: 
assumes "V tr = vl1 ## v" 
shows "\<exists> trn tr1 tr2. tr = tr1 @ [trn] @ tr2 \<and> \<phi> trn \<and> f trn = v \<and> V tr1 = vl1 \<and> never \<phi> tr2" 
using assms filtermap_eq_RCons[of \<phi> f] unfolding V_def by blast

lemma V_eq_Cons_RCons: 
assumes "V tr = v # vl1 ## w" 
shows "\<exists> trv trnv tr1 trnw trw. 
   tr = trv @ [trnv] @ tr1 @ [trnw] @ trw \<and> 
   never \<phi> trv \<and> \<phi> trnv \<and> f trnv = v \<and> V tr1 = vl1 \<and> \<phi> trnw \<and> f trnw = w \<and> never \<phi> trw" 
using assms filtermap_eq_Cons_RCons[of \<phi> f] unfolding V_def by blast

lemma O_append: "O (tr @ tr1) = O tr @ O tr1" 
unfolding O_def using filtermap_append by auto

lemma O_Nil_list_ex: "O tr = [] \<longleftrightarrow> \<not> list_ex \<gamma> tr"
unfolding O_def using filtermap_Nil_list_ex by auto

lemma O_Nil_never: "O tr = [] \<longleftrightarrow> never \<gamma> tr"
unfolding O_def using filtermap_Nil_never by auto

lemma length_O: "length (O tr) \<le> length tr"
by (auto simp: O_def length_filtermap) 

lemma O_list_all: "O tr = map g tr \<longleftrightarrow> list_all \<gamma> tr"
by (auto simp: O_def length_filtermap)

lemma O_eq_Cons: 
assumes "O tr = obs # obsl1" 
shows "\<exists> trn tr2 tr1. tr = tr2 @ [trn] @ tr1 \<and> never \<gamma> tr2 \<and> \<gamma> trn \<and> g trn = obs \<and> O tr1 = obsl1" 
using assms filtermap_eq_Cons unfolding O_def by auto 

lemma O_eq_append: 
assumes "O tr = obsl1 @ obsl2" 
shows "\<exists> tr1 tr2. tr = tr1 @ tr2 \<and> O tr1 = obsl1 \<and> O tr2 = obsl2" 
using assms filtermap_eq_append[of \<gamma> g] unfolding O_def by auto

lemma O_eq_RCons: 
assumes "O tr = oul1 ## ou" 
shows "\<exists> trn tr1 tr2. tr = tr1 @ [trn] @ tr2 \<and> \<gamma> trn \<and> g trn = ou \<and> O tr1 = oul1 \<and> never \<gamma> tr2" 
using assms filtermap_eq_RCons[of \<gamma> g] unfolding O_def by blast

lemma O_eq_Cons_RCons: 
assumes "O tr0 = ou # oul1 ## ouu" 
shows "\<exists> tr trn tr1 trnn trr. 
   tr0 = tr @ [trn] @ tr1 @ [trnn] @ trr \<and> 
   never \<gamma> tr \<and> \<gamma> trn \<and> g trn = ou \<and> O tr1 = oul1 \<and> \<gamma> trnn \<and> g trnn = ouu \<and> never \<gamma> trr" 
using assms filtermap_eq_Cons_RCons[of \<gamma> g] unfolding O_def by blast

lemma O_eq_Cons_RCons_append: 
assumes "O tr0 = ou # oul1 ## ouu @ oull" 
shows "\<exists> tr trn tr1 trnn trr. 
   tr0 = tr @ [trn] @ tr1 @ [trnn] @ trr \<and> 
   never \<gamma> tr \<and> \<gamma> trn \<and> g trn = ou \<and> O tr1 = oul1 \<and> \<gamma> trnn \<and> g trnn = ouu \<and> O trr = oull" 
proof-
  from O_eq_append[of tr0 "ou # oul1 ## ouu" oull] assms
  obtain tr00 trrr where 1: "tr0 = tr00 @ trrr" 
  and 2: "O tr00 = ou # oul1 ## ouu" and 3: "O trrr = oull" by auto
  from O_eq_Cons_RCons[OF 2] obtain tr trn tr1 trnn trr where 
  4:"tr00 = tr @ [trn] @ tr1 @ [trnn] @ trr \<and>
     never \<gamma> tr \<and>
     \<gamma> trn \<and> g trn = ou \<and> O tr1 = oul1 \<and> \<gamma> trnn \<and> g trnn = ouu \<and> never \<gamma> trr" by auto
  show ?thesis apply(rule exI[of _ tr], rule exI[of _ trn], rule exI[of _ tr1], 
     rule exI[of _ trnn], rule exI[of _ "trr @ trrr"]) 
  using 1 3 4 by (simp add: O_append O_Nil_never)
qed

lemma O_Nil_tr_Nil: "O tr \<noteq> [] \<Longrightarrow> tr \<noteq> []"
by (induction tr) auto

lemma V_Cons_eq_append: "V (trn # tr) = V [trn] @ V tr"
by (cases "\<phi> trn") auto

lemma set_V: "set (V tr) \<subseteq> {f trn | trn . trn \<in>\<in> tr \<and> \<phi> trn}"
using set_filtermap unfolding V_def .

lemma set_O: "set (O tr) \<subseteq> {g trn | trn . trn \<in>\<in> tr \<and> \<gamma> trn}"
using set_filtermap unfolding O_def .

lemma list_ex_length_O: 
assumes "list_ex \<gamma> tr"  shows "length (O tr) > 0"
by (metis assms O_Nil_list_ex length_greater_0_conv)

lemma list_ex_iff_length_O: 
"list_ex \<gamma> tr \<longleftrightarrow> length (O tr) > 0"
by (metis assms O_Nil_list_ex length_greater_0_conv)

lemma length1_O_list_ex_iff: 
"length (O tr) > 1 \<Longrightarrow> list_ex \<gamma> tr"
unfolding list_ex_iff_length_O by auto

lemma list_all_O_map: "list_all \<gamma> tr \<Longrightarrow> O tr = map g tr"
using assms O_list_all by auto

lemma never_O_Nil: "never \<gamma> tr \<Longrightarrow> O tr = []"
using assms O_Nil_never by auto

lemma list_all_V_map: "list_all \<phi> tr \<Longrightarrow> V tr = map f tr"
using assms V_list_all by auto

lemma never_V_Nil: "never \<phi> tr \<Longrightarrow> V tr = []"
using assms V_Nil_never by auto

(* Reachable states by transitions satisfying T: *)
inductive reachNT:: "'state \<Rightarrow> bool" where
Istate: "reachNT istate"
|
Step:
"\<lbrakk>reachNT (srcOf trn); step (srcOf trn) (actOf trn) (outOf trn) (tgtOf trn); \<not> T trn\<rbrakk>
 \<Longrightarrow> reachNT (tgtOf trn)"

lemma reachNT_PairI:
assumes "reachNT s" and "step s a ou s'" and "\<not> T (Trans s a ou s')"
shows "reachNT s'"
by (metis Step actOf.simps assms outOf.simps srcOf.simps tgtOf.simps)

lemma reachNT_reach: assumes "reachNT s"  shows "reach s"
using assms apply induct by (auto intro: reach.intros)

lemma reachNT_stateO_aux:
  assumes "reachNT s"
  shows "s = istate \<or> (\<exists>sh a ou. reach sh \<and> step sh a ou s \<and> \<not>T (Trans sh a ou s))"
  using assms apply induct
  apply clarsimp
  apply (metis reachNT_reach actOf.simps outOf.simps srcOf.simps tgtOf.simps validTrans.cases)
  done

lemma reachNT_state_cases[cases set,consumes 1, case_names init step]:
  assumes "reachNT s"
  obtains "s = istate"
  | sh a ou where "reach sh" "step sh a ou s" "\<not>T (Trans sh a ou s)"
  by (metis reachNT_stateO_aux[OF assms])

(* This is assumed to be an invariant only modulo non T  *)
definition invarNT where
"invarNT Inv \<equiv> \<forall> s a ou s'. reachNT s \<and> Inv s \<and> \<not> T (Trans s a ou s') \<and> step s a ou s' \<longrightarrow> Inv s'"

lemma invarNT_disj:
assumes "invarNT Inv1" and "invarNT Inv2"
shows "invarNT (\<lambda> s. Inv1 s \<or> Inv2 s)"
using assms unfolding invarNT_def by blast

lemma invarNT_conj:
assumes "invarNT Inv1" and "invarNT Inv2"
shows "invarNT (\<lambda> s. Inv1 s \<and> Inv2 s)"
using assms unfolding invarNT_def by blast

lemma holdsIstate_invarNT:
assumes h: "holdsIstate Inv" and i: "invarNT Inv" and a: "reachNT s"
shows "Inv s"
using a apply (induct rule: reachNT.induct)
using h i unfolding holdsIstate_def invarNT_def
apply metis
by (metis i invarNT_def srcOf.simps tgtOf.simps validTrans validTrans.elims(2))

end (* context BD_Security_IO_Aut *)

sublocale BD_Security_NIO_Aut < BD_Security where 
(* 'traces = ('state,'act,'out) trace, 'values = 'value list, 'observations = 'obs list   *)
validSystemTrace = "validFrom istate" and V = V and O = O and B = B and TT = "never T"
done

context BD_Security_NIO_Aut 
begin 

(* Note: 
lemma secure_def: 
"secure =
 (\<forall> tr vl vl1.
   validFrom istate tr \<and> never T tr \<and> B vl vl1 \<and> V tr = vl \<longrightarrow>
   (\<exists> tr1. validFrom istate tr1 \<and> O tr1 = O tr \<and> V tr1 = vl1))"
unfolding secure_def ..
*)

lemma V_iff_non_\<phi>[simp]: "V (trn # tr) = V tr \<longleftrightarrow> \<not> \<phi> trn"
by (cases "\<phi> trn") auto

lemma V_imp_\<phi>: "V (trn # tr) = v # V tr \<Longrightarrow> \<phi> trn"
by (cases "\<phi> trn") auto

lemma V_imp_Nil: "V (trn # tr) = [] \<Longrightarrow> V tr = []"
by (metis V_simps list.distinct trans.exhaust)

lemma V_iff_Nil[simp]: "V (trn # tr) = [] \<longleftrightarrow> \<not> \<phi> trn \<and> V tr = []"
by (metis V_iff_non_\<phi> V_imp_Nil)

end (* context BD_Security_IO_Aut *)


section{* Unwinding method for proving security *}

context BD_Security_NIO_Aut
begin

definition consume :: "('state,'act,'out) trans \<Rightarrow> 'value list \<Rightarrow> 'value list \<Rightarrow> bool" where
"consume trn vl vl' \<equiv>
 if \<phi> trn then vl \<noteq> [] \<and> f trn = hd vl \<and> vl' = tl vl
 else vl' = vl"

lemma length_consume[simp]:
"consume trn vl vl' \<Longrightarrow> length vl' < Suc (length vl)"
unfolding consume_def by (auto split: if_splits)

lemma ex_consume_\<phi>:
assumes "\<not> \<phi> trn"
obtains vl' where "consume trn vl vl'"
using assms unfolding consume_def by auto

lemma ex_consume_NO:
assumes "vl \<noteq> []" and "f trn = hd vl"
obtains vl' where "consume trn vl vl'"
using assms unfolding consume_def by (cases "\<phi> trn") auto

(* independent action: *)
definition iaction where
"iaction \<Delta> s vl s1 vl1 \<equiv>
 \<exists> a1 ou1 s1' vl1'.
   let trn1 = Trans s1 a1 ou1 s1' in
   validTrans trn1 \<and>
   \<phi> trn1 \<and> consume trn1 vl1 vl1' \<and>
   \<not> \<gamma> trn1
   \<and>
   \<Delta> s vl s1' vl1'"

lemma iactionI[intro?]:
assumes "step s1 a1 ou1 s1'" and "\<phi> (Trans s1 a1 ou1 s1')"
and "consume (Trans s1 a1 ou1 s1') vl1 vl1'"
and "\<not> \<gamma> (Trans s1 a1 ou1 s1')" and "\<Delta> s vl s1' vl1'"
shows "iaction \<Delta> s vl s1 vl1"
unfolding iaction_def using assms by auto

definition match where
"match \<Delta> s s1 vl1 a ou s' vl' \<equiv>
 \<exists> a1 ou1 s1' vl1'.
    let trn = Trans s a ou s'; trn1 = Trans s1 a1 ou1 s1' in
    validTrans trn1 \<and>
    consume trn1 vl1 vl1' \<and>
    \<gamma> trn = \<gamma> trn1 \<and> (\<gamma> trn \<longrightarrow> g trn = g trn1) \<and>
    \<Delta> s' vl' s1' vl1'"

lemma matchI[intro?]:
assumes "validTrans (Trans s1 a1 ou1 s1')"
and "consume (Trans s1 a1 ou1 s1') vl1 vl1'" and "\<gamma> (Trans s a ou s') = \<gamma> (Trans s1 a1 ou1 s1')"
and "\<gamma> (Trans s a ou s') \<Longrightarrow> g (Trans s a ou s') = g (Trans s1 a1 ou1 s1')"
and "\<Delta> s' vl' s1' vl1'"
shows "match \<Delta> s s1 vl1 a ou s' vl'"
unfolding match_def using assms by auto

definition ignore where
"ignore \<Delta> s s1 vl1 a ou s' vl' \<equiv>
 \<not> \<gamma> (Trans s a ou s') \<and>
 \<Delta> s' vl' s1 vl1"

lemma ignoreI[intro?]:
assumes "\<not> \<gamma> (Trans s a ou s')" and "\<Delta> s' vl' s1 vl1"
shows "ignore \<Delta> s s1 vl1 a ou s' vl'"
unfolding ignore_def using assms by auto

(* reaction: *)
definition reaction where
"reaction \<Delta> s vl s1 vl1 \<equiv>
 \<forall> a ou s' vl'.
   let trn = Trans s a ou s' in
   validTrans trn \<and> \<not> T trn \<and>
   consume trn vl vl'
   \<longrightarrow>
   match \<Delta> s s1 vl1 a ou s' vl'
   \<or>
   ignore \<Delta> s s1 vl1 a ou s' vl'"

lemma reactionI[intro?]:
assumes
"\<And>a ou s' vl'.
   \<lbrakk>step s a ou s'; \<not> T (Trans s a ou s');
    consume (Trans s a ou s') vl vl'\<rbrakk>
   \<Longrightarrow>
   match \<Delta> s s1 vl1 a ou s' vl' \<or> ignore \<Delta> s s1 vl1 a ou s' vl'"
shows "reaction \<Delta> s vl s1 vl1"
using assms unfolding reaction_def by auto

definition "exit" :: "'state \<Rightarrow> 'value \<Rightarrow> bool" where
"exit s v \<equiv> \<forall> tr trn. validFrom s (tr ## trn) \<and> never T (tr ## trn) \<and> \<phi> trn \<longrightarrow> f trn \<noteq> v"

lemma exit_coind:
assumes K: "K s"
and I: "\<And> trn. \<lbrakk>K (srcOf trn); validTrans trn; \<not> T trn\<rbrakk>
        \<Longrightarrow> (\<phi> trn \<longrightarrow> f trn \<noteq> v) \<and> K (tgtOf trn)"
shows "exit s v"
using K unfolding exit_def proof(intro allI conjI impI)
  fix tr trn assume "K s" and "validFrom s (tr ## trn) \<and> never T (tr ## trn) \<and> \<phi> trn"
  thus "f trn \<noteq> v"
  apply (induction tr arbitrary: s trn)
  using I unfolding validFrom_def apply auto
  by (metis neq_Nil_conv rotate1.simps(2) rotate1_is_Nil_conv valid_ConsE)
qed

definition noVal where
"noVal K v \<equiv>
 \<forall> s a ou s'. reachNT s \<and> K s \<and> step s a ou s' \<and> \<phi> (Trans s a ou s') \<longrightarrow> f (Trans s a ou s') \<noteq> v"

lemma noVal_disj:
assumes "noVal Inv1 v" and "noVal Inv2 v"
shows "noVal (\<lambda> s. Inv1 s \<or> Inv2 s) v"
using assms unfolding noVal_def by metis

lemma noVal_conj:
assumes "noVal Inv1 v" and "noVal Inv2 v"
shows "noVal (\<lambda> s. Inv1 s \<and> Inv2 s) v"
using assms unfolding noVal_def by blast

(* Often encountered sufficient criterion for noVal: *)
definition no\<phi> where
"no\<phi> K \<equiv> \<forall> s a ou s'. reachNT s \<and> K s \<and> step s a ou s' \<longrightarrow> \<not> \<phi> (Trans s a ou s')"

lemma no\<phi>_noVal: "no\<phi> K \<Longrightarrow> noVal K v"
unfolding no\<phi>_def noVal_def by auto

(* intro rule for quick inline checks: *)
lemma exitI[consumes 2, induct pred: "exit"]:
assumes rs: "reachNT s" and K: "K s"
and I:
"\<And> s a ou s'.
   \<lbrakk>reach s; reachNT s; step s a ou s'; K s\<rbrakk>
   \<Longrightarrow> (\<phi> (Trans s a ou s') \<longrightarrow> f (Trans s a ou s') \<noteq> v) \<and> K s'"
shows "exit s v"
proof-
  let ?K = "\<lambda> s. reachNT s \<and> K s"
  show ?thesis apply(rule exit_coind[of ?K])
  using assms apply - apply simp apply (case_tac trn, simp_all)
  by (metis reachNT_PairI reachNT_reach)
qed

(* intro rule for more elaborate checks: *)
lemma exitI2:
assumes rs: "reachNT s" and K: "K s"
and "invarNT K" and "noVal K v"
shows "exit s v"
proof-
  let ?K = "\<lambda> s. reachNT s \<and> K s"
  show ?thesis apply(rule exit_coind[of ?K])
  using assms unfolding invarNT_def noVal_def apply - apply simp
  by (case_tac trn, simp_all) (metis reachNT_PairI)+
qed

(* Binary version of the invariant: *)
definition noVal2 where
"noVal2 K v \<equiv>
 \<forall> s a ou s'. reachNT s \<and> K s v \<and> step s a ou s' \<and> \<phi> (Trans s a ou s') \<longrightarrow> f (Trans s a ou s') \<noteq> v"

lemma noVal2_disj:
assumes "noVal2 Inv1 v" and "noVal2 Inv2 v"
shows "noVal2 (\<lambda> s v. Inv1 s v \<or> Inv2 s v) v"
using assms unfolding noVal2_def by metis

lemma noVal2_conj:
assumes "noVal2 Inv1 v" and "noVal2 Inv2 v"
shows "noVal2 (\<lambda> s v. Inv1 s v \<and> Inv2 s v) v"
using assms unfolding noVal2_def by blast

lemma noVal_noVal2: "noVal K v \<Longrightarrow> noVal2 (\<lambda> s v. K s) v"
unfolding noVal_def noVal2_def by auto

lemma exitI_noVal2[consumes 2, induct pred: "exit"]:
assumes rs: "reachNT s" and K: "K s v"
and I:
"\<And> s a ou s'.
   \<lbrakk>reach s; reachNT s; step s a ou s'; K s v\<rbrakk>
   \<Longrightarrow> (\<phi> (Trans s a ou s') \<longrightarrow> f (Trans s a ou s') \<noteq> v) \<and> K s' v"
shows "exit s v"
proof-
  let ?K = "\<lambda> s. reachNT s \<and> K s v"
  show ?thesis apply(rule exit_coind[of ?K])
  using assms apply - apply simp apply (case_tac trn, simp_all)
  by (metis reachNT_PairI reachNT_reach)
qed

lemma exitI2_noVal2:
assumes rs: "reachNT s" and K: "K s v"
and "invarNT (\<lambda> s. K s v)" and "noVal2 K v"
shows "exit s v"
proof-
  let ?K = "\<lambda> s. reachNT s \<and> K s v"
  show ?thesis apply(rule exit_coind[of ?K])
  using assms unfolding invarNT_def noVal2_def apply - apply simp
  by (case_tac trn, simp_all) (metis reachNT_PairI)+
qed

(* end binary version *)

lemma exit_validFrom:
assumes vl: "vl \<noteq> []" and i: "exit s (hd vl)" and v: "validFrom s tr" and V: "V tr = vl"
and T: "never T tr"
shows False
using i v V T proof(induction tr arbitrary: s)
  case Nil thus ?case by (metis V_simps(1) vl)
next
  case (Cons trn tr s)
  show ?case
  proof(cases "\<phi> trn")
    case True
    hence "f trn = hd vl" using Cons by (metis V_simps list.sel(1))
    moreover have "validFrom s [trn]" using `validFrom s (trn # tr)`
    unfolding validFrom_def by auto
    ultimately show ?thesis using Cons True unfolding exit_def
    apply(elim allE[of _ "[]"]) by auto
  next
    case False
    hence "V tr = vl" using Cons by auto
    moreover have "never T tr" by (metis Cons.prems list_all_simps)
    moreover from `validFrom s (trn # tr)` have "validFrom (tgtOf trn) tr" and s: "s = srcOf trn"
    apply (metis list.distinct(1) validFrom_def valid_ConsE)
    by (metis Cons.prems(2) list.sel(1) list.distinct(1) validFrom_def)
    moreover have "exit (tgtOf trn) (hd vl)" using `exit s (hd vl)`
    unfolding exit_def s
    by (metis Cons append_Cons list.distinct list_all_simps
    srcOf.simps tgtOf.simps validFrom_def validFrom_validTrans validTrans.cases valid_ConsE)
    ultimately show ?thesis using Cons(1) by auto
  qed
qed

definition unwind where
"unwind \<Delta> \<equiv>
 \<forall> s vl s1 vl1.
   reachNT s \<and> reach s1 \<and> \<Delta> s vl s1 vl1
   \<longrightarrow>
   (vl \<noteq> [] \<and> exit s (hd vl))
   \<or>
   iaction \<Delta> s vl s1 vl1
   \<or>
   ((vl \<noteq> [] \<or> vl1 = []) \<and> reaction \<Delta> s vl s1 vl1)"

lemma unwindI[intro?]:
assumes
"\<And> s vl s1 vl1.
   \<lbrakk>reachNT s; reach s1; \<Delta> s vl s1 vl1\<rbrakk>
   \<Longrightarrow>
   (vl \<noteq> [] \<and> exit s (hd vl))
   \<or>
   iaction \<Delta> s vl s1 vl1
   \<or>
   ((vl \<noteq> [] \<or> vl1 = []) \<and> reaction \<Delta> s vl s1 vl1)"
shows "unwind \<Delta>"
using assms unfolding unwind_def by auto

lemma unwind_trace:
assumes unwind: "unwind \<Delta>" and "reachNT s" and "reach s1" and "\<Delta> s vl s1 vl1"
and "validFrom s tr" and "never T tr" and "V tr = vl"
shows "\<exists>tr1. validFrom s1 tr1 \<and> O tr1 = O tr \<and> V tr1 = vl1"
proof-
  let ?S = "\<lambda> tr vl1.
  \<forall> s vl s1. reachNT s \<and> reach s1 \<and> \<Delta> s vl s1 vl1 \<and> validFrom s tr \<and> never T tr \<and> V tr = vl \<longrightarrow>
          (\<exists>tr1. validFrom s1 tr1 \<and> O tr1 = O tr \<and> V tr1 = vl1)"
  let ?f = "\<lambda> tr vl1. length tr + length vl1"
  have "?S tr vl1"
  proof(induct rule: measure_induct2[of ?f ?S])
    case (IH tr vl1)
    show ?case
    proof(intro allI impI, elim conjE)
      fix s vl s1 assume rs: "reachNT s" and rs1: "reach s1" and \<Delta>: "\<Delta> s vl s1 vl1"
      and v: "validFrom s tr" and NT: "never T tr" and V: "V tr = vl"
      hence "(vl \<noteq> [] \<and> exit s (hd vl)) \<or>
             iaction \<Delta> s vl s1 vl1 \<or>
             (reaction \<Delta> s vl s1 vl1 \<and> \<not> iaction \<Delta> s vl s1 vl1)"
      (is "?exit \<or> ?iact \<or> ?react \<and> _")
      using unwind unfolding unwind_def by metis
      thus "\<exists>tr1. validFrom s1 tr1 \<and> O tr1 = O tr \<and> V tr1 = vl1"
      proof safe
        assume "vl \<noteq> []" and "exit s (hd vl)"
        hence False using v V exit_validFrom NT by auto
        thus ?thesis by auto
      next
        assume ?iact
        thus ?thesis  unfolding iaction_def Let_def proof safe
          fix a1 :: 'act and ou1 :: 'out and s1' :: 'state and vl1'
          let ?trn1 = "Trans s1 a1 ou1 s1'"
          assume vtrans1: "validTrans ?trn1" and \<phi>1: "\<phi> ?trn1"
          and c: "consume ?trn1 vl1 vl1'" and \<gamma>: "\<not> \<gamma> ?trn1" and \<Delta>: "\<Delta> s vl s1' vl1'"
          from \<phi>1 c obtain v1 where vl1: "vl1 = v1 # vl1'" and f1: "f ?trn1 = v1"
          unfolding consume_def by (cases vl1) auto
          have rs1': "reach s1'" using rs1 vtrans1 by (auto intro: reach_PairI)
          obtain tr1 where v1: "validFrom s1' tr1" and O: "O tr1 = O tr" and V: "V tr1 = vl1'"
          using IH[of tr vl1'] rs rs1' \<Delta> v NT V unfolding vl1 by auto
          show ?thesis
          apply(rule exI[of _ "?trn1 # tr1"])
          using vtrans1 v1 O \<gamma> V \<phi>1 f1 unfolding vl1 by auto
        qed
      next
        assume react: ?react and iact: "\<not> ?iact"
        show ?thesis
        proof(cases tr)
          case Nil note tr = Nil
          hence vl: "vl = []" using V by simp
          show ?thesis proof(cases vl1)
            case Nil note vl1 = Nil
            show ?thesis using IH[of tr vl1] \<Delta> V NT V unfolding tr vl1 by auto
          next
            case Cons
            hence False using vl unwind rs rs1 \<Delta> iact unfolding unwind_def by auto
            thus ?thesis by auto
          qed
        next
          case (Cons trn tr') note tr = Cons
          show ?thesis
          proof(cases trn)
            case (Trans ss a ou s') note trn = Trans let ?trn = "Trans s a ou s'"
            have ss: "ss = s" using trn v unfolding tr validFrom_def by auto
            have Ta: "\<not> T ?trn" and s: "s = srcOf trn" and vtrans: "validTrans ?trn"
            and v': "validFrom s' tr'" and NT': "never T tr'"
            using v NT V unfolding tr validFrom_def trn by auto
            have rs': "reachNT s'" using rs vtrans Ta by (auto intro: reachNT_PairI)
            {assume "\<phi> ?trn" hence "vl \<noteq> [] \<and> f ?trn = hd vl" using V unfolding tr trn ss by auto
            }
            then obtain vl' where c: "consume ?trn vl vl'"
            using ex_consume_\<phi> ex_consume_NO by metis
            have V': "V tr' = vl'" using V c unfolding tr trn ss consume_def
            apply(cases "\<phi> ?trn", simp_all) by (metis list.sel(2-3))
            have "match \<Delta> s s1 vl1 a ou s' vl' \<or> ignore \<Delta> s s1 vl1 a ou s' vl'" (is "?match \<or> ?ignore")
            using react unfolding reaction_def using vtrans Ta c by auto
            thus ?thesis proof safe
              assume ?match
              thus ?thesis unfolding match_def Let_def proof (elim exE conjE)
                fix a1 :: 'act and ou1 :: 'out and s1' :: 'state and vl1'
                let ?trn1 = "Trans s1 a1 ou1 s1'"
                assume \<Delta>: "\<Delta> s' vl' s1' vl1'" and vtrans1: "validTrans ?trn1"
                and c1: "consume ?trn1 vl1 vl1'" and \<gamma>: "\<gamma> ?trn = \<gamma> ?trn1"
                and g: "\<gamma> ?trn \<longrightarrow> g ?trn = g ?trn1"
                have rs1': "reach s1'" using rs rs1 vtrans vtrans1 by (auto intro: reach_PairI)
                obtain tr1 where v1: "validFrom s1' tr1" and O: "O tr1 = O tr'" and V: "V tr1 = vl1'"
                using IH[of tr' vl1'] rs' rs1' \<Delta> v' NT' V' c1 unfolding tr by auto
                have "V (?trn1 # tr1) = vl1"
                using c1 V unfolding consume_def by (cases "\<phi> ?trn1") auto
                thus ?thesis
                apply(intro exI[of _ "Trans s1 a1 ou1 s1' # tr1"])
                using vtrans1 v1 O \<gamma> g V unfolding tr trn ss by auto
              qed
            next
              assume ?ignore
              thus ?thesis unfolding ignore_def Let_def proof (elim exE conjE)
                assume \<gamma>: "\<not> \<gamma> ?trn" and \<Delta>: "\<Delta> s' vl' s1 vl1"
                obtain tr1 where v1: "validFrom s1 tr1" and O: "O tr1 = O tr'" and V: "V tr1 = vl1"
                using IH[of tr' vl1] rs' rs1 \<Delta> v' NT' V' c unfolding tr by auto
                show ?thesis
                apply(intro exI[of _ tr1])
                using v1 O V \<gamma> unfolding tr trn ss by auto
              qed
            qed
          qed
        qed
      qed
    qed
  qed
  thus ?thesis using assms by auto
qed

theorem unwind_secure:
assumes init: "\<And> vl vl1. B vl vl1 \<Longrightarrow> \<Delta> istate vl istate vl1"
and unwind: "unwind \<Delta>"
shows secure
using assms unwind_trace unfolding secure_def by (blast intro: reach.Istate reachNT.Istate)



end (* locale BD_Security_IO_Aut *)


end