theory Borrowing_Security imports BD_Security_IO
begin

(* Preliminaries *)

context IO_Automaton begin

declare validTrans_Trans_srcOf_actOf_tgtOf[simp]
declare validTrans_step_srcOf_actOf_tgtOf[simp]
declare sstep_Cons[simp]
declare length_fst_sstep[simp]

end (* context IO_Automaton *)


context BD_Security_IO_Aut begin

declare O_append[simp]
declare V_append[simp]
declare validFrom_Cons[simp]
declare validFrom_append[simp]

declare list_all_O_map[simp]
declare never_O_Nil[simp]
declare list_all_V_map[simp]
declare never_V_Nil[simp]

end (* BD_Security_IO_Aut *)


(* hide_const adjust *)


locale BD_Security2 = 
  Prime: BD_Security validSystemTraces' V' O' B TT' + 
  Orig?: BD_Security validSystemTraces V O B TT (* Note: the "Orig" prefix will be optional *)
for 
   validSystemTraces :: "'traces \<Rightarrow> bool" 
 and 
   V :: "'traces \<Rightarrow> 'values" and O :: "'traces \<Rightarrow> 'observations"
 and 
   TT :: "'traces \<Rightarrow> bool"
 and 
   B :: "'values \<Rightarrow> 'values \<Rightarrow> bool"
 and 
(*  *)
   validSystemTraces' :: "'traces' \<Rightarrow> bool" 
 and 
   V' :: "'traces' \<Rightarrow> 'values" and O' :: "'traces' \<Rightarrow> 'observations'"
 and 
   TT' :: "'traces' \<Rightarrow> bool"
+ 
fixes 
   \<pi> :: "'traces' \<Rightarrow> 'traces"
 and 
   \<rho> :: "'observations' \<Rightarrow> 'observations \<Rightarrow> bool"
begin

abbreviation secure' where "secure' \<equiv> Prime.secure"

definition "\<pi>_validSystemTraces \<equiv> \<forall> tr'. validSystemTraces' tr' \<longrightarrow> validSystemTraces (\<pi> tr')"
definition "\<pi>_TT \<equiv> \<forall> tr'. validSystemTraces' tr' \<and> TT' tr' \<longrightarrow> TT (\<pi> tr')"
definition "V_\<pi> \<equiv> \<forall> tr'. validSystemTraces' tr' \<longrightarrow> V (\<pi> tr') = V' tr'"
definition "\<pi>_\<rho> \<equiv> \<forall> tr'. validSystemTraces' tr' \<longrightarrow> \<rho> (O' tr') (O (\<pi> tr'))"

(* A realistic ``double'' pullback involving both observations and values: *)
definition "pullbackOV \<equiv>
  \<forall> tr ol'. 
    validSystemTraces tr \<and> \<rho> ol' (O tr)  
    \<longrightarrow>
    (\<exists> tr'. validSystemTraces' tr' \<and> O' tr' = ol' \<and> V' tr' = V tr)"

theorem borrow_secure_pullbackOV: 
assumes \<pi>_TT: \<pi>_TT and \<pi>_v: \<pi>_validSystemTraces
and V_\<pi>: V_\<pi> and \<pi>_\<rho>: \<pi>_\<rho> and p: pullbackOV
and s: secure  
shows secure'
unfolding Prime.secure_def proof safe 
  fix tr' vl1
  assume v': "validSystemTraces' tr'"
  and TT': "TT' tr'" and B: "B (V' tr') vl1"
  def tr \<equiv> "\<pi> tr'"  
  have "B (V tr) vl1" and "validSystemTraces tr" and "TT tr"
  using B v' TT' V_\<pi> \<pi>_v \<pi>_TT unfolding V_\<pi>_def tr_def \<pi>_TT_def \<pi>_validSystemTraces_def by auto
  then obtain tr1 where 1: "validSystemTraces tr1" and 2: "O tr1 = O tr" and 3: "V tr1 = vl1"
  using s unfolding secure_def by auto
  have "\<rho> (O' tr') (O tr)" using \<pi>_\<rho> v' unfolding \<pi>_\<rho>_def tr_def by auto
  then obtain tr1' where "validSystemTraces' tr1' \<and> O' tr1' = O' tr' \<and> V' tr1' = V tr1"
  using 1 p unfolding pullbackOV_def 2[symmetric] by auto
  thus "\<exists>tr1'. validSystemTraces' tr1' \<and> O' tr1' = O' tr' \<and> V' tr1' = vl1"
  unfolding 3 by auto
qed

(* Parenthesis (not used in further development: 
   The ``ideal'' pullback: *)
definition "pullback \<equiv>
  \<forall> tr ol'. 
    validSystemTraces tr \<and> \<rho> ol' (O tr)  
    \<longrightarrow>
    (\<exists> tr'. validSystemTraces' tr' \<and> O' tr' = ol' \<and> \<pi> tr' = tr)"

lemma pullback_pullbackOV:
assumes V_\<pi>: V_\<pi> and pullback
shows pullbackOV
using assms unfolding V_\<pi>_def pullbackOV_def pullback_def by force

corollary borrow_secure_ideal: 
assumes \<pi>_TT: \<pi>_TT and \<pi>_v: \<pi>_validSystemTraces
and V_\<pi>: V_\<pi> and \<pi>_\<rho>: \<pi>_\<rho> and p: pullback
and s: secure  
shows secure'
using assms borrow_secure_pullbackOV pullback_pullbackOV by auto
(* End of paranthesis *)

(* An ``adjustment'' applied to an original system trace to applied a trace that is ``P'' 
 w.r.t. the new system.  *)
definition "adjust \<equiv> \<lambda> P :: 'observations' \<Rightarrow> 'traces \<Rightarrow> bool. 
  \<forall>tr ol'. 
     validSystemTraces tr \<and> \<rho> ol' (O tr)
     \<longrightarrow> 
     (\<exists>tr1. validSystemTraces tr1 \<and> O tr1 = O tr \<and> V tr1 = V tr \<and> P ol' tr1)"

definition "pull \<equiv> \<lambda> P :: 'observations' \<Rightarrow> 'traces \<Rightarrow> bool. 
  \<forall> tr ol'. 
    validSystemTraces tr \<and> \<rho> ol' (O tr) \<and> P ol' tr   
    \<longrightarrow>
    (\<exists> tr'. validSystemTraces' tr' \<and> O' tr' = ol' \<and> \<pi> tr' = tr)"

lemma pull_adjust_pullback:
fixes P :: "'observations' \<Rightarrow> 'traces \<Rightarrow> bool"
assumes V_\<pi>: V_\<pi> and p: "pull P" and c: "adjust P"
shows pullbackOV
unfolding pullbackOV_def proof safe
  fix tr ol'
  assume "validSystemTraces tr" and 0: "\<rho> ol' (O tr)"
  then obtain tr1 where 1: "validSystemTraces tr1" and 2: "O tr1 = O tr" and 3: "V tr1 = V tr" 
  and 4: "P ol' tr1" using c unfolding adjust_def by auto
  have "\<rho> ol' (O tr1)" using 0 unfolding 2 .
  with 1 4 obtain tr' where 5: "validSystemTraces' tr'" and 6: "O' tr' = ol'" and 7: "\<pi> tr' = tr1" 
  using p unfolding pull_def by auto
  have "V' tr' = V tr" using V_\<pi> 3 5 7 unfolding V_\<pi>_def by auto
  thus "\<exists>tr'. validSystemTraces' tr' \<and> O' tr' = ol' \<and> V' tr' = V tr"
  using 5 6 by auto
qed

corollary borrow_secure: 
assumes \<pi>_TT and \<pi>_validSystemTraces and \<sigma>_B and V_\<pi> and \<pi>_\<rho> 
and "pull P" and "adjust P"
and s: secure  
shows secure'
using assms borrow_secure_pullbackOV pull_adjust_pullback by auto
  
end (* context BD_Security2 *) 

(* The standard "g" (i.e., observation mapping) function *)
definition g :: "('state,'act,'out) trans \<Rightarrow> ('act \<times> 'out)" where 
"g trn = (actOf trn, outOf trn)"

lemma g[simp]: "g (Trans s a ou s1) = (a,ou)"
unfolding g_def by auto

(* A trick aimed at making Isabelle's locale abbreviation mechanism work as desired: 
Since we want, for each constant or theorem X in IO_Automaton and BD_Security_IO_Aut, 
to have X abbreviate Orig.X and not Prime.X, we have to instantiate our locale first 
with step' and then with step. But step' is defined based on sstep (which is defined in IO_Automaton based 
on step)! So we first leave sstep and traceOf generic (in locale Extension_aux), then instantiate to IO_Automaton step', 
then to IO_Automaton step, and finally instantiate sstep and traceOf. 
The desired end locale is Extension.  *)

datatype ('act, 'newact) act' = OldA 'act | NewA 'newact
datatype ('out, 'newout) out' = OldO 'out | NewO 'newout

locale Extension_aux = 
fixes 
   istate :: 'state and step :: "'state \<Rightarrow> 'act \<Rightarrow> 'out \<times> 'state"
 and 
   \<phi> :: "('state,'act,'out) trans \<Rightarrow> bool" and f :: "('state,'act,'out) trans \<Rightarrow> 'value"
 and  (* 'obs will be 'act \<times> 'out and g will be the projection *)
   \<gamma> :: "('state,'act,'out) trans \<Rightarrow> bool" 
 and 
   T :: "('state,'act,'out) trans \<Rightarrow> bool"
 and 
   B :: "'value list \<Rightarrow> 'value list \<Rightarrow> bool"
and 
  \<epsilon>o :: "'newact \<Rightarrow> 'out list \<Rightarrow> 'newout" 
and 
  \<epsilon>a :: "'newact \<Rightarrow> 'state \<Rightarrow> 'act list"
and 
  sstep :: "'state \<Rightarrow> 'act list \<Rightarrow> 'out list \<times> 'state"
and 
  traceOf :: "'state \<Rightarrow> 'act list \<Rightarrow> ('state,'act,'out) trace"
begin 

definition step' :: "'state \<Rightarrow> ('act, 'newact) act' \<Rightarrow> ('out, 'newout) out' \<times> 'state" where  
[code]: 
"step' s a' = 
 (case a' of OldA a \<Rightarrow> (case step s a of (ou,s1) \<Rightarrow> (OldO ou, s1))
            |NewA na \<Rightarrow> (case sstep s (\<epsilon>a na s) of (oul,s1) \<Rightarrow> (NewO (\<epsilon>o na oul), s1)))"

fun T'  :: "('state, ('act, 'newact) act', ('out, 'newout) out') trans \<Rightarrow> bool" where 
"T' (Trans s a' ou' s1) = 
 (case a' of OldA a \<Rightarrow> T (Trans s a (fst (step s a)) (snd (step s a)))
           | NewA na \<Rightarrow> list_ex T (traceOf s ((\<epsilon>a na s))))"

fun \<phi>' :: "('state, ('act, 'newact) act', ('out, 'newout) out') trans \<Rightarrow> bool"  where 
"\<phi>' (Trans s a' ou' s1) = 
 (case a' of OldA a \<Rightarrow> \<phi> (Trans s a (fst (step s a)) (snd (step s a)))
            |NewA na \<Rightarrow> list_ex \<phi> (traceOf s (\<epsilon>a na s)))"

fun f' :: "('state, ('act, 'newact) act', ('out, 'newout) out') trans \<Rightarrow> 'value" where 
"f' (Trans s a' ou' s1) = 
 (case a' of OldA a \<Rightarrow> f (Trans s a (fst (step s a)) (snd (step s a)))
            |NewA na \<Rightarrow> f (SOME trn. trn \<in>\<in> traceOf s (\<epsilon>a na s) \<and> \<phi> trn))"

fun \<gamma>' :: "('state, ('act, 'newact) act', ('out, 'newout) out') trans \<Rightarrow> bool" where 
"\<gamma>' (Trans s a' ou' s1) = 
 (case a' of OldA a \<Rightarrow> \<gamma> (Trans s a (fst (step s a)) (snd (step s a)))
           | NewA na \<Rightarrow> list_ex \<gamma> (traceOf s ((\<epsilon>a na s))))"

abbreviation "O \<equiv> BD_Security_IO_Aut.O \<gamma> g"

fun g' :: "('state, ('act, 'newact) act', ('out, 'newout) out') trans \<Rightarrow> ('act, 'newact) act' \<times> ('out, 'newout) out'" where
  "g' (Trans s (OldA a) (OldO ou) s1) = (OldA a, OldO ou)"
| "g' (Trans s (NewA na) (NewO nou) s1) = (NewA na, NewO (\<epsilon>o na (map snd (O (traceOf s (\<epsilon>a na s))))))"

end (* context Extension_aux *)


sublocale Extension_aux < Prime: BD_Security_IO_Aut 
where istate = istate and step = step' and B = B and T = T'
and \<phi> = \<phi>' and f = f' and \<gamma> = \<gamma>' and g = g' 
done

sublocale Extension_aux < Orig?: BD_Security_IO_Aut 
where istate = istate and step = step and B = B and T = T
and \<phi> = \<phi> and f = f and \<gamma> = \<gamma> and g = g 
done

locale Extension = 
Extension_aux istate step \<phi> f \<gamma> T B \<epsilon>o \<epsilon>a "IO_Automaton.sstep step" "IO_Automaton.traceOf step"
for istate :: 'state and step :: "'state \<Rightarrow> 'act \<Rightarrow> 'out \<times> 'state"
 and \<phi> and f :: "('state, 'act, 'out) trans \<Rightarrow> 'value" and \<gamma>
 and T B 
 and \<epsilon>o :: "'newact \<Rightarrow> 'out list \<Rightarrow> 'newout" 
 and \<epsilon>a :: "'newact \<Rightarrow> 'state \<Rightarrow> 'act list"


context Extension begin

abbreviation valid' where "valid' \<equiv> Prime.valid"
abbreviation validFrom' where "validFrom' \<equiv> Prime.validFrom"
abbreviation validTrans' where "validTrans' \<equiv> Prime.validTrans"
abbreviation sstep' where "sstep' \<equiv> Prime.sstep" 
abbreviation traceOf' where "traceOf' \<equiv> Prime.traceOf"
abbreviation V' where "V' \<equiv> Prime.V"   abbreviation O' where "O' \<equiv> Prime.O"
abbreviation reach' where "reach' \<equiv> Prime.reach"
abbreviation secure' where "secure' \<equiv> Prime.secure"

(* THE ASSUMPTIONS: *)

(* In the original system, the filter for observations depends only on the actions, not on the state: *)
definition "\<gamma>_act \<equiv> \<forall> trn trn1. actOf trn = actOf trn1 \<longrightarrow> \<gamma> trn = \<gamma> trn1"
(* Each new chain of actions produces at most one secret: *)
definition \<phi>_\<epsilon>a  where "\<phi>_\<epsilon>a na ss \<equiv> 
  \<forall> i j. let tr = traceOf ss (\<epsilon>a na ss) in
      i < j \<and> j < length tr \<and> \<phi> (tr!i) \<and> \<phi> (tr!j) \<longrightarrow> i = j"

definition \<gamma>a :: "'act \<Rightarrow> bool" where
  "\<gamma>a a = (\<exists>s ou s'. \<gamma> (Trans s a ou s'))"

lemma \<gamma>_act_\<gamma>_\<gamma>a:
assumes \<gamma>_act
shows "\<gamma> trn = \<gamma>a (actOf trn)"
using assms unfolding \<gamma>_act_def \<gamma>a_def 
by (metis actOf.simps)

(* The local_adjust property and its auxiliaries. It says that (old-system) sequences of actions 
corresponding to new actions can be taken consecutively (atomically) without affecting the values 
or the returned state.*)
definition "local_right na ss s tr tr1 tr2 \<equiv> 
  let al = \<epsilon>a na ss; tra = traceOf ss al; oul = fst (sstep ss al) in 
  let s1 = (if tr1 = [] then s else tgtOf (last tr1)); 
  trr = traceOf s1 (\<epsilon>a na s1); s2 = tgtOf (last trr) in 
  validFrom s tr1 \<and> validFrom s2 tr2 \<and> never \<gamma> tr1 \<and> never \<gamma> tr2 \<and> 
  O tra = O trr \<and> V (tr1 @ trr @ tr2) = V tr \<and> tgtOf (last (tr1 @ trr @ tr2)) = tgtOf (last tr)"

definition "local_adjust_param na ss s tr \<equiv> 
  let al = \<epsilon>a na ss; oul = fst (sstep ss al) in 
    reach ss \<and> reach s \<and> list_ex \<gamma> (traceOf ss al) \<and> 
    validFrom s tr \<and> O tr = O (traceOf ss al) \<longrightarrow> 
    (\<exists>tr1 tr2. local_right na ss s tr tr1 tr2)"

definition "local_adjust \<equiv> \<forall> na ss s tr. local_adjust_param na ss s tr"

(* Main definitions: *)
fun \<pi> :: "('state,('act,'newact) act',('out, 'newout) out') trace \<Rightarrow> ('state,'act,'out) trace" where 
"\<pi> [] = []"
|
"\<pi> ((Trans s a' ou' s1) # tr') = 
 (case a' of OldA a \<Rightarrow> Trans s a (fst (step s a)) (snd (step s a)) # \<pi> tr' 
            |NewA na \<Rightarrow> traceOf s (\<epsilon>a na s) @ \<pi> tr')"

fun pre_\<rho> :: "'state \<Rightarrow> 'newact \<Rightarrow> 'newout \<Rightarrow> ('act \<times> 'out) list \<Rightarrow> bool" where 
"pre_\<rho> s na nou al_oul = 
  (let al = \<epsilon>a na s; oul = fst (sstep s al) in
       reach s \<and> nou = \<epsilon>o na oul \<and> al_oul = O (traceOf s al) \<and> list_ex \<gamma> (traceOf s al))"

inductive \<rho> :: "(('act, 'newact) act' \<times> ('out,'newout) out') list \<Rightarrow> ('act \<times> 'out) list \<Rightarrow> bool" where
Nil[simp,intro!]: "\<rho> [] []"
|
Step_Old[intro]: 
"\<rho> obsl' obsl1
 \<Longrightarrow> \<rho> ((OldA a,OldO ou) # obsl') ((a,ou) # obsl1)"
|
Step_New[intro]: 
"\<lbrakk>\<rho> obsl' obsl1; pre_\<rho> s na nou al_oul\<rbrakk> 
 \<Longrightarrow> \<rho> ((NewA na, NewO (\<epsilon>o na (map snd (O (traceOf s (\<epsilon>a na s)))))) # obsl') (al_oul @ obsl1)"


inductive P :: "(('act, 'newact) act' \<times> ('out,'newout) out') list \<Rightarrow> ('state,'act,'out) trace \<Rightarrow> bool" where 
Nil[simp,intro!]: "P [] []"
|
Step_notObs[intro]: 
"\<lbrakk>\<not> \<gamma> trn; P obsl' tr\<rbrakk> 
 \<Longrightarrow> P obsl' (trn # tr)"
|
Step_Old[intro]: 
"\<lbrakk>\<gamma> trn; g trn = (a, ou); P obsl' tr\<rbrakk>  
 \<Longrightarrow> P ((OldA a,OldO ou) # obsl') (trn # tr)"
|
Step_New[intro]: (* For P traces, the chains of old actions corresponding 
to new actions should be produced consecutively:    *)
"\<lbrakk>reach ss; reach s; al = \<epsilon>a na ss; nou = \<epsilon>o na (fst (sstep ss al));
  O (traceOf ss (\<epsilon>a na ss)) = O (traceOf s (\<epsilon>a na s));
  list_ex \<gamma> (traceOf ss al); P obsl' tr1\<rbrakk> 
 \<Longrightarrow> P ((NewA na, NewO (\<epsilon>o na (map snd (O (traceOf s (\<epsilon>a na s)))))) # obsl') 
       ((traceOf s (\<epsilon>a na s)) @ tr1)"


sublocale B? : BD_Security2
where 
validSystemTraces = "validFrom istate" and 
V = V and O = O and TT = "never T" and  B = B  and 
validSystemTraces' = "validFrom' istate" and 
V' = V' and O' = O' and TT' = "never T'" 
and \<pi> = \<pi> and \<rho> = \<rho> 
done
 

(* Preliminary facts: *)
(* They should be eventually moved to an instance of IO_Automaton having standard g: *)
lemma \<gamma>_act_Trans_iff:
  assumes "\<gamma>_act"
  shows "\<gamma> (Trans s1 a o1 s1') = \<gamma> (Trans s2 a o2 s2')"
using assms unfolding \<gamma>_act_def by auto

lemma \<gamma>_act_TransI:
  assumes "\<gamma>_act"
      and "\<gamma> (Trans s1 a o1 s1')"
  shows "\<gamma> (Trans s2 a o2 s2')"
using assms by (auto iff: \<gamma>_act_Trans_iff)

lemma \<gamma>_act_TransE:
  assumes "\<gamma>_act"
      and "\<gamma> (Trans s1 a o1 s1')"
      and "\<not> \<gamma> (Trans s2 a o2 s2')"
  shows "Q"
using assms by (auto iff: \<gamma>_act_Trans_iff)

lemma snd_step_srcOf_tgtOf[simp]: 
"validTrans trn \<Longrightarrow> snd (step (srcOf trn) (fst (g trn))) = tgtOf trn"
by (cases trn) auto

lemma snd_sstep_map_fst_g[simp]: 
"validFrom s tr \<Longrightarrow> tr \<noteq> [] \<Longrightarrow> snd (sstep s (map (fst \<circ> g) tr)) = tgtOf (last tr)"
by (induct tr arbitrary: s) auto

lemma fst_g_actOf[simp]: "fst (g trn) = actOf trn"
by (cases trn) simp

lemma fst_o_g_actOf[simp]: "fst \<circ> g = actOf"
by (rule ext) auto

lemma snd_g_outOf[simp]: "validTrans trn \<Longrightarrow> snd (g trn) = outOf trn"
by (cases trn) simp

lemma \<gamma>_act_O_traceOf: 
assumes g: \<gamma>_act and f: "fst (sstep s al) = fst (sstep ss al)"
shows "O (traceOf s al) = O (traceOf ss al)"
using f
proof(induct al arbitrary: s ss) 
  case (Cons a al)
  thus ?case proof(cases "step s a")
    case (Pair ou s1) note sa = Pair
    show ?thesis
    proof(cases "step ss a") 
       case (Pair ouu s11) note ssa = Pair
       hence [simp]: "ouu = ou" 
       using sa Cons by simp 
       have 1: "fst (Orig.sstep s1 al) = fst (Orig.sstep s11 al)" 
       using Cons(2) using sa ssa by simp
       show ?thesis
       proof(cases "\<gamma> (Trans s a ou s1)") 
         case True hence [simp]: "\<gamma> (Trans ss a ou s11)" 
         using g unfolding \<gamma>_act_def by (metis actOf.simps)       
         show ?thesis using sa ssa using Cons(1)[OF 1] True by simp
       next
         case False hence [simp]: "\<not> \<gamma> (Trans ss a ou s11)" 
         using g unfolding \<gamma>_act_def by (metis actOf.simps)       
         show ?thesis using sa ssa using Cons(1)[OF 1] False by simp
      qed
    qed 
  qed
qed auto

(* If our extended system only deals with sequences of observations, then we can 
easily check smth. stronger than \<phi>_\<epsilon>a: *)
lemma never_\<phi>_\<epsilon>aI: 
assumes "never \<phi> (traceOf ss (\<epsilon>a na ss))"
shows "\<phi>_\<epsilon>a na ss"
using assms unfolding \<phi>_\<epsilon>a_def never_list_ex list_ex_length by auto


(* Facts about P and \<pi>: *)

(* Not used, but interesting: 
lemma P_\<rho>: 
assumes i: \<epsilon>a_inj 
and p: "P obsl' tr" shows "\<rho> obsl' (O tr)"
using p proof(induct rule: P.induct)
  case (Step_New ss al na oul s obsl' tr1)
  thus ?case by (auto intro!: \<rho>.Step_New) (metis i[unfolded \<epsilon>a_inj_def])
qed auto
*)

lemma \<pi>_append: "\<pi> (tr' @ tr1') = \<pi> tr' @ \<pi> tr1'"
proof (induct tr')
  case (Cons trn' tr')
  show ?case proof (cases trn')
    case (Trans s a' ou' s1)
    thus ?thesis using Cons by (cases a') auto
  qed
qed auto

corollary \<pi>_ConsR: "\<pi> (tr' ## trn') = \<pi> tr' @ \<pi> [trn']"
using \<pi>_append . 

lemma P_Nil[simp]:
assumes "never \<gamma> tr"
shows "P [] tr" 
using assms by (induct tr) auto

lemma P_append[simp]:
assumes "P obsl1' tr1" and "P obsl2' tr2"
shows "P (obsl1' @ obsl2') (tr1 @ tr2)"
using assms by (induct obsl1' tr1 rule: P.induct) (auto intro!: P.Step_New)

lemma \<pi>_Cons: "\<pi> (trn' # tr') = \<pi> [trn'] @ \<pi> tr'"
by (metis \<pi>_append append_Cons append_Nil)

lemma validFrom_\<pi>:
assumes "validFrom' s tr'"
shows "validFrom s (\<pi> tr')"
using assms proof(induct tr' arbitrary: s) 
  case (Cons trn' tr')
  hence trn': "validTrans' trn'" and s: "srcOf trn' = s" and tr' :"validFrom' (tgtOf trn') tr'"
  unfolding Prime.validFrom_Cons by auto
  show ?case proof (cases trn')
    case (Trans ss a' ou' s1)
    hence ss: "ss = s" using trn' s by auto
    hence st': "step' s a' = (ou',s1)" using trn' unfolding Trans by auto
    have tr' :"validFrom' s1 tr'" using tr' unfolding Trans by simp
    show ?thesis proof (cases a')
      case (OldA a)
      show ?thesis using st' Cons(1)[OF tr'] unfolding Trans ss OldA 
      by (cases "step s a") (auto simp add: step'_def)
    next
      case (NewA na)
      show ?thesis using st' Cons(1)[OF tr'] unfolding Trans ss NewA 
      by (cases "sstep s (\<epsilon>a na s)", cases "\<epsilon>a na s = []") (auto simp add: step'_def)
    qed
  qed
qed auto

lemma \<pi>_validSystemTraces: \<pi>_validSystemTraces
using validFrom_\<pi> unfolding \<pi>_validSystemTraces_def by auto

lemma \<pi>_TT: \<pi>_TT
unfolding \<pi>_TT_def proof safe
  fix tr' assume "validFrom' istate tr'" and 1: "never T' tr'"
  from 1 show "never T (\<pi> tr')" proof(induction tr')
    case (Cons trn' tr')
    thus ?case 
    proof(cases trn')
      case (Trans s a' ou' s1)
      show ?thesis using Cons unfolding Trans 
      by (cases a') (auto simp add: never_list_ex)
    qed
  qed auto
qed

lemma \<phi>_\<epsilon>a_list_ex_unique_nth: 
fixes s na defines "tr \<equiv> traceOf s (\<epsilon>a na s)"
assumes i: "\<phi>_\<epsilon>a na s" and l: "list_ex \<phi> tr"
shows "\<exists>!i. i < length tr \<and> \<phi> (tr!i)"
proof-
  have "\<exists> i. i < length tr \<and> \<phi> (tr!i)" using l unfolding list_ex_length by auto
  thus ?thesis using i unfolding tr_def \<phi>_\<epsilon>a_def 
  by auto (metis linorder_neqE_nat)
qed


lemma \<phi>_\<epsilon>a_list_ex_unique: 
fixes s na defines "tr \<equiv> traceOf s (\<epsilon>a na s)"
assumes i: "\<phi>_\<epsilon>a na s" and l: "list_ex \<phi> tr"
shows "\<exists>! trn. trn \<in>\<in> tr \<and> \<phi> trn"
using assms \<phi>_\<epsilon>a_list_ex_unique_nth length_traceOf unfolding in_set_conv_nth 
by simp (metis length_traceOf)

lemma traceOf_V_list_ex_SOME[simp]:
fixes s na defines "tr \<equiv> traceOf s (\<epsilon>a na s)"
assumes i: "\<phi>_\<epsilon>a na s" and l: "list_ex \<phi> tr"
shows "V tr = [f (SOME trn. trn \<in>\<in> tr \<and> \<phi> trn)]"
proof-
  note 1 = \<phi>_\<epsilon>a_list_ex_unique[OF i l[unfolded tr_def]]
  note 2 = \<phi>_\<epsilon>a_list_ex_unique_nth[OF i l[unfolded tr_def]]
  obtain i where tri: "i < length tr" "\<phi> (tr!i)" using l unfolding list_ex_length by auto
  have s: "(SOME trn. trn \<in>\<in> tr \<and> \<phi> trn) = tr!i" 
  using tri 1 
  unfolding tr_def by (intro some1_equality) auto
  obtain tr1 tr2 where tr: "tr = tr1 @ [tr!i] @ tr2" using tri 
  by (metis append_Cons append_Nil id_take_nth_drop)
  have "never \<phi> tr1 \<and> never \<phi> tr2" unfolding never_list_ex 
  using append_ex_unique_list_ex[OF 2 tr[unfolded tr_def] tri(2)[unfolded tr_def]] .
  thus ?thesis unfolding s using tri by (subst tr) simp
qed

lemma V_\<pi>_aux: 
assumes i: "\<And> na ss. reach ss \<Longrightarrow> \<phi>_\<epsilon>a na ss"
and s: "reach s" and v: "validFrom' s tr'"
shows "V (\<pi> tr') = V' tr'"
using s v proof(induction tr' arbitrary: s)
  case (Cons trn' tr' ss)
  show ?case 
  proof(cases trn')
    case (Trans s a' ou' s1)
    with `validFrom' ss (trn' # tr')` have ss[simp]: "ss = s" and 
    trn': "validTrans' trn'" and tr': "validFrom' s1 tr'" by auto
    show ?thesis
    proof(cases a')
      case (OldA a)
      have 1: "\<phi>' trn' = \<phi> (Trans s a (fst (step s a)) (snd (step s a)))" 
      unfolding Trans OldA by simp
      have "reach s1" using trn' `reach ss` unfolding ss Trans OldA 
      by (auto simp: step'_def split: prod.splits intro: reach_PairI) 
      hence [simp]: "V (\<pi> tr') = V' tr'" using tr' Cons by auto
      show ?thesis unfolding Trans OldA using 1 by auto
    next
      case (NewA na) 
      have s1: "reach s1" using trn' `reach ss` unfolding ss Trans NewA 
      by (auto simp: step'_def split: prod.splits intro: reach_sstep)
      hence [simp]: "V (\<pi> tr') = V' tr'" using tr' Cons by auto
      let ?tr = "traceOf s (\<epsilon>a na s)"
      let ?trn = "SOME trn. trn \<in>\<in> ?tr \<and> \<phi> trn"
      show ?thesis proof(cases "list_ex \<phi> ?tr")
        case True
        hence "V ?tr = [f ?trn]" 
        using traceOf_V_list_ex_SOME `reach ss`[unfolded ss] i by simp
        thus ?thesis using True unfolding Trans NewA by simp
      next 
        case False
        hence "V ?tr = []" using V_Nil_list_ex by simp
        thus ?thesis using False unfolding Trans NewA by simp
      qed
    qed
  qed
qed auto

lemma V_\<pi>: assumes "\<And> na ss. reach ss \<Longrightarrow> \<phi>_\<epsilon>a na ss"   
shows V_\<pi> 
using assms V_\<pi>_aux reach.Istate unfolding V_\<pi>_def by auto

lemma length_filtermap_filter: "length (filtermap pred func tr) = length (filter pred tr)"
by (induction tr) auto

lemma list_all_fst_O_traceOf[simp]:
assumes "list_all \<gamma> (traceOf s al)" 
shows "map fst (O (traceOf s al)) = al"
using assms proof(induct al arbitrary: s)
  case (Cons a al)
  obtain ou s1 where sa: "step s a = (ou, s1)" by force
  show ?case
  using sa Cons by (cases "\<gamma> (Trans s a ou s1)") auto
qed auto

lemma fst_O_actOf:
  assumes "a \<in>\<in> map fst (O tr)"
  shows "a \<in>\<in> map actOf tr"
using assms proof (induction tr)
  case (Cons trn tr) then show ?case by (cases trn, cases "\<gamma> trn") auto
qed auto

lemma map_fst_g_traceOf[simp]:
"map (fst \<circ> g) (traceOf s al) = al"
by (induct al arbitrary: s) (auto split: prod.splits)

lemma map_snd_g_traceOf[simp]:
"map (snd \<circ> g) (traceOf s al) = fst (sstep s al)"
by (induct al arbitrary: s) (auto split: prod.splits)

lemma traceOf_map_fst_g[simp]: 
assumes "validFrom s tr"
shows "traceOf s (map (fst o g) tr) = tr"
using assms proof(induction tr arbitrary: s) 
  case (Cons trn tr s)
  thus ?case 
  proof(cases trn)
    case (Trans ss a ou s1)
    hence "validTrans trn" and ss: "ss = s" and tr: "validFrom s1 tr"
    using `validFrom s (trn # tr)` unfolding validFrom_Cons by auto
    hence "step s a = (ou, s1)" using Trans by auto
    thus ?thesis unfolding Trans ss using Cons tr by auto
  qed
qed auto

lemma fst_sstep_g[simp]: 
assumes "validFrom s tr"
shows "fst (sstep s (map (fst \<circ> g) tr)) = map (snd \<circ> g) tr"
using assms proof(induction tr arbitrary: s)
  case (Cons trn tr s)
  hence 1: "validTrans trn \<and> srcOf trn = s \<and> validFrom (tgtOf trn) tr" by auto
  thus ?case proof(cases trn) 
    case (Trans ss a ou s1)
    thus ?thesis using Cons 1 by (force split: prod.splits) 
  qed
qed auto

lemma list_all_snd_O_traceOf[simp]:
assumes "list_all \<gamma> (traceOf s al)" and "sstep s al = (ol, s1)"
shows "map snd (O (traceOf s al)) = ol"
using assms proof(induct al arbitrary: s ol s1)
  case (Cons a al s ol s1)
  moreover obtain ou s2 where sa: "step s a = (ou, s2)" by force
  ultimately show ?case
  by (cases "\<gamma> (Trans s a ou s2)") (auto split: prod.splits)
qed auto

lemma \<pi>_\<rho>_aux: 
assumes "reach s" and "validFrom' s tr'"
shows "\<rho> (O' tr') (O (\<pi> tr'))"
using assms proof(induct tr' arbitrary: s)
  case (Cons trn' tr' s)
  hence trn': "validTrans' trn'" and ist: "srcOf trn' = s" and tr' :"validFrom' (tgtOf trn') tr'"
  by auto
  show ?case proof(cases trn')
    case (Trans ss a' ou' s1)
    hence ss: "ss = s" using trn' ist by auto
    hence st': "step' s a' = (ou',s1)" using trn' unfolding Trans by auto
    have tr' :"validFrom' s1 tr'" using tr' unfolding Trans by simp 
    show ?thesis proof (cases a')
      case (OldA a)
      then obtain ou where ou'st: "ou' = OldO ou" "step s a = (ou,s1)" 
      using st' by (cases "step s a") (auto simp add: step'_def)
      have s1: "reach s1" using `reach s` trn' unfolding Trans ss OldA 
      by (auto simp add: step'_def split: prod.splits) (metis reach_PairI)
      have [simp]: "\<rho> (O' tr') (BD_Security_IO_Aut.O \<gamma> g (\<pi> tr'))" using Cons(1)[OF s1 tr'] .
      show ?thesis using st' Cons(1)[OF s1 tr'] unfolding Trans ss OldA s1
      by (cases "\<gamma> (Trans s a ou s1)") (auto simp add: ou'st)
    next
      case (NewA na)
      then obtain oul where ou'st: "ou' = NewO (\<epsilon>o na oul)" "sstep s (\<epsilon>a na s) = (oul,s1)" 
      using st' by (cases "sstep s (\<epsilon>a na s)") (auto simp add: step'_def)
      have s1: "reach s1" using `reach s` trn' unfolding Trans ss NewA 
      by (auto simp add: step'_def split: prod.splits) (metis reach_sstep)
      have [simp]: "\<rho> (O' tr') (BD_Security_IO_Aut.O \<gamma> g (\<pi> tr'))" using Cons(1)[OF s1 tr'] .
      show ?thesis 
      unfolding Trans ss NewA using O_Nil_list_ex[of "traceOf s (\<epsilon>a na s)"] s1 `reach s`
      by (cases "list_ex \<gamma> (Orig.traceOf s (\<epsilon>a na s))") (auto simp add: ou'st intro!: \<rho>.Step_New exI[of _ s])      
    qed
  qed
qed auto

lemma \<pi>_\<rho>: assumes \<epsilon>a_\<gamma>   shows \<pi>_\<rho>
using assms \<pi>_\<rho>_aux reach.Istate unfolding \<pi>_\<rho>_def by auto

lemma pre_\<rho>_list_ex_ex: 
assumes s: "reach s" and ss: "reach ss"
and nou: "nou = \<epsilon>o na (fst (sstep ss (\<epsilon>a na ss)))"
and O: "O (traceOf ss (\<epsilon>a na ss)) = O (traceOf s (\<epsilon>a na s))"
defines "tr \<equiv> traceOf s (\<epsilon>a na s)"
assumes l: "list_ex \<gamma> tr"
shows 
"\<exists> trn'. 
   validTrans' trn' \<and> srcOf trn' = s \<and> 
   tgtOf trn' = tgtOf (last tr) \<and> 
   \<gamma>' trn' \<and> g' trn' = (NewA na, NewO (\<epsilon>o na (map snd (O (traceOf s (\<epsilon>a na s)))))) \<and> \<pi> [trn'] = tr"
proof- 
  let ?trn' = "Trans s (NewA na) (NewO (\<epsilon>o na (fst (sstep s (\<epsilon>a na s))))) (snd (sstep s (\<epsilon>a na s)))"
  show ?thesis apply(intro exI[of _ ?trn']) 
  using assms apply (cases "sstep s (\<epsilon>a na s)", auto simp add: step'_def)
  by (metis traceOf.simps(1) list_ex_simps(2) sstep_tgtOf_traceOf[of "\<epsilon>a na s" s])
qed

lemma pull_aux: 
assumes pv: "P ol' tr" "reach s" "validFrom s tr" 
(* The hypothesis "\<rho> ol' (O tr)" would be redundant, since 
for our particular definition of P it happens to be implied by "P ol' tr" *)
shows "\<exists>tr'. validFrom' s tr' \<and> O' tr' = ol' \<and> \<pi> tr' = tr"
using pv proof(induction tr arbitrary: s rule: P.induct)
  case (Nil s) thus ?case by (intro exI[of _ "[]"]) auto
next
  case (Step_notObs trn obsl' tr s) 
  let ?s1 = "tgtOf trn"
  have trn: "validTrans trn" and ist: "srcOf trn = s" and tr :"validFrom ?s1 tr" 
  using `validFrom s (trn # tr)` by auto   
  hence [simp]: "step s (actOf trn) = (outOf trn, tgtOf trn)" by auto
  have s1: "reach ?s1" using `reach s` trn ist valid_reach_src_tgt[of "[trn]"] by auto
  from Step_notObs s1 obtain tr' where 1: "validFrom' ?s1 tr' \<and> O' tr' = obsl' \<and> \<pi> tr' = tr" by auto
  let ?trn' = "Trans s (OldA (actOf trn)) (OldO (outOf trn)) ?s1"
  show ?case 
  using  `\<not> \<gamma> trn` trn ist 1 by (intro exI[of _ "?trn' # tr'"]) (fastforce simp: step'_def)
next
  case (Step_Old trn a ou obsl' tr s) 
  let ?s1 = "tgtOf trn"
  have trn: "validTrans trn" and ist: "srcOf trn = s" and tr :"validFrom ?s1 tr" 
  using `validFrom s (trn # tr)` by auto   
  hence [simp]: "step s (actOf trn) = (outOf trn, tgtOf trn)" by auto
  have s1: "reach ?s1" using `reach s` trn ist valid_reach_src_tgt[of "[trn]"] by auto
  from Step_Old s1 obtain tr' where 1: "validFrom' ?s1 tr' \<and> O' tr' = obsl' \<and> \<pi> tr' = tr" by auto
  let ?trn' = "Trans s (OldA (actOf trn)) (OldO (outOf trn)) ?s1"
  show ?case proof(intro exI[of _ "?trn' # tr'"], safe) 
    show "validFrom' s (?trn' # tr')" using 1 by (simp add: step'_def)
  next
    have "\<gamma>' ?trn'" using ist trn `\<gamma> trn` by auto
    thus "O' (?trn' # tr') = (OldA a,OldO ou) # obsl'" using trn 1 `g trn = (a, ou)` 
    by (cases trn) auto    
  qed(insert 1 trn ist, auto)
next
  case (Step_New ss s al na nou obsl' tr1 sa)
  hence al[simp]: "al \<noteq> []" by auto
  with Step_New(3,6) have "O (traceOf ss (\<epsilon>a na ss)) \<noteq> []" using O_Nil_list_ex by auto
  with Step_New(5) have l': "list_ex \<gamma> (traceOf s (\<epsilon>a na s))" using O_Nil_list_ex by auto
  with Step_New(5) have \<epsilon>a_s: "\<epsilon>a na s \<noteq> []" by auto
  let ?tr = "traceOf s (\<epsilon>a na s)"  let ?s1 = "tgtOf (last ?tr)"
  have tr: "validFrom s ?tr" and tr1 :"validFrom ?s1 tr1" and sa[simp]: "sa = s"
  using `validFrom sa (?tr @ tr1)` \<epsilon>a_s by simp_all (metis validFrom_def traceOf_Nil validFrom_traceOf al)
  have s1: "reach ?s1" using `reach s` tr valid_reach_src_tgt[of ?tr] \<epsilon>a_s by auto
  thm Step_New(2,1,3-5,6)[unfolded `al = \<epsilon>a na ss`]
  from pre_\<rho>_list_ex_ex[OF Step_New(2,1,4,5)[unfolded `al = \<epsilon>a na ss`] l']
  obtain trn' where 
    trn': "validTrans' trn' \<and> srcOf trn' = s \<and> tgtOf trn' = ?s1 \<and> \<gamma>' trn' \<and> 
         g' trn' = (NewA na, NewO (\<epsilon>o na (map snd (O (traceOf s (\<epsilon>a na s)))))) \<and> \<pi> [trn'] = ?tr" 
  unfolding `al = \<epsilon>a na ss` by auto
  obtain tr1' where tr1': "validFrom' ?s1 tr1' \<and> O' tr1' = obsl' \<and> \<pi> tr1' = tr1" 
  using Step_New(8)[OF s1 tr1] by auto
  show ?case using trn' tr1' by (intro exI[of _ "trn' # tr1'"]) (subst \<pi>_Cons, simp)
qed

lemma pull: "pull P"
using pull_aux reach.Istate unfolding pull_def by auto 

lemma adjust_aux:
assumes a : local_adjust
and s: "reach s" and v: "validFrom s tr" and r: "\<rho> obsl' (O tr)"
shows "\<exists>tr1. validFrom s tr1 \<and> O tr1 = O tr \<and> V tr1 = V tr \<and> P obsl' tr1" (is "?phi s tr")
proof-
  {fix ol assume "\<rho> obsl' ol"
   hence "\<And> s tr. O tr = ol \<Longrightarrow> reach s \<Longrightarrow> validFrom s tr \<Longrightarrow> ?phi s tr"
   proof(induction rule: \<rho>.induct)
     case (Nil s tr) thus ?case by (intro exI[of _ "tr"]) (auto simp: O_Nil_never)
   next
     case (Step_Old obsl' obsl1 a ou s tr)
     from O_eq_Cons[OF `O tr = (a, ou) # obsl1`] obtain trn trl trr where 
     tr: "tr = trl @ (trn # trr)" and ntrl: "never \<gamma> trl" 
     and trn: "\<gamma> trn" "g trn = (a, ou)" and Otr1: "O trr = obsl1" by auto
     from ntrl `validFrom s tr` have trl: "P [] trl" "validFrom s trl" and vtrn: "validTrans trn"
     unfolding tr by auto
     let ?s1 = "tgtOf trn"
     have vtr1: "validFrom ?s1 trr" using `validFrom s tr` unfolding tr by auto
     have s1: "reach ?s1" using `validFrom s tr` unfolding tr 
     by simp (metis validFrom_def validTrans valid_reach_src_tgt reach_PairI Step_Old.prems(2))+
     from Step_Old(2)[OF Otr1 s1 vtr1] obtain trr1 where trr1: 
     "validFrom ?s1 trr1 \<and> O trr1 = O trr \<and> V trr1 = V trr" and P: "P obsl' trr1" by auto
     show ?case proof(rule exI[of _ "trl @ [trn] @ trr1"], safe)
       show "validFrom s (trl @ [trn] @ trr1)" 
       using trl trr1 `validFrom s tr` unfolding tr by auto 
       show "O (trl @ [trn] @ trr1) = O tr" unfolding tr using trr1 trn by simp
       show "V (trl @ [trn] @ trr1) = V tr" unfolding tr using trr1 trn  by(cases "\<phi> trn") auto
       have "P ((OldA a,OldO ou) # obsl') (trn # trr1)" using P.Step_Old[OF trn P] .
       from P_append[OF trl(1) this] 
       show  "P ((OldA a,OldO ou) # obsl') (trl @ [trn] @ trr1)" by simp
     qed
   next
     case (Step_New obsl' obsl1 ss na nou al_oul s Tr) 
     from ` pre_\<rho> ss na nou al_oul` obtain al oul where ss: "reach ss" and 
     al: "al = \<epsilon>a na ss" and oul: "oul = fst (sstep ss al)"
     and nou: "nou = \<epsilon>o na oul" and l: "list_ex \<gamma> (traceOf ss al)"
     and al_oul: "al_oul = O (traceOf ss al)" by auto
     hence al_oulNE: "al_oul \<noteq> []" using O_Nil_list_ex by auto
     have alNE: "al \<noteq> []" using l by auto
     from O_eq_append[OF `O Tr = al_oul @ obsl1`] obtain tr Tr1 where 
     Tr: "Tr = tr @ Tr1" and Otr: "O tr = al_oul" and OTr1: "O Tr1 = obsl1" by auto
     hence trNE[simp]: "tr \<noteq> []" using al_oulNE by (cases tr) auto
     let ?S1 = "tgtOf (last tr)"
     have tr: "validFrom s tr" and Tr1: "validFrom ?S1 Tr1" 
     using `validFrom s Tr` al_oul unfolding Tr by auto
     have S1: "reach ?S1" using tr `reach s` valid_reach_src_tgt[of tr]  
     by (metis validFrom_def trNE)
     from Step_New(3)[OF OTr1 S1 Tr1] obtain Tr1a where 
     Tr1a: "validFrom ?S1 Tr1a" and OTr1a: "O Tr1a = O Tr1"
     and VTr1a: "V Tr1a = V Tr1" and P: "P obsl' Tr1a" by auto
     from a[unfolded local_adjust_def] have a: "local_adjust_param na ss s tr" by auto
     from l tr Otr[unfolded al_oul] al this[unfolded local_adjust_param_def] `reach s` ss
     obtain tr1 tr2 where lr: "local_right na ss s tr tr1 tr2" by auto
     let ?s1 = "if tr1 = [] then s else tgtOf (last tr1)" let ?trr = "traceOf ?s1 (\<epsilon>a na ?s1)" 
     let ?s2 = "tgtOf (last ?trr)"
     from lr[unfolded local_right_def] have tr1: "validFrom s tr1"
     and tr2: "validFrom ?s2 tr2"
     and ntr1: "never \<gamma> tr1" and ntr2: "never \<gamma> tr2" and Otrr: "O ?trr = al_oul"
     and V: "V (tr1 @ ?trr @ tr2) = V tr" 
     and tgtOf: "tgtOf (last (tr1 @ ?trr @ tr2)) = ?S1"
     using al oul al_oul by auto
     have s1: "reach ?s1" using `reach s` tr1 valid_reach_src_tgt[of tr1] 
     by (cases "tr1 = []") (auto simp: validFrom_def)
     then have OO: "O (traceOf ?s1 (\<epsilon>a na ?s1)) = O (traceOf ss (\<epsilon>a na ss))" 
     using ss l Otrr al_oul al by auto  
     from Otrr have O: "O (tr1 @ ?trr @ tr2) = O tr" using ntr1 ntr2 unfolding Otr by simp
     show ?case proof(intro exI[of _ "(tr1 @ ?trr @ tr2) @ Tr1a"] conjI)
       have "validFrom s tr1" by fact
       moreover have "validFrom ?s1 ?trr" by (metis validFrom_traceOf)
       ultimately have "validFrom s (tr1 @ ?trr)" by auto
       moreover have NE: "?trr \<noteq> []" using Otrr al_oulNE by auto
       moreover have "validFrom ?s2 tr2" by fact
       ultimately have "validFrom s (tr1 @ ?trr @ tr2)" by auto
       moreover have "validFrom (tgtOf (last (tr1 @ ?trr @ tr2))) Tr1a" using Tr1a tgtOf by auto
       moreover have "tr1 @ ?trr @ tr2 \<noteq> []" using NE by simp
       ultimately show "validFrom s ((tr1 @ ?trr @ tr2) @ Tr1a)" by (subst validFrom_append) blast
     next
       show "O ((tr1 @ ?trr @ tr2) @ Tr1a) = O Tr" by (subst O_append, unfold O OTr1a Tr) simp
     next
       show "V ((tr1 @ ?trr @ tr2) @ Tr1a) = V Tr" using V VTr1a Tr by simp
     next
       have 1: "P obsl' (tr2 @ Tr1a)" using P_append[OF P_Nil[OF ntr2] P] by simp
       have "list_ex \<gamma> ?trr" using Otrr al_oulNE by (metis O_Nil_list_ex)
       from P.Step_New[OF ss s1 al nou[unfolded oul] Otrr[unfolded al_oul al, symmetric] l 1]
       show "P ((NewA na, NewO (\<epsilon>o na (map snd (O (traceOf ss (\<epsilon>a na ss)))))) # obsl') 
               ((tr1 @ ?trr @ tr2) @ Tr1a)" 
      using OO P_append[OF P_Nil[OF ntr1] _] by simp
     qed
   qed
  }
  thus ?thesis using assms by auto
qed

lemma adjust: 
assumes a : local_adjust
shows "adjust P"
using assms adjust_aux reach.Istate unfolding adjust_def by auto

theorem stepwise_borrow_secure:
assumes (* should be immediate to check: *) 
"\<And> na ss. reach ss \<Longrightarrow> \<phi>_\<epsilon>a na ss" 
and  (* could be tedious to check: *) local_adjust
shows "secure \<Longrightarrow> secure'"
using assms \<pi>_TT \<pi>_validSystemTraces V_\<pi> \<pi>_\<rho> pull adjust
by (intro borrow_secure) auto


(* Next we focus on further simplifying local_adjust: *)

lemma validFrom_traceOf_al:
  assumes "validFrom s tr"
  obtains al where "tr = traceOf s al"
using assms proof (induction tr arbitrary: s thesis)
  case Nil
    show thesis by (intro Nil(1)[of "[]"]) auto
next
  case (Cons trn tr s thesis)
    obtain a where "trn = transOf s a" using Cons(3) unfolding validFrom_Cons by (cases trn, auto) (metis fst_conv snd_conv)
    moreover then have "validFrom (snd (step s a)) tr" using Cons(3) unfolding validFrom_Cons by auto
    moreover then obtain al where "tr = traceOf (snd (step s a)) al" using Cons(1) by blast
    ultimately have "trn # tr = traceOf s (a # al)" by (cases "step s a") auto
    from Cons(2)[OF this] show thesis .
qed



(* Below, Bw indicates the desire to move backward and 
          Fw the desire to move forward
*)
(*definition "O_\<epsilon>a na ss Bw Fw
 \<equiv> \<forall>s tr.
      reach s \<and> validFrom s tr 
    \<and> O tr = O (traceOf ss (\<epsilon>a na ss))
    \<and> list_ex \<gamma> (traceOf ss (\<epsilon>a na ss))
    \<and> list_all (\<lambda>trn. \<gamma> trn \<or> (\<not>Bw trn \<and> \<not>Fw trn)) tr
  \<longrightarrow> O (traceOf s (\<epsilon>a na s)) = O tr
    \<and> V (traceOf s (\<epsilon>a na s)) = V tr
    \<and> tgtOf (last (traceOf s (\<epsilon>a na s))) = tgtOf (last tr)"  *)

definition "equivFiltermap pred func a b \<equiv> 
  (pred a \<longleftrightarrow> pred b) \<and> (pred a \<longrightarrow> func a = func b)"

lemma equivFiltermap_V: 
assumes "\<not> (\<phi> trn \<and> \<phi> trn')" and 
"equivFiltermap \<phi> f trn1 trn" and "equivFiltermap \<phi> f trn1' trn'"
shows "V [trn, trn'] = V [trn1', trn1]"
using assms unfolding equivFiltermap_def by auto

lemma equivFiltermap_O: 
assumes "\<not> (\<gamma> trn \<and> \<gamma> trn')" and 
"equivFiltermap \<gamma> g trn1 trn" and "equivFiltermap \<gamma> g trn1' trn'"
shows "O [trn, trn'] = O [trn1', trn1]"
using assms unfolding equivFiltermap_def by auto

definition "\<epsilon>a_commute_goal na ss obsl Bw Fw Inv s a a'
 \<equiv> let trn = transOf s a; trn' = transOf (snd (step s a)) a' in
      (let trn1' = transOf s a'; trn1 = transOf (snd (step s a')) a in
       (commute s a a' 
        \<and> (Bw (obsl @ O [trn1']) trn1 \<longleftrightarrow> Bw obsl trn) \<and> (Bw obsl trn1' \<longleftrightarrow> Bw (obsl @ O [trn]) trn')
        \<and> (Fw (obsl @ O [trn1']) trn1 \<longleftrightarrow> Fw obsl trn) \<and> (Fw obsl trn1' \<longleftrightarrow> Fw (obsl @ O [trn]) trn')
        \<and> (Inv s \<longrightarrow> Inv (snd (step s a)) \<longrightarrow> Inv (snd (step s a')))  
        \<and> \<not> (\<phi> trn \<and> \<phi> trn') \<and> \<not> (\<gamma> trn \<and> \<gamma> trn')
        \<and> equivFiltermap \<phi> f trn1 trn \<and> equivFiltermap \<gamma> g trn1 trn
        \<and> equivFiltermap \<phi> f trn1' trn' \<and> equivFiltermap \<gamma> g trn1' trn')
       \<or> 
       (absorb s a a' \<and> \<not> \<gamma> trn \<and> \<not> \<phi> trn 
        \<and> (Bw obsl trn1' \<longrightarrow> Bw obsl trn \<or> Bw (obsl @ O [trn]) trn')
        (*\<and> (Fw trn1' \<longleftrightarrow> Fw trn') *)
        \<and> (Inv s \<longrightarrow> Inv (snd (step s a)) \<longrightarrow> Inv (snd (step s a')))
        \<and> equivFiltermap \<phi> f trn1' trn' \<and> equivFiltermap \<gamma> g trn1' trn'))" 

lemma commute_\<epsilon>a_commute_goalI:
fixes na ss Bw Fw Inv s a a'
defines trn: "trn \<equiv> transOf s a" and trn': "trn' \<equiv> transOf (snd (step s a)) a'"
    and trn1': "trn1' \<equiv> transOf s a'" and trn1: "trn1 \<equiv> transOf (snd (step s a')) a"
assumes "commute s a a'"
        "Bw (obsl @ O [trn1']) trn1 \<longleftrightarrow> Bw obsl trn" "Bw obsl trn1' \<longleftrightarrow> Bw (obsl @ O [trn]) trn'"
        "Fw (obsl @ O [trn1']) trn1 \<longleftrightarrow> Fw obsl trn" "Fw obsl trn1' \<longleftrightarrow> Fw (obsl @ O [trn]) trn'"
        "Inv s \<Longrightarrow> Inv (snd (step s a)) \<Longrightarrow> Inv (snd (step s a'))"
        "\<not> (\<phi> trn \<and> \<phi> trn')" "\<not> (\<gamma> trn \<and> \<gamma> trn')"
        "equivFiltermap \<phi> f trn1 trn" "equivFiltermap \<gamma> g trn1 trn"
        "equivFiltermap \<phi> f trn1' trn'" "equivFiltermap \<gamma> g trn1' trn'"
shows "\<epsilon>a_commute_goal na ss obsl Bw Fw Inv s a a'"
using assms unfolding \<epsilon>a_commute_goal_def trn trn' trn1' trn1 by auto

lemma absorb_\<epsilon>a_commute_goalI:
fixes na ss Bw Fw Inv s a a'
defines trn: "trn \<equiv> transOf s a" and trn': "trn' \<equiv> transOf (snd (step s a)) a'"
    and trn1': "trn1' \<equiv> transOf s a'" and trn1: "trn1 \<equiv> transOf (snd (step s a')) a"
assumes "absorb s a a'" "\<not> \<gamma> trn" "\<not> \<phi> trn"
        "Bw obsl trn1' \<longrightarrow> Bw obsl trn \<or> Bw (obsl @ O [trn]) trn'" (*"Fw trn1' \<longleftrightarrow> Fw trn'"*)
        "Inv s \<Longrightarrow> Inv (snd (step s a)) \<Longrightarrow> Inv (snd (step s a'))"
        "equivFiltermap \<phi> f trn1' trn' \<and> equivFiltermap \<gamma> g trn1' trn'"
shows "\<epsilon>a_commute_goal na ss obsl Bw Fw Inv s a a'"
using assms unfolding \<epsilon>a_commute_goal_def trn trn' trn1' trn1 by auto

lemma non_\<gamma>_equivFiltermap[simp]: 
assumes "\<not> \<gamma> trn" and "\<not> \<gamma> trn1"
shows "equivFiltermap \<gamma> g trn1 trn"
using assms unfolding equivFiltermap_def by auto

lemma non_\<phi>_equivFiltermap[simp]: 
assumes "\<not> \<phi> trn" and "\<not> \<phi> trn1"
shows "equivFiltermap \<phi> f trn1 trn"
using assms unfolding equivFiltermap_def by auto


(* Helper predicate stating that S, which might depend on a list of previous observations obsl,
   never holds for any unobservable transition in tr *)
fun never_n\<gamma> where
  "never_n\<gamma> S obsl (trn # tr) = (if \<gamma> trn then never_n\<gamma> S (obsl ## g trn) tr
                                 else (\<not>S obsl trn) \<and> never_n\<gamma> S obsl tr)"
| "never_n\<gamma> S obsl [] = True"

lemma never_n\<gamma>_append:
"never_n\<gamma> S obsl (tr1 @ tr2) \<longleftrightarrow> never_n\<gamma> S obsl tr1 \<and> never_n\<gamma> S (obsl @ O tr1) tr2"
by (induction tr1 arbitrary: obsl) auto

fun filter_n\<gamma> where
  "filter_n\<gamma> S obsl (trn # tr) = (if \<gamma> trn then filter_n\<gamma> S (obsl ## g trn) tr
                                  else (if S obsl trn then trn # filter_n\<gamma> S obsl tr
                                        else filter_n\<gamma> S obsl tr))"
| "filter_n\<gamma> S obsl [] = []"

lemma filter_n\<gamma>_append:
"filter_n\<gamma> S obsl (tr1 @ tr2) = filter_n\<gamma> S obsl tr1 @ filter_n\<gamma> S (obsl @ O tr1) tr2"
by (induction tr1 arbitrary: obsl) auto

lemma never_n\<gamma>_filter_n\<gamma>_Nil: "never_n\<gamma> S obsl tr \<longleftrightarrow> filter_n\<gamma> S obsl tr = []"
by (induction tr arbitrary: obsl) auto

lemma Cons_eq_filter_n\<gamma>D:
assumes "trn # trns = filter_n\<gamma> S obsl tr"
obtains trl trr where "tr = trl @ trn # trr" and "never_n\<gamma> S obsl trl"
  and "S (obsl @ O trl) trn" and "\<not>\<gamma> trn"
  and "filter_n\<gamma> S (obsl @ O trl) trr = trns"
using assms proof (induction tr arbitrary: obsl)
  case (Cons trn' tr)
  consider (\<gamma>) "\<gamma> trn'" | (S) "S obsl trn'" "\<not>\<gamma> trn'" | (n\<gamma>S) "\<not>S obsl trn'" "\<not>\<gamma> trn'" by auto
  then show ?case proof cases
    case \<gamma>
    show ?thesis proof (rule Cons.IH[of "obsl ## g trn'"])
      show "trn # trns = filter_n\<gamma> S (obsl ## g trn') tr" using \<gamma> Cons(3) by auto
    next
      fix trl trr
      assume "tr = trl @ trn # trr" "never_n\<gamma> S (obsl ## g trn') trl"
        "S (obsl ## g trn' @ O trl) trn" "\<not>\<gamma> trn" "filter_n\<gamma> S (obsl ## g trn' @ O trl) trr = trns"
      then show ?thesis using \<gamma> by (intro Cons(2)[of "trn' # trl" trr]) auto
    qed
  next
    case S
    moreover then have "trn' = trn" and "filter_n\<gamma> S obsl tr = trns" using Cons(3) by auto
    ultimately show ?thesis by (intro Cons(2)[of "[]" tr]) auto
  next
    case n\<gamma>S
    show ?thesis proof (rule Cons.IH[of obsl])
      show "trn # trns = filter_n\<gamma> S obsl tr" using n\<gamma>S Cons(3) by auto
    next
      fix trl trr
      assume "tr = trl @ trn # trr" "never_n\<gamma> S obsl trl"
        "S (obsl @ O trl) trn" "\<not>\<gamma> trn" "filter_n\<gamma> S (obsl @ O trl) trr = trns"
      then show ?thesis using n\<gamma>S by (intro Cons(2)[of "trn' # trl" trr]) auto
    qed
  qed
qed auto

lemma snoc_eq_filter_n\<gamma>D:
assumes "trns ## trn = filter_n\<gamma> S obsl tr"
obtains trl trr where "tr = trl @ trn # trr" and "never_n\<gamma> S (obsl @ (O (trl ## trn))) trr"
  and "S (obsl @ O trl) trn" and "\<not>\<gamma> trn"
  and "filter_n\<gamma> S obsl trl = trns"
using assms proof (induction tr arbitrary: obsl rule: rev_induct)
  case (snoc trn' tr obsl)
  consider (\<gamma>) "\<gamma> trn'" | (S) "S (obsl @ O tr) trn'" "\<not>\<gamma> trn'"
    | (n\<gamma>S) "\<not>S (obsl @ O tr) trn'" "\<not>\<gamma> trn'" by auto
  then show ?case proof cases
    case \<gamma>
    show ?thesis proof (rule snoc.IH[of "obsl"])
      show "trns ## trn = filter_n\<gamma> S obsl tr" using \<gamma> snoc(3) by (auto simp: filter_n\<gamma>_append)
    next
      fix trl trr
      assume "tr = trl @ trn # trr" "never_n\<gamma> S (obsl @ (O (trl ## trn))) trr"
        "S (obsl @ O trl) trn" "\<not>\<gamma> trn" "filter_n\<gamma> S obsl trl = trns"
      then show ?thesis using \<gamma> by (intro snoc(2)[of trl "trr ## trn'"]) (auto simp: never_n\<gamma>_append)
    qed
  next
    case S
    moreover then have "trn' = trn" and "filter_n\<gamma> S obsl tr = trns" using snoc(3)
      by (auto simp: filter_n\<gamma>_append split: if_splits)
    ultimately show ?thesis by (intro snoc(2)[of tr "[]"]) auto
  next
    case n\<gamma>S
    show ?thesis proof (rule snoc.IH[of obsl])
      show "trns ## trn = filter_n\<gamma> S obsl tr" using n\<gamma>S snoc(3) by (auto simp: filter_n\<gamma>_append)
    next
      fix trl trr thm snoc(2)[of trl "trr ## trn'"]
      assume "tr = trl @ trn # trr" "never_n\<gamma> S (obsl @ (O (trl ## trn))) trr"
        "S (obsl @ O trl) trn" "\<not>\<gamma> trn" "filter_n\<gamma> S obsl trl = trns"
      then show ?thesis using n\<gamma>S
        by (intro snoc(2)[of trl "trr ## trn'"]) (auto simp: never_n\<gamma>_append)
    qed
  qed
qed auto

(* Factoring in an invariant: *)

definition "O_\<epsilon>a_inv na ss Bw Fw Inv
 \<equiv> \<forall>s tr.
      reach s \<and> validFrom s tr 
    \<and> O tr = O (traceOf ss (\<epsilon>a na ss))
    \<and> list_ex \<gamma> (traceOf ss (\<epsilon>a na ss))
    \<and> never_n\<gamma> Bw [] tr \<and> never_n\<gamma> Fw [] tr
    \<and> list_all (\<lambda>trn. Inv (srcOf trn) \<and> Inv (tgtOf trn)) tr
    (*\<and> list_all (\<lambda>trn. (\<gamma> trn \<or> (\<not>Bw trn \<and> \<not>Fw trn)) \<and> Inv (srcOf trn) \<and> Inv (tgtOf trn)) tr*)
  \<longrightarrow> O (traceOf s (\<epsilon>a na s)) = O tr
    \<and> V (traceOf s (\<epsilon>a na s)) = V tr
    \<and> tgtOf (last (traceOf s (\<epsilon>a na s))) = tgtOf (last tr)"  

lemma O_\<epsilon>a_invE:
assumes "O_\<epsilon>a_inv na ss Bw Fw Inv"
and "reach s" "O tr = O (traceOf ss (\<epsilon>a na ss))" "validFrom s tr" 
"list_ex \<gamma> (traceOf ss (\<epsilon>a na ss))" (*"list_all (\<lambda>trn. (\<gamma> trn \<or> (\<not>Bw trn \<and> \<not>Fw trn)) \<and> Inv (srcOf trn) \<and> Inv (tgtOf trn)) tr"*)
"never_n\<gamma> Bw [] tr" "never_n\<gamma> Fw [] tr" "list_all (\<lambda>trn. Inv (srcOf trn) \<and> Inv (tgtOf trn)) tr"
shows "O (traceOf s (\<epsilon>a na s)) = O tr \<and> V (traceOf s (\<epsilon>a na s)) = V tr \<and> tgtOf (last (traceOf s (\<epsilon>a na s))) = tgtOf (last tr)"
using assms unfolding O_\<epsilon>a_inv_def by blast

primrec changesState :: "('state, 'act, 'out) trans \<Rightarrow> bool" where
  "changesState (Trans s a ou s') = (s' \<noteq> s)"

definition "\<epsilon>a_commute_inv na ss Bw Fw Inv 
 \<equiv> \<forall>s0 s a a' tr1 tr2.
   let trn = transOf s a; trn' = transOf (snd (step s a)) a';
       Bwn\<gamma> = \<lambda>obsl trn. Bw obsl trn \<and> \<not>\<gamma> trn; Fwn\<gamma> = \<lambda>obsl trn. Fw obsl trn \<and> \<not>\<gamma> trn in
      reach s0 \<and> validFrom s0 (tr1 @ trn # trn' # tr2) \<and> 
      list_all (\<lambda> trn. Inv (srcOf trn) \<and> Inv (tgtOf trn)) (tr1 @ trn # trn' # tr2)
      \<and> (\<gamma> trn \<longrightarrow> a \<in>\<in> \<epsilon>a na ss) \<and> (\<gamma> trn' \<longrightarrow> a' \<in>\<in> \<epsilon>a na ss)
      \<and> ((Bwn\<gamma> (O (tr1 ## trn)) trn' \<and> \<not>Bwn\<gamma> (O tr1) trn \<and> (changesState trn \<or> \<gamma> trn \<or> \<phi> trn) \<and> (changesState trn' \<or> \<phi> trn') \<and> never_n\<gamma> Bw [] tr1)
         \<or> 
         (Fwn\<gamma> (O tr1) trn \<and> \<not>Fwn\<gamma> (O (tr1 ## trn)) trn' \<and> (changesState trn \<or> \<phi> trn) \<and> (changesState trn' \<or> \<gamma> trn' \<or> \<phi> trn') \<and> never_n\<gamma> Fw (O (tr1 @ [trn, trn'])) tr2)
        )
      \<and> O (tr1 @ trn # trn' # tr2) = O (traceOf ss (\<epsilon>a na ss))
      \<and> length (O (traceOf ss (\<epsilon>a na ss))) > 1
  \<longrightarrow> \<epsilon>a_commute_goal na ss (O tr1) Bw Fw Inv s a a'"

lemma \<epsilon>a_commute_invE:
  fixes na ss Bw Fw Inv s0 s a a' tr1 tr2
  defines Bwn\<gamma>: "Bwn\<gamma> \<equiv> (\<lambda>obsl trn. Bw obsl trn \<and> \<not>\<gamma> trn)"
      and Fwn\<gamma>: "Fwn\<gamma> \<equiv> (\<lambda>obsl trn. Fw obsl trn \<and> \<not>\<gamma> trn)"
  assumes c: "\<epsilon>a_commute_inv na ss Bw Fw Inv"
      and prems: "reach s0" "validFrom s0 (tr1 @ trn # trn' # tr2)"
          "list_all (\<lambda> trn. Inv (srcOf trn) \<and> Inv (tgtOf trn)) (tr1 @ trn # trn' # tr2)"
          "\<gamma> trn \<longrightarrow> actOf trn \<in>\<in> \<epsilon>a na ss" "\<gamma> trn' \<longrightarrow> actOf trn' \<in>\<in> \<epsilon>a na ss"
          "(Bwn\<gamma> (O (tr1 ## trn)) trn' \<and> \<not>Bwn\<gamma> (O tr1) trn \<and> (changesState trn \<or> \<gamma> trn \<or> \<phi> trn) \<and> (changesState trn' \<or> \<phi> trn') \<and> never_n\<gamma> Bw [] tr1)
         \<or> (Fwn\<gamma> (O tr1) trn \<and> \<not>Fwn\<gamma> (O (tr1 ## trn)) trn' \<and> (changesState trn \<or> \<phi> trn) \<and> (changesState trn' \<or> \<gamma> trn' \<or> \<phi> trn') \<and> never_n\<gamma> Fw (O (tr1 @ [trn, trn'])) tr2)"
          "O (tr1 @ trn # trn' # tr2) = O (traceOf ss (\<epsilon>a na ss))"
          "length (O (traceOf ss (\<epsilon>a na ss))) > 1"
  shows "\<epsilon>a_commute_goal na ss (O tr1) Bw Fw Inv (srcOf trn) (actOf trn) (actOf trn')"
proof -
  from prems(2) obtain s a where trn: "trn = transOf s a" by (cases trn, auto) (metis fst_conv snd_conv)+
  with prems(2) obtain a' where trn': "trn' = transOf (snd (step s a)) a'" by (cases trn', auto) (metis fst_conv snd_conv)+
  have s_a_a': "srcOf trn = s" "actOf trn = a" "actOf trn' = a'" unfolding trn trn' by auto
  from c prems show ?thesis unfolding s_a_a' unfolding \<epsilon>a_commute_inv_def Let_def s_a_a' trn trn' Bwn\<gamma> Fwn\<gamma> by blast
qed

lemma \<epsilon>a_commute_invI_length[simp, intro!]:
assumes "length (O (traceOf ss (\<epsilon>a na ss))) \<le> 1"
shows "\<epsilon>a_commute_inv na ss Bw Fw Inv"
using assms unfolding \<epsilon>a_commute_inv_def by auto

lemma \<epsilon>a_commute_invI[intro, case_names Bw Fw]:
assumes 
Bw: "\<And> s0 s a a' tr1 tr2.
      \<lbrakk>length (O (traceOf ss (\<epsilon>a na ss))) > 1;
       reach s0; validFrom s0 tr1; (tr1 \<noteq> [] \<longrightarrow> tgtOf (last tr1) = s);
       reach s; validFrom (snd (step (snd (step s a)) a')) tr2;
       list_all (\<lambda>trn. Inv (srcOf trn) \<and> Inv (tgtOf trn)) tr1;
       list_all (\<lambda>trn. Inv (srcOf trn) \<and> Inv (tgtOf trn)) tr2;
       Inv s; Inv ((snd (step s a))); Inv (snd (step (snd (step s a)) a'));
       \<gamma> (transOf s a) \<Longrightarrow> 
         a \<in>\<in> \<epsilon>a na ss \<and> (a, fst (step s a)) \<in>\<in> O (traceOf ss (\<epsilon>a na ss));
       Bw (O tr1) (transOf s a) \<Longrightarrow> \<gamma> (transOf s a); 
       Bw (O (tr1 ## transOf s a)) (transOf (snd (step s a)) a'); \<not> \<gamma> (transOf (snd (step s a)) a');
       changesState (transOf s a) \<or> \<gamma> (transOf s a) \<or> \<phi> (transOf s a);
       changesState (transOf (snd (step s a)) a') \<or> \<phi> (transOf (snd (step s a)) a');
       never_n\<gamma> Bw [] tr1;
       O (tr1 @ traceOf s [a, a'] @ tr2) = O (traceOf ss (\<epsilon>a na ss))
      \<rbrakk>
     \<Longrightarrow> \<epsilon>a_commute_goal na ss (O tr1) Bw Fw Inv s a a'"
and 
Fw: "\<And> s0 s a a' tr1 tr2.
      \<lbrakk>length (O (traceOf ss (\<epsilon>a na ss))) > 1;
       reach s0; validFrom s0 tr1; (tr1 \<noteq> [] \<longrightarrow> tgtOf (last tr1) = s);
       reach s; validFrom (snd (step (snd (step s a)) a')) tr2;
       list_all (\<lambda>trn. Inv (srcOf trn) \<and> Inv (tgtOf trn)) tr1;
       list_all (\<lambda>trn. Inv (srcOf trn) \<and> Inv (tgtOf trn)) tr2;
       Inv s; Inv ((snd (step s a))); Inv (snd (step (snd (step s a)) a'));
       \<gamma> (transOf (snd (step s a)) a') \<Longrightarrow> 
          a' \<in>\<in> \<epsilon>a na ss \<and> (a', fst (step (snd (step s a)) a')) \<in>\<in> O (traceOf ss (\<epsilon>a na ss));
       Fw (O tr1) (transOf s a); \<not> \<gamma> (transOf s a); 
       Fw (O (tr1 ## transOf s a)) (transOf (snd (step s a)) a') \<Longrightarrow> \<gamma> (transOf (snd (step s a)) a');
       changesState (transOf s a) \<or> \<phi> (transOf s a);
       changesState (transOf (snd (step s a)) a') \<or> \<gamma> (transOf (snd (step s a)) a') \<or> \<phi> (transOf (snd (step s a)) a');
       never_n\<gamma> Fw (O (tr1 @ traceOf s [a, a'])) tr2;
       O (tr1 @ traceOf s [a, a'] @ tr2) = O (traceOf ss (\<epsilon>a na ss))
      \<rbrakk>
     \<Longrightarrow> \<epsilon>a_commute_goal na ss (O tr1) Bw Fw Inv s a a'"
shows "\<epsilon>a_commute_inv na ss Bw Fw Inv"
unfolding \<epsilon>a_commute_inv_def Let_def proof(intro allI impI, elim conjE)
  fix s0 s a a' tr1 tr2
  assume s0: "reach s0" 
  and v: "validFrom s0 (tr1 @ transOf s a # transOf (snd (step s a)) a' # tr2)"
  (is "validFrom s0 (tr1 @ transOf s a # ?tr)")
  and \<gamma>: "\<gamma> (transOf s a) \<longrightarrow> a \<in>\<in> \<epsilon>a na ss"
         "\<gamma> (transOf (snd (step s a)) a') \<longrightarrow> a' \<in>\<in> \<epsilon>a na ss"
  and l: "length (O (traceOf ss (\<epsilon>a na ss))) > 1"
  and O: "O (tr1 @ transOf s a # transOf (snd (step s a)) a' # tr2) =
          O (traceOf ss (\<epsilon>a na ss))"
  and Inv: "list_all (\<lambda> trn. Inv (srcOf trn) \<and> Inv (tgtOf trn)) (tr1 @ transOf s a # transOf (snd (step s a)) a' # tr2)"
  and D: "(Bw (O (tr1 ## transOf s a)) (transOf (snd (step s a)) a') \<and>
        \<not> \<gamma> (transOf (snd (step s a)) a')) \<and>
       \<not> (Bw (O tr1) (transOf s a) \<and> \<not> \<gamma> (transOf s a)) \<and>
       (changesState (transOf s a) \<or> \<gamma> (transOf s a) \<or> \<phi> (transOf s a)) \<and>
       (changesState (transOf (snd (step s a)) a') \<or> \<phi> (transOf (snd (step s a)) a')) \<and>
       never_n\<gamma> Bw [] tr1
       \<or>
       (Fw (O tr1) (transOf s a) \<and> \<not> \<gamma> (transOf s a)) \<and>
       \<not> (Fw (O (tr1 ## transOf s a)) (transOf (snd (step s a)) a') \<and>
           \<not> \<gamma> (transOf (snd (step s a)) a')) \<and>
       (changesState (transOf s a) \<or> \<phi> (transOf s a)) \<and>
       (changesState (transOf (snd (step s a)) a') \<or> \<gamma> (transOf (snd (step s a)) a') \<or> \<phi> (transOf (snd (step s a)) a')) \<and>
       never_n\<gamma> Fw (O (tr1 @ [transOf s a, transOf (snd (step s a)) a'])) tr2" (is "?Bw \<or> ?Fw")
  (*have Inv: "Inv s" "Inv (snd (step s a))" "Inv (snd (step (snd (step s a)) a'))" using Inv unfolding list_all_iff by auto*)
  {assume "\<gamma> (transOf s a)"
   hence "(a, fst (step s a)) \<in>\<in> O (traceOf ss (\<epsilon>a na ss))"
   unfolding O[symmetric] O_append by auto
  } note 1 = this
  {assume "\<gamma> (transOf (snd (step s a)) a')"
   hence "(a', fst (step (snd (step s a)) a')) \<in>\<in> O (traceOf ss (\<epsilon>a na ss))"
   unfolding O[symmetric] O_append by (cases "\<gamma> (transOf s a)", auto)
  } note 2 = this
  have "s = s0 \<or> (tr1 \<noteq> [] \<and> validFrom s0 tr1 \<and> tgtOf (last tr1) = s)"
  using v unfolding validFrom_append by auto
  hence s: "reach s" using s0 by (metis validFrom_def valid_reach_src_tgt)
  show "\<epsilon>a_commute_goal na ss (O tr1) Bw Fw Inv s a a'"
  using D proof
    assume ?Bw with s0 v s \<gamma> l 1 Inv O show ?thesis by (intro Bw, unfold traceOf_Cons) auto
  next
    assume ?Fw with s0 v s \<gamma> l 2 Inv O show ?thesis by (intro Fw, unfold traceOf_Cons) auto
  qed
qed

definition "\<epsilon>a_inv na ss Inv \<equiv>
\<forall> s tr. 
  list_ex \<gamma> (traceOf ss (\<epsilon>a na ss)) \<and>
  reach s \<and> validFrom s tr \<and> O tr = O (traceOf ss (\<epsilon>a na ss)) \<and> \<gamma> (hd tr) \<and> \<gamma> (last tr)
  \<longrightarrow> list_all (\<lambda> trn. Inv (srcOf trn) \<and>  Inv (tgtOf trn)) tr"

lemma \<epsilon>a_invE:
  assumes "\<epsilon>a_inv na ss Inv"
      and "list_ex \<gamma> (traceOf ss (\<epsilon>a na ss))"
      and "validFrom s tr" "reach s" "O tr = O (traceOf ss (\<epsilon>a na ss))" "\<gamma> (hd tr)" "\<gamma> (last tr)"
  shows "list_all (\<lambda> trn. Inv (srcOf trn) \<and>  Inv (tgtOf trn)) tr"
using assms unfolding \<epsilon>a_inv_def by blast

definition "O_split Inv s tr trl trm trr
 \<equiv> validFrom s (trl @ trm @ trr)
    \<and> V (trl @ trm @ trr) = V tr
    \<and> never \<gamma> trl \<and> O trm = O tr \<and> never \<gamma> trr
    \<and> list_all (\<lambda>trn. Inv (srcOf trn) \<and> Inv (tgtOf trn)) trm
    \<and> tgtOf (last (trl @ trm @ trr)) = tgtOf (last tr)"

lemma O_split_left_neutral:
  assumes "O_split Inv s tr trl (trn # trm) trr"
      and "\<not>\<gamma> trn"
  shows "O_split Inv s tr (trl ## trn) trm trr"
using assms unfolding O_split_def by auto

lemma O_split_right_neutral:
  assumes "O_split Inv s tr trl (trm ## trn) trr"
      and "\<not>\<gamma> trn"
  shows "O_split Inv s tr trl trm (trn # trr)"
using assms unfolding O_split_def by auto

lemma O_split_replace_trm:
  assumes O: "O_split Inv s tr trl trm trr"
      and \<gamma>: "O tr \<noteq> []"
      and s': "validFrom s' trm'" "(trl = [] \<and> s' = s) \<or> (trl \<noteq> [] \<and> s' = tgtOf (last trl))"
      and trm': "O trm' = O trm \<and> V trm' = V trm \<and> tgtOf (last trm') = tgtOf (last trm)"
                "list_all (\<lambda>trn. Inv' (srcOf trn) \<and> Inv' (tgtOf trn)) trm'"
  shows "O_split Inv' s tr trl trm' trr"
proof -
  from \<gamma> have "O trm \<noteq> []" using O unfolding O_split_def by auto
  moreover then have "O trm' \<noteq> []" using trm'(1) by simp
  ultimately have trm_not_Nil: "trm \<noteq> []" "trm' \<noteq> []" by (auto simp add: O_Nil_list_ex)
  moreover from O have "validFrom s (trl @ trm @ trr)" unfolding O_split_def by auto
  ultimately have "validFrom s (trl @ trm' @ trr)" using s' trm'(1) by auto
  then show ?thesis using O trm' trm_not_Nil unfolding O_split_def by (cases "trr = []") auto
qed

lemma O_split_drop:
  assumes O_split: "O_split Inv s0 tr trl (trm1 @ trn # trn' # trm2) trr"
      and trn: "\<not> changesState trn" "\<not> \<gamma> trn" "\<not> \<phi> trn"
  shows "O_split Inv s0 tr trl (trm1 @ trn' # trm2) trr"
proof -
  have O: "O (trm1 @ trn # trn' # trm2) = O (trm1 @ trn' # trm2)"
   and V: "V (trl @ (trm1 @ trn # trn' # trm2) @ trr) = V (trl @ (trm1 @ trn' # trm2) @ trr)"
    using trn(2,3) by auto
  have "validFrom s0 (trl @ (trm1 @ trn' # trm2) @ trr)"
    using O_split trn(1) unfolding O_split_def by (cases trn) auto
  then show ?thesis using O_split unfolding O_split_def O V by auto
qed

lemma O_split_drop':
  assumes O_split: "O_split Inv s0 tr trl (trm1 @ trn # trn' # trm2) trr"
      and trn: "\<not> changesState trn'" "\<not> \<gamma> trn'" "\<not> \<phi> trn'"
  shows "O_split Inv s0 tr trl (trm1 @ trn # trm2) trr"
proof -
  have O: "O (trm1 @ trn # trn' # trm2) = O (trm1 @ trn # trm2)"
   and V: "V (trl @ (trm1 @ trn # trn' # trm2) @ trr) = V (trl @ (trm1 @ trn # trm2) @ trr)"
    using trn(2,3) by (auto simp add: O_Cons_unfold V_Cons_unfold)
  have tgt: "tgtOf (last (trl @ (trm1 @ trn # trm2) @ trr)) = tgtOf (last (trl @ (trm1 @ trn # trn' # trm2) @ trr))" (*"tgtOf trn' = tgtOf trn"*)
    using O_split trn(1) unfolding O_split_def by (cases trn') auto
  have "validFrom s0 (trl @ (trm1 @ trn # trm2) @ trr)"
    using O_split trn(1) unfolding O_split_def by (cases trn') auto
  moreover have "list_all (\<lambda>trn. Inv (srcOf trn) \<and> Inv (tgtOf trn)) (trm1 @ trn # trm2)"
    using O_split unfolding O_split_def by auto
  ultimately show ?thesis using O_split unfolding O_split_def O V tgt by blast
qed

lemma O_split_commute:
  assumes O_split: "O_split Inv s0 tr trl (trm1 @ trn # trn' # trm2) trr"
      and \<epsilon>a_commute_goal: "\<epsilon>a_commute_goal na ss (O trm1) Bw Fw Inv (srcOf trn) (actOf trn) (actOf trn')"
  defines trn1': "trn1' \<equiv> transOf (srcOf trn) (actOf trn')"
      and trn1: "trn1 \<equiv> transOf (snd (step (srcOf trn) (actOf trn'))) (actOf trn)"
  shows "(O_split Inv s0 tr trl (trm1 @ trn1' # trn1 # trm2) trr
          \<and> Bw (O (trm1 ## trn1')) trn1 = Bw (O trm1) trn \<and> Fw (O (trm1 ## trn1')) trn1 = Fw (O trm1) trn
          \<and> Bw (O trm1) trn1' = Bw (O (trm1 ## trn)) trn' \<and> Fw (O trm1) trn1' = Fw (O (trm1 ## trn)) trn'
          \<and> equivFiltermap \<phi> f trn1 trn \<and> equivFiltermap \<gamma> g trn1 trn
          \<and> equivFiltermap \<phi> f trn1' trn' \<and> equivFiltermap \<gamma> g trn1' trn')
       \<or> (O_split Inv s0 tr trl (trm1 @ trn1' # trm2) trr
          \<and> \<not> \<gamma> trn \<and> \<not> \<phi> trn \<and> (Bw (O trm1) trn1' \<longrightarrow> Bw (O trm1) trn \<or> Bw (O (trm1 ## trn)) trn') (*\<and> Fw trn1' = Fw trn'*)
          \<and> equivFiltermap \<phi> f trn1' trn' \<and> equivFiltermap \<gamma> g trn1' trn')"
proof -
  from O_split have "validTrans trn" unfolding O_split_def by auto
  then obtain s a where trn: "trn = transOf s a" by (cases trn, auto) (metis fst_conv snd_conv)
  with O_split have "validTrans trn'" "srcOf trn' = snd (step s a)" unfolding O_split_def by auto
  then obtain a' where trn': "trn' = transOf (snd (step s a)) a'" by (cases trn', auto) (metis fst_conv snd_conv)
  have trn_a': "trn1' = transOf s a'" and trn_a: "trn1 = transOf (snd (step s a')) a" unfolding trn1' trn1 trn trn' by auto
  from \<epsilon>a_commute_goal
  have "(commute s a a' 
        \<and> (Bw (O (trm1 ## trn1')) trn1 \<longleftrightarrow> Bw (O trm1) trn) \<and> (Bw (O trm1) trn1' \<longleftrightarrow> Bw (O (trm1 ## trn)) trn')
        \<and> (Fw (O (trm1 ## trn1')) trn1 \<longleftrightarrow> Fw (O trm1) trn) \<and> (Fw (O trm1) trn1' \<longleftrightarrow> Fw (O (trm1 ## trn)) trn')
        \<and> (Inv s \<longrightarrow> Inv (snd (step s a)) \<longrightarrow> Inv (snd (step s a')))  
        \<and> \<not> (\<phi> trn \<and> \<phi> trn') \<and> \<not> (\<gamma> trn \<and> \<gamma> trn')
        \<and> equivFiltermap \<phi> f trn1 trn \<and> equivFiltermap \<gamma> g trn1 trn
        \<and> equivFiltermap \<phi> f trn1' trn' \<and> equivFiltermap \<gamma> g trn1' trn')
       \<or> 
       (absorb s a a' \<and> \<not> \<gamma> trn \<and> \<not> \<phi> trn 
        \<and> (Bw (O trm1) trn1' \<longrightarrow> Bw (O trm1) trn \<or> Bw (O (trm1 ## trn)) trn')
        (*\<and> (Fw trn1' \<longleftrightarrow> Fw trn')*)
        \<and> (Inv s \<longrightarrow> Inv (snd (step s a)) \<longrightarrow> Inv (snd (step s a')))  
        \<and> equivFiltermap \<phi> f trn1' trn' \<and> equivFiltermap \<gamma> g trn1' trn')"
       (is "?commute \<or> ?absorb")
    using trn trn' unfolding \<epsilon>a_commute_goal_def trn1 trn1' by auto
  then show ?thesis
  proof
    assume c: "?commute"
    moreover have "O_split Inv s0 tr trl (trm1 @ trn1' # trn1 # trm2) trr" unfolding O_split_def
    proof (intro conjI)
      from c have c_trn: "commute s a a'" by simp
      from O_split have "validFrom s0 ((trl @ trm1) @ trn # trn' # (trm2 @ trr))"
        unfolding O_split_def by auto
      then have "validFrom s0 ((trl @ trm1) @ trn1' # trn1 # (trm2 @ trr))"
        using c_trn unfolding trn trn' trn_a trn_a' by (elim validFrom_commute)
      then show "validFrom s0 (trl @ (trm1 @ trn1' # trn1 # trm2) @ trr)" by simp
      have "V (trl @ (trm1 @ trn1' # trn1 # trm2) @ trr) = V (trl @ trm1) @ V [trn1', trn1] @ V (trm2 @ trr)" by (auto simp add: V_Cons_unfold)
      also have "V [trn1', trn1] = V [trn, trn']" using c by (intro equivFiltermap_V[symmetric]) auto
      also have "V (trl @ trm1) @ V [trn, trn'] @ V (trm2 @ trr) = V (trl @ (trm1 @ trn # trn' # trm2) @ trr)" by (auto simp add: V_Cons_unfold)
      finally show "V (trl @ (trm1 @ trn1' # trn1 # trm2) @ trr) = Orig.V tr" using O_split unfolding O_split_def by auto
      have "O (trm1 @ trn1' # trn1 # trm2) = O trm1 @ O [trn1', trn1] @ O trm2" by (auto simp add: O_Cons_unfold)
      also have "O [trn1', trn1] = O [trn, trn']" using c by (intro equivFiltermap_O[symmetric]) auto
      also have "O trm1 @ O [trn, trn'] @ O trm2 = O (trm1 @ trn # trn' # trm2)" by (auto simp add: O_Cons_unfold)
      finally show "O (trm1 @ trn1' # trn1 # trm2) = O tr" using O_split unfolding O_split_def by auto
      have "tgtOf (last (trl @ (trm1 @ trn # trn' # trm2) @ trr)) = tgtOf (last tr)" using O_split unfolding O_split_def by auto
      moreover have "tgtOf trn1 = tgtOf trn'" using trn trn' c_trn unfolding commute_def trn_a trn_a' by auto
      ultimately show "tgtOf (last (trl @ (trm1 @ trn1' # trn1 # trm2) @ trr)) = tgtOf (last tr)" by auto
      have "Inv s" "Inv (snd (step s a))" "Inv (snd (step (snd (step s a)) a'))"
        using O_split unfolding O_split_def trn trn' by auto
      moreover then have "Inv (snd (step s a'))"
                     and "snd (step (snd (step s a')) a) = snd (step (snd (step s a)) a')"
        using c unfolding commute_def by auto
      ultimately have "Inv (srcOf trn1)" "Inv (tgtOf trn1)" "Inv (srcOf trn1')" "Inv (tgtOf trn1')"
        unfolding trn_a trn_a' trn trn' by auto
      then show "list_all (\<lambda>trn. Inv (srcOf trn) \<and> Inv (tgtOf trn)) (trm1 @ trn1' # trn1 # trm2)"
            and "never \<gamma> trl" and "never \<gamma> trr"
        using O_split unfolding O_split_def by auto
    qed
    (*moreover have "\<gamma> trn1 = \<gamma> trn"  "\<gamma> trn1' = \<gamma> trn'"  "\<phi> trn1 = \<phi> trn"  "\<phi> trn1' = \<phi> trn'"
      using c unfolding equivFiltermap_def by auto*)
    ultimately show ?thesis by blast
  next
    assume a: "?absorb"
    moreover have "O_split Inv s0 tr trl (trm1 @ trn1' # trm2) trr" unfolding O_split_def
    proof (intro conjI)
      from a have a_trn: "absorb s a a'" by simp
      from O_split have "validFrom s0 ((trl @ trm1) @ trn # trn' # (trm2 @ trr))"
        unfolding O_split_def by auto
      then have "validFrom s0 ((trl @ trm1) @ trn1' # (trm2 @ trr))"
        using a_trn unfolding trn trn' trn_a trn_a' by (elim validFrom_absorb)
      then show "validFrom s0 (trl @ (trm1 @ trn1' # trm2) @ trr)" by auto
      from a have "equivFiltermap \<phi> f trn1' trn'" "\<not> \<phi> trn" by auto
      then have "V (trl @ (trm1 @ trn1' # trm2) @ trr) = V (trl @ (trm1 @ trn # trn' # trm2) @ trr)"
        unfolding equivFiltermap_def by auto
      then show "V (trl @ (trm1 @ trn1' # trm2) @ trr) = V tr" using O_split unfolding O_split_def by auto
      from a have "equivFiltermap \<gamma> g trn1' trn'" "\<not> \<gamma> trn" by auto
      then have "O (trm1 @ trn1' # trm2) = O (trm1 @ trn # trn' # trm2)"
        unfolding equivFiltermap_def by auto
      then show "O (trm1 @ trn1' # trm2) = O tr" using O_split unfolding O_split_def by auto
      have "tgtOf (last (trl @ (trm1 @ trn # trn' # trm2) @ trr)) = tgtOf (last tr)" using O_split unfolding O_split_def by auto
      moreover have "tgtOf trn1' = tgtOf trn'" using trn trn' a_trn unfolding absorb_def trn_a trn_a' by auto
      ultimately show "tgtOf (last (trl @ (trm1 @ trn1' # trm2) @ trr)) = tgtOf (last tr)" by auto
      have "Inv s" "Inv (snd (step s a))" using O_split unfolding O_split_def trn trn' by auto
      moreover then have "Inv (snd (step s a'))" using a by auto
      ultimately show "list_all (\<lambda>trn. Inv (srcOf trn) \<and> Inv (tgtOf trn)) (trm1 @ trn1' # trm2)"
                  and "never \<gamma> trl" and "never \<gamma> trr"
        using O_split unfolding O_split_def trn_a' by auto
    qed
    (*moreover have "\<gamma> trn1' = \<gamma> trn'"  "\<phi> trn1' = \<phi> trn'"
      using a unfolding equivFiltermap_def by auto*)
    ultimately show ?thesis by blast
  qed
qed

lemma length_filtermap_hd_last_gt_1:
  assumes "length tr > 1"
      and "pred (hd tr)" "pred (last tr)"
  shows "length (filtermap pred func tr) > 1"
proof -
  from assms(1) obtain x tr' where "tr = x # tr'" by (cases tr) auto
  with assms(1) obtain y tr'' where "tr = x # tr'' ## y" by (cases tr' rule: rev_cases) auto
  with assms(2,3) have "filtermap pred func tr = func x # filtermap pred func tr'' ## func y" by auto
  then show ?thesis by auto
qed

thm commute_def
thm \<epsilon>a_commute_goal_def
thm O_\<epsilon>a_inv_def
thm \<epsilon>a_inv_def
thm \<epsilon>a_commute_inv_def

(* can also use: *)
definition "\<epsilon>a_noStateChange na ss \<equiv> \<forall> a s. a \<in>\<in> \<epsilon>a na ss \<longrightarrow> snd (step s a) = s"
definition "no_\<phi>_\<epsilon>a na ss \<equiv> \<forall> trn. actOf trn \<in>\<in> \<epsilon>a na ss \<longrightarrow> \<not> \<phi> trn"

lemma local_adjust
  unfolding local_adjust_def local_adjust_param_def
  apply clarsimp 
  unfolding local_right_def[abs_def] 
oops

lemma \<epsilon>a_commute_inv_local_adjust:
  assumes \<epsilon>a_inv: "\<And> na ss. reach ss \<Longrightarrow> \<epsilon>a_inv na ss (Inv na ss)"
  and \<epsilon>a_commute_inv: "\<And> na ss. reach ss \<Longrightarrow> \<epsilon>a_commute_inv na ss (Bw na ss) (Fw na ss) (Inv na ss)" 
  and O_\<epsilon>a_inv: "\<And> na ss. reach ss \<Longrightarrow> O_\<epsilon>a_inv na ss (Bw na ss) (Fw na ss) (Inv na ss)"
  shows local_adjust
unfolding local_adjust_def local_adjust_param_def Let_def
proof (intro allI impI, elim conjE)
  fix na ss s tr
  assume r: "reach ss" "reach s"
     and \<gamma>_ss: "list_ex \<gamma> (traceOf ss (\<epsilon>a na ss))"
     and s_tr: "validFrom s tr"
     and O: "O tr = O (traceOf ss (\<epsilon>a na ss))" (is "O tr = ?Oss")
  then have Oss_not_Nil: "?Oss \<noteq> []" by (simp add: O_Nil_list_ex)
  let ?Bwn\<gamma> = "\<lambda>obsl trn. Bw na ss obsl trn \<and> \<not>\<gamma> trn"
  let ?Fwn\<gamma> = "\<lambda>obsl trn. Fw na ss obsl trn \<and> \<not>\<gamma> trn"
  let ?BwFwn\<gamma> = "\<lambda>obsl trn. ?Bwn\<gamma> obsl trn \<or> ?Fwn\<gamma> obsl trn"
  let ?Inv = "\<lambda>trn. Inv na ss (srcOf trn) \<and> Inv na ss (tgtOf trn)"
  let ?s' = "\<lambda>s tr1. if tr1 = [] then s else tgtOf (last tr1)"
  let ?swap' = "\<lambda>trn trn'. transOf (srcOf trn) (actOf trn')"
  let ?swap = "\<lambda>trn trn'. transOf (snd (step (srcOf trn) (actOf trn'))) (actOf trn)"
  let ?O_split_commute = "\<lambda>trl trm1 trn trn' trm2 trr.
    let trn1' = ?swap' trn trn'; trn1 = ?swap trn trn' in
    (O_split (Inv na ss) s tr trl (trm1 @ trn1' # trn1 # trm2) trr
     \<and> Bw na ss (O (trm1 ## trn1')) trn1 = Bw na ss (O trm1) trn
     \<and> Fw na ss (O (trm1 ## trn1')) trn1 = Fw na ss (O trm1) trn
     \<and> Bw na ss (O trm1) trn1' = Bw na ss (O (trm1 ## trn)) trn'
     \<and> Fw na ss (O trm1) trn1' = Fw na ss (O (trm1 ## trn)) trn'
     \<and> equivFiltermap \<phi> f trn1 trn \<and> equivFiltermap \<gamma> g trn1 trn
     \<and> equivFiltermap \<phi> f trn1' trn' \<and> equivFiltermap \<gamma> g trn1' trn')"
  let ?O_split_absorb = "\<lambda>trl trm1 trn trn' trm2 trr.
    let trn1' = ?swap' trn trn'; trn1 = ?swap trn trn' in
    (O_split (Inv na ss) s tr trl (trm1 @ trn1' # trm2) trr
     \<and> \<not>\<gamma> trn \<and> \<not>\<phi> trn
     \<and> (Bw na ss (O trm1) trn1' \<longrightarrow> Bw na ss (O trm1) trn \<or> Bw na ss (O (trm1 ## trn)) trn') (*\<and> Fw na ss trn1' = Fw na ss trn'*)
     \<and> equivFiltermap \<phi> f trn1' trn' \<and> equivFiltermap \<gamma> g trn1' trn')"
  let ?O_split_absorb' = "\<lambda>trl trm1 trn trn' trm2 trr.
    (O_split (Inv na ss) s tr trl (trm1 @ trn # trm2) trr \<and> \<not>\<gamma> trn' \<and> \<not>\<phi> trn')"

  { fix tr trn
    assume O: "O tr = O (traceOf ss (\<epsilon>a na ss))"
       and trn: "trn \<in>\<in> tr" and \<gamma>: "\<gamma> trn"
    have a_O: "actOf trn \<in>\<in> map fst ?Oss"
      unfolding O[symmetric] unfolding O_def filtermap_map_filter
      using trn \<gamma> by auto
    have "actOf trn \<in>\<in> \<epsilon>a na ss"
      using fst_O_actOf[OF a_O] by auto
  } note \<gamma>_\<epsilon>a = this

  { fix trl trm1 trn trn' trm2 trr
    assume split: "O_split (Inv na ss) s tr trl (trm1 @ trn # trn' # trm2) trr"
       and \<gamma>_hd: "\<gamma> (hd (trm1 @ trn # trn' # trm2))"
       and \<gamma>_last: "\<gamma> (last (trm1 @ trn # trn' # trm2))"
       and Bw_Fw: "(Bw na ss (O (trm1 ## trn)) trn' \<and> \<not> \<gamma> trn' \<and> (Bw na ss (O trm1) trn \<longrightarrow> \<gamma> trn) \<and> never_n\<gamma> (Bw na ss) [] trm1)
                 \<or> (Fw na ss (O trm1) trn \<and> \<not> \<gamma> trn \<and> (Fw na ss (O (trm1 ## trn)) trn' \<longrightarrow> \<gamma> trn') \<and> never_n\<gamma> (Fw na ss) (O (trm1 @ [trn, trn']))  trm2)"
         (is "?Bw \<or> ?Fw")
    have O_trm: "O (trm1 @ trn # trn' # trm2) = ?Oss" using split unfolding O_split_def O by blast
    with \<gamma>_hd \<gamma>_last have Oss_gt_1: "length ?Oss > 1"
      unfolding O_trm[symmetric] unfolding O_def
      by (intro length_filtermap_hd_last_gt_1) auto
    from Bw_Fw have "?O_split_commute trl trm1 trn trn' trm2 trr
                   \<or> ?O_split_absorb trl trm1 trn trn' trm2 trr
                   \<or> ?O_split_absorb' trl trm1 trn trn' trm2 trr"
    proof (elim disjE)
      assume Bw: "?Bw"
      consider (cS) "(changesState trn \<or> \<gamma> trn \<or> \<phi> trn) \<and> (changesState trn' \<or> \<phi> trn')"
             | (drop) "\<not>(changesState trn \<or> \<gamma> trn \<or> \<phi> trn)" | (drop') "\<not>(changesState trn' \<or> \<phi> trn')"
        by auto
      then show ?thesis proof cases
        case cS
        then have "\<epsilon>a_commute_goal na ss (O trm1) (Bw na ss) (Fw na ss) (Inv na ss) (srcOf trn) (actOf trn) (actOf trn')"
          using \<epsilon>a_commute_inv[OF r(1), of na] Bw split O_trm Oss_gt_1
        proof (elim \<epsilon>a_commute_invE)
          show "reach (?s' s trl)"
           and "validFrom (?s' s trl) (trm1 @ trn # trn' # trm2)"
            using r(2) split unfolding O_split_def by (auto intro: validFrom_reach)
          show "\<gamma> trn \<longrightarrow> actOf trn \<in>\<in> \<epsilon>a na ss" and "\<gamma> trn' \<longrightarrow> actOf trn' \<in>\<in> \<epsilon>a na ss"
            using \<gamma>_\<epsilon>a[OF O_trm] by auto
        qed (auto simp add: O_split_def)
        then have "?O_split_commute trl trm1 trn trn' trm2 trr
                 \<or> ?O_split_absorb trl trm1 trn trn' trm2 trr"
          using split unfolding Let_def by (intro O_split_commute)
        then show ?thesis by blast
      next
        case drop
        then have "transOf (srcOf trn) (actOf trn') = trn'" using split
          by (cases trn; cases trn') (auto simp: O_split_def)
        with drop Bw show ?thesis using O_split_drop[OF split] by (auto simp: equivFiltermap_def)
      next
        case drop'
        with Bw split show ?thesis by (blast intro: O_split_drop')
      qed
    next
      assume Fw: "?Fw"
      consider (cS) "(changesState trn \<or> \<phi> trn) \<and> (changesState trn' \<or> \<gamma> trn' \<or> \<phi> trn')"
             | (drop) "\<not>(changesState trn \<or> \<phi> trn)" | (drop') "\<not>(changesState trn' \<or> \<gamma> trn' \<or> \<phi> trn')"
        by auto
      then show ?thesis proof cases
        case cS
        then have "\<epsilon>a_commute_goal na ss (O trm1) (Bw na ss) (Fw na ss) (Inv na ss) (srcOf trn) (actOf trn) (actOf trn')"
          using \<epsilon>a_commute_inv[OF r(1), of na] Fw O_trm Oss_gt_1
        proof (elim \<epsilon>a_commute_invE)
          show "reach (?s' s trl)"
           and "validFrom (?s' s trl) (trm1 @ trn # trn' # trm2)"
            using r(2) split unfolding O_split_def by (auto intro: validFrom_reach)
          show "\<gamma> trn \<longrightarrow> actOf trn \<in>\<in> \<epsilon>a na ss" and "\<gamma> trn' \<longrightarrow> actOf trn' \<in>\<in> \<epsilon>a na ss"
            using \<gamma>_\<epsilon>a[OF O_trm] by auto
          show "list_all (\<lambda>trn. Inv na ss (srcOf trn) \<and> Inv na ss (tgtOf trn)) (trm1 @ trn # trn' # trm2)"
            using split unfolding O_split_def by auto
        qed auto
        then have "?O_split_commute trl trm1 trn trn' trm2 trr
                 \<or> ?O_split_absorb trl trm1 trn trn' trm2 trr"
          using split unfolding Let_def by (intro O_split_commute)
        then show ?thesis by blast
      next
        case drop
        have "trn' = transOf (tgtOf trn) (actOf trn')" using split unfolding O_split_def by (cases trn') auto
        then have "?swap' trn trn' = trn'" using drop by (cases trn) auto
        with Fw drop split have "?O_split_absorb trl trm1 trn trn' trm2 trr" by (auto intro: O_split_drop simp: equivFiltermap_def)
        then show ?thesis by blast
      next
        case drop'
        with Fw split show ?thesis by (blast intro: O_split_drop')
      qed
    qed
  } note O_splitE = this

  obtain tr1 trm tr2
  where split: "O_split (Inv na ss) s tr tr1 trm tr2"
    and trm: "never_n\<gamma> (Bw na ss) [] trm \<and> never_n\<gamma> (Fw na ss) [] trm"
  proof (cases "length ?Oss > 1")
    assume "\<not> (length ?Oss > 1)"
    with Oss_not_Nil O obtain obs where obs: "O tr = [obs]" by (cases ?Oss) auto
    obtain tr1 trn tr2 where "tr = tr1 @ trn # tr2" and "never \<gamma> tr1" and "never \<gamma> tr2" and "\<gamma> trn" and "O tr = O [trn]"
      using O_eq_Cons[OF obs] by (auto simp add: O_Nil_never)
    moreover then have "list_all ?Inv [trn]"
      using r(2) O s_tr \<gamma>_ss
      by (intro \<epsilon>a_invE[OF \<epsilon>a_inv[OF r(1)], of na "srcOf trn"]) (auto intro: validFrom_reach)
    ultimately show thesis using O s_tr by (intro that[of "tr1" "[trn]" "tr2"]) (auto simp add: O_split_def)
  next
    assume "length ?Oss > 1"
    then obtain o1 Om on
    where Om: "?Oss = o1 # Om ## on" by (blast elim: length_gt_1_Cons_snoc)
    then obtain tr1 trn1 tro trnn tr2
    where tr: "tr = tr1 @ (trn1 # tro ## trnn) @ tr2" (is "tr = tr1 @ ?tr @ tr2")
      and trn1: "\<gamma> trn1" "g trn1 = o1" and trnn: "\<gamma> trnn" "g trnn = on"
      and O_tro: "O tro = Om"
      and tr1: "never \<gamma> tr1" and tr2: "never \<gamma> tr2"
      unfolding O[symmetric] by (auto dest!: O_eq_Cons_RCons)
    have "reach (srcOf trn1)" and "validFrom (srcOf trn1) ?tr"
      using r(2) s_tr unfolding tr by (auto intro: validFrom_reach)
    moreover then have "list_all ?Inv ?tr"
      using \<epsilon>a_inv[OF r(1)] \<gamma>_ss Om O_tro trn1 trnn by (intro \<epsilon>a_invE) auto
    then have "O_split (Inv na ss) s tr tr1 (trn1 # tro ## trnn) tr2"
      using s_tr tr1 tr2 trn1 trnn unfolding O_split_def tr by auto
    then obtain trl trn1' tro'
    where "O_split (Inv na ss) s tr (tr1 @ trl) (trn1' # tro' ## trnn) tr2"
      and tro': "never_n\<gamma> (Bw na ss) (O [trn1']) tro'"
      and trn1': "\<gamma> trn1'"
    using trn1(1)
    proof (induction "filter_n\<gamma> (Bw na ss) (O [trn1]) tro" arbitrary: tro tr1 trn1)
      case (Nil tro trn1 tr1)
        then show ?case by (intro Nil(2)[of "[]" trn1 tro]) (auto simp: never_n\<gamma>_filter_n\<gamma>_Nil)
    next
      case (Cons trnb trbs tro trn1 tr1)
        obtain tro1 tro2
        where tro: "tro = tro1 @ trnb # tro2" and tro1: "never_n\<gamma> (Bw na ss) (O [trn1]) tro1"
          and Bwn\<gamma>_trnb: "?Bwn\<gamma> (O (trn1 # tro1)) trnb"
          and tro2: "trnb # trbs = filter_n\<gamma> (Bw na ss) (O [trn1]) (tro1 @ trnb # tro2)"
          using Cons(2,5) by (auto elim: Cons_eq_filter_n\<gamma>D)

        from Cons(4,5)[unfolded tro] tro1 tro2 Bwn\<gamma>_trnb
        obtain trb trn1' tro1'
        where tro1': "O_split (Inv na ss) s tr (tr1 @ trb) (trn1' # tro1' @ tro2 ## trnn) tr2"
                     "never_n\<gamma> (Bw na ss) (O [trn1']) tro1'"
          and trn1': "\<gamma> trn1'"
        proof (induction tro1 arbitrary: tro2 trnb rule: rev_induct)
          case (Nil tro2 trnb)
            with Nil have "O_split (Inv na ss) s tr tr1 ([] @ trn1 # trnb # (tro2 ## trnn)) tr2"
              using trnn(1) by auto
            from O_splitE[OF this] Nil(3,6) trnn(1)
            have "?O_split_commute tr1 [] trn1 trnb (tro2 ## trnn) tr2
                \<or> ?O_split_absorb' tr1 [] trn1 trnb (tro2 ## trnn) tr2" by auto
            then show thesis
            proof
              assume "?O_split_commute tr1 [] trn1 trnb (tro2 ## trnn) tr2"
              then show thesis using Nil(3,6)
                by (intro Nil(1)[of "[?swap' trn1 trnb]" "?swap trn1 trnb" "[]"])
                   (auto intro: O_split_left_neutral simp: equivFiltermap_def)
            next
              assume "?O_split_absorb' tr1 [] trn1 trnb (tro2 ## trnn) tr2"
              then show thesis using Nil(3) by (intro Nil(1)[of "[]" trn1 "[]"]) auto
            qed
        next
          case (snoc trno tro1'' tro2 trnb)
            from snoc have split: "O_split (Inv na ss) s tr tr1 ((trn1 # tro1'') @ trno # trnb # (tro2 ## trnn)) tr2"
              and Bwn\<gamma>_trno: "Bw na ss (O (trn1 # tro1'')) trno \<longrightarrow> \<gamma> trno"
              and tro1'': "never_n\<gamma> (Bw na ss) [] (trn1 # tro1'')"
              by (auto simp: never_n\<gamma>_append split: if_splits)
            then have "?O_split_commute tr1 (trn1 # tro1'') trno trnb (tro2 ## trnn) tr2
                     \<or> ?O_split_absorb tr1 (trn1 # tro1'') trno trnb (tro2 ## trnn) tr2
                     \<or> ?O_split_absorb' tr1 (trn1 # tro1'') trno trnb (tro2 ## trnn) tr2"
              using snoc(4,7) trnn(1) by (intro O_splitE) auto
            then show thesis
            proof (elim disjE)
              assume c: "?O_split_commute tr1 (trn1 # tro1'') trno trnb (tro2 ## trnn) tr2"
              then have trno':
                "equivFiltermap \<gamma> g (?swap trno trnb) trno"
                "Bw na ss (O ((trn1 # tro1'') ## (?swap' trno trnb))) (?swap trno trnb)
                = Bw na ss (O (trn1 # tro1'')) trno"
                and trnb':
                "\<not>\<gamma> (?swap' trno trnb)"
                "Bw na ss (local.O (trn1 # tro1'')) (?swap' trno trnb)"
                using snoc(7) by (auto simp: equivFiltermap_def)
              show thesis proof (rule snoc(1))
                show "O_split (Inv na ss) s tr tr1 (trn1 # (tro1'' @ ?swap' trno trnb # (?swap trno trnb # tro2)) ## trnn) tr2" using c by auto
                show "\<gamma> trn1" "never_n\<gamma> (Bw na ss) (O [trn1]) tro1''" using snoc(4,5) by (auto simp: never_n\<gamma>_append)
                show "?swap' trno trnb # trbs = filter_n\<gamma> (Bw na ss) (O [trn1]) (tro1'' @ ?swap' trno trnb # ?swap trno trnb # tro2)" (*using snoc(5,6,7,8) c Bwn\<gamma>_trno*)
                  using snoc(5,6,7) trno' trnb' Bwn\<gamma>_trno
                  unfolding never_n\<gamma>_filter_n\<gamma>_Nil
                  by (auto simp: equivFiltermap_def filter_n\<gamma>_append O_Cons_unfold split: if_splits)
                show Bwn\<gamma>_trnb': "?Bwn\<gamma> (O (trn1 # tro1'')) (?swap' trno trnb)"
                  using snoc(7) c by (auto simp: equivFiltermap_def)
              next
                fix trb trn1' tro1'
                assume IH: "never_n\<gamma> (Bw na ss) (O [trn1']) tro1'" "\<gamma> trn1'"
                  "O_split (Inv na ss) s tr (tr1 @ trb) (trn1' # tro1' @ (?swap trno trnb # tro2) ## trnn) tr2"
                moreover have "O (trn1 # tro1'' ## trno) = O (trn1' # tro1' ## ?swap trno trnb)"
                proof -
                  from IH split have "O (trn1' # tro1' @ ?swap trno trnb # tro2 ## trnn) =
                                      O (((trn1 # tro1'') @ trno # trnb # tro2 ## trnn))"
                    unfolding O_split_def by auto
                  then show "O (trn1 # tro1'' ## trno) = O (trn1' # tro1' ## ?swap trno trnb)"
                    using IH(2) snoc(4,7) by (auto simp: O_Cons_unfold split: if_splits)
                qed
                moreover have "\<not>\<gamma> (?swap' trno trnb)" using snoc(7) c by (auto simp: equivFiltermap_def)
                ultimately show thesis using Bwn\<gamma>_trno trno' snoc(4)
                  by (intro snoc(2)[of trb trn1' "tro1' ## ?swap trno trnb"])
                     (auto simp: never_n\<gamma>_append equivFiltermap_def)
              qed
            next
              assume a: "?O_split_absorb tr1 (trn1 # tro1'') trno trnb (tro2 ## trnn) tr2"
              show thesis proof (cases "?Bwn\<gamma> (O (trn1 # tro1'')) (?swap' trno trnb)")
                case True
                show thesis proof (rule snoc(1))
                  show "O_split (Inv na ss) s tr tr1 (trn1 # (tro1'' @ ?swap' trno trnb # tro2) ## trnn) tr2" using a by auto
                  show "\<gamma> trn1" "never_n\<gamma> (Bw na ss) (O [trn1]) tro1''" using snoc(4,5)
                    by (auto simp: never_n\<gamma>_append)
                  show "?swap' trno trnb # trbs
                    = filter_n\<gamma> (Bw na ss) (O [trn1]) (tro1'' @ ?swap' trno trnb # tro2)"
                    using snoc(5,6) a True
                    unfolding never_n\<gamma>_filter_n\<gamma>_Nil
                    by (auto simp: filter_n\<gamma>_append O_Cons_unfold equivFiltermap_def split: if_splits)
                  show "?Bwn\<gamma> (O (trn1 # tro1'')) (?swap' trno trnb)" using True by auto
                next
                  fix trb trn1' tro1'
                  assume "O_split (Inv na ss) s tr (tr1 @ trb) (trn1' # tro1' @ tro2 ## trnn) tr2"
                         "never_n\<gamma> (Bw na ss) (O [trn1']) tro1'" "\<gamma> trn1'"
                  then show thesis using snoc(5) a by (intro snoc(2)[of trb trn1' tro1']) auto
                qed
              next
                case False
                then show thesis using snoc(4,5) a
                  by (intro snoc(2)[of "[]" trn1 "tro1'' ## ?swap' trno trnb"])
                     (auto simp: never_n\<gamma>_append)
              qed
            next
              assume a': "?O_split_absorb' tr1 (trn1 # tro1'') trno trnb (tro2 ## trnn) tr2"
              with snoc(4,5) show thesis by (intro snoc(2)[of "[]" trn1 "tro1'' ## trno"]) auto
            qed
        qed
        show thesis
        proof (rule Cons(1)[of trn1' "tro1' @ tro2" "tr1 @ trb"])
          show "O_split (Inv na ss) s tr (tr1 @ trb) (trn1' # (tro1' @ tro2) ## trnn) tr2"
            using tro1'(1) by simp
          then have "O (trn1' # (tro1' @ tro2) ## trnn) = O (trn1 # tro ## trnn)"
            using Cons(4) unfolding O_split_def by auto
          then have "g trn1' = g trn1" "O tro1' = O tro1"
            using Cons(5) trn1' Bwn\<gamma>_trnb unfolding tro by auto
          then show "trbs = filter_n\<gamma> (Bw na ss) (O [trn1']) (tro1' @ tro2)"
            using tro2 tro1'(2) tro1 never_n\<gamma>_filter_n\<gamma>_Nil Bwn\<gamma>_trnb
            using trn1' Cons(5)
            by (auto simp: filter_n\<gamma>_append O_Cons_unfold split: if_splits)
          show "\<gamma> trn1'" using trn1' .
        next
          fix trl trn1' tro'
          assume "O_split (Inv na ss) s tr ((tr1 @ trb) @ trl) (trn1' # tro' ## trnn) tr2"
                 "never_n\<gamma> (Bw na ss) (O [trn1']) tro'" "\<gamma> trn1'"
          then show thesis by (intro Cons(3)[of "trb @ trl" trn1' tro']) auto
        qed
    qed

    from this trnn(1)
    obtain trm trnn' trr
    where "O_split (Inv na ss) s tr (tr1 @ trl) (trn1' # trm ## trnn') (trr @ tr2)"
      and "never_n\<gamma> (Bw na ss) [] (trn1' # trm ## trnn')"
      and "never_n\<gamma> (Fw na ss) [] (trn1' # trm ## trnn')"
    proof (induction "filter_n\<gamma> (Fw na ss) (O [trn1']) tro'" arbitrary: tro' tr2 trnn rule: rev_induct)
      case (Nil tro' tr2 trnn)
      then show ?case
        by (intro Nil(2)[of tro' trnn "[]"]) (auto simp: never_n\<gamma>_filter_n\<gamma>_Nil filter_n\<gamma>_append)
    next
      case (snoc trnf trfs tro' tr2 trnn)
      obtain tro1 tro2 where tro': "tro' = tro1 @ trnf # tro2"
        and never_Bw_tro2: "never_n\<gamma> (Bw na ss) (O (trn1' # tro1 ## trnf)) tro2"
        and never_Fw_tro2: "never_n\<gamma> (Fw na ss) (O (trn1' # tro1 ## trnf)) tro2"
        and Fwn\<gamma>_trnf: "?Fwn\<gamma> (O (trn1' # tro1)) trnf"
        and tro1: "trfs ## trnf = filter_n\<gamma> (Fw na ss) (O [trn1']) (tro1 ## trnf)"
        using snoc(2,5)
        by (auto elim: snoc_eq_filter_n\<gamma>D split: if_splits
                 simp: O_Cons_unfold never_n\<gamma>_filter_n\<gamma>_Nil filter_n\<gamma>_append)

      from snoc(4,5,7)[unfolded tro'] tro1 never_Bw_tro2 never_Fw_tro2 Fwn\<gamma>_trnf
      obtain trf trnn' tro2'
      where tro2': "O_split (Inv na ss) s tr (tr1 @ trl) (trn1' # tro1 @ tro2' ## trnn') (trf @ tr2)"
                   "never_n\<gamma> (Bw na ss) (O (trn1' # tro1)) tro2'"
                   "never_n\<gamma> (Fw na ss) (O (trn1' # tro1)) tro2'"
        and trnn': "\<gamma> trnn'"
      proof (induction tro2 arbitrary: tro1 trnf)
        case (Nil tro1 trnf)
        with Nil have "O_split (Inv na ss) s tr (tr1 @ trl) ((trn1' # tro1) @ trnf # trnn # []) tr2"
          using trnn(1) by auto
        from O_splitE[OF this] Nil(4,8) snoc(6)
        have "?O_split_commute (tr1 @ trl) (trn1' # tro1) trnf trnn [] tr2
            \<or> ?O_split_absorb (tr1 @ trl) (trn1' # tro1) trnf trnn [] tr2" by auto
        then show thesis
        proof
          assume "?O_split_commute (tr1 @ trl) (trn1' # tro1) trnf trnn [] tr2"
          moreover then have "\<not>\<gamma> (?swap trnf trnn)" "\<gamma> (?swap' trnf trnn)"
            using Nil(4,8) by (auto simp: equivFiltermap_def)
          ultimately show thesis
            by (intro Nil(1)[of "[]" "?swap' trnf trnn" "[?swap trnf trnn]"])
               (auto intro: O_split_right_neutral)
        next
          assume "?O_split_absorb (tr1 @ trl) (trn1' # tro1) trnf trnn [] tr2"
          then show thesis using Nil(4)
            by (intro Nil(1)[of "[]" "?swap' trnf trnn" "[]"]) (auto simp: equivFiltermap_def)
        qed
      next
        case (Cons trno tro2'' tro1 trnf)
        from Cons.prems have "O_split (Inv na ss) s tr (tr1 @ trl) ((trn1' # tro1) @ trnf # trno # (tro2'' ## trnn)) tr2"
          and "\<not>?BwFwn\<gamma> (O (trn1' # tro1 ## trnf)) trno"
          and "never_n\<gamma> (Bw na ss) (O (trn1' # tro1 @ [trnf, trno])) (tro2'' ## trnn)"
          and "never_n\<gamma> (Fw na ss) (O (trn1' # tro1 @ [trnf, trno])) (tro2'' ## trnn)"
          by (auto simp: never_n\<gamma>_append O_Cons_unfold split: if_splits)
        with O_splitE[OF this(1)] consider
            (c) "?O_split_commute (tr1 @ trl) (trn1' # tro1) trnf trno (tro2'' ## trnn) tr2"
          | (a) "?O_split_absorb (tr1 @ trl) (trn1' # tro1) trnf trno (tro2'' ## trnn) tr2"
          | (a') "?O_split_absorb' (tr1 @ trl) (trn1' # tro1) trnf trno (tro2'' ## trnn) tr2"
          using Cons(5,9) snoc(6) by auto
        then show thesis proof cases
          case c
          then have trno':
            "equivFiltermap \<gamma> g (?swap' trnf trno) trno"
            "Fw na ss (O ((trn1' # tro1) ## trnf)) (?swap' trnf trno)
            = Fw na ss (O (trn1' # tro1)) trno"
            "Bw na ss (O ((trn1' # tro1) ## trnf)) (?swap' trnf trno)
            = Bw na ss (O (trn1' # tro1)) trno"
            and trnf':
            "\<not>\<gamma> (?swap trnf trno)"
            "Fw na ss (O ((trn1' # tro1) ## ?swap' trnf trno)) (?swap trnf trno)"
            "\<not>Bw na ss (O ((trn1' # tro1) ## ?swap' trnf trno)) (?swap trnf trno)"
            "\<not>Bw na ss (O (trn1' # tro1)) trnf"
            using Cons(4,9) by (auto simp: never_n\<gamma>_append O_Cons_unfold equivFiltermap_def split: if_splits)
          show thesis proof (rule Cons(1))
            show "O_split (Inv na ss) s tr (tr1 @ trl) (trn1' # ((tro1 ## ?swap' trnf trno) @ ?swap trnf trno # tro2'') ## trnn) tr2" using c by auto
            show "\<gamma> trnn" "never_n\<gamma> (Bw na ss) (local.O [trn1']) ((tro1 ## ?swap' trnf trno) @ ?swap trnf trno # tro2'')"
              using Cons(4,5,9) trnf' trno'
              by (auto simp: never_n\<gamma>_append O_Cons_unfold equivFiltermap_def split: if_splits)
            show "trfs ## ?swap trnf trno = filter_n\<gamma> (Fw na ss) (local.O [trn1']) ((tro1 ## ?swap' trnf trno) ## ?swap trnf trno)"
              using Cons(6-9) trno' trnf'
              by (auto simp: filter_n\<gamma>_append O_Cons_unfold equivFiltermap_def split: if_splits)
            show "never_n\<gamma> (Bw na ss) (O (trn1' # (tro1 ## ?swap' trnf trno) ## ?swap trnf trno)) tro2''"
              using Cons(7,9) trno' trnf' by (auto simp: O_Cons_unfold equivFiltermap_def split: if_splits)
            show "never_n\<gamma> (Fw na ss) (O (trn1' # (tro1 ## ?swap' trnf trno) ## ?swap trnf trno)) tro2''"
              using Cons(8,9) trno' trnf' by (auto simp: O_Cons_unfold equivFiltermap_def split: if_splits)
            show "?Fwn\<gamma> (O (trn1' # tro1 ## ?swap' trnf trno)) (?swap trnf trno)"
              using Cons(9) trnf' by auto
          next
            fix tro2' trnn' trf
            assume "O_split (Inv na ss) s tr (tr1 @ trl) (trn1' # tro1 ## ?swap' trnf trno @ tro2' ## trnn') (trf @ tr2)"
                   "never_n\<gamma> (Bw na ss) (O (trn1' # tro1 ## ?swap' trnf trno)) tro2'" "\<gamma> trnn'"
                   "never_n\<gamma> (Fw na ss) (O (trn1' # tro1 ## ?swap' trnf trno)) tro2'"
            then show thesis using Cons(7,8,9) trno' trnf'
              by (intro Cons(2)[of "?swap' trnf trno # tro2'" trnn' trf])
                 (auto simp: never_n\<gamma>_append O_Cons_unfold equivFiltermap_def split: if_splits)
          qed
        next
          assume a: "?O_split_absorb (tr1 @ trl) (trn1' # tro1) trnf trno (tro2'' ## trnn) tr2"
          then have nBwg\<gamma>_trno': "\<not>?Bwn\<gamma> (O (trn1' # tro1)) (?swap' trnf trno)"
            using Cons(4,9) by (auto simp: equivFiltermap_def O_Cons_unfold never_n\<gamma>_append split: if_splits)
          show thesis proof (cases "?Fwn\<gamma> (O (trn1' # tro1)) (?swap' trnf trno)")
            case True
            show thesis proof (rule Cons(1)[of tro1 "?swap' trnf trno"])
              show "O_split (Inv na ss) s tr (tr1 @ trl) (trn1' # (tro1 @ ?swap' trnf trno # tro2'') ## trnn) tr2" using a by auto
              show "never_n\<gamma> (Bw na ss) (O [trn1']) (tro1 @ ?swap' trnf trno # tro2'')"
                using Cons(4) a
                by (auto simp: never_n\<gamma>_append O_Cons_unfold equivFiltermap_def split: if_splits)
              show "\<gamma> trnn" using Cons(5) .
              show "trfs ## ?swap' trnf trno = filter_n\<gamma> (Fw na ss) (O [trn1']) (tro1 ## ?swap' trnf trno)"
                using Cons(6,9) True by (auto simp: filter_n\<gamma>_append O_Cons_unfold)
              show "never_n\<gamma> (Bw na ss) (O (trn1' # tro1 ## ?swap' trnf trno)) tro2''"
                using Cons(7) a by (auto simp: O_Cons_unfold equivFiltermap_def split: if_splits)
              show "never_n\<gamma> (Fw na ss) (O (trn1' # tro1 ## ?swap' trnf trno)) tro2''"
                using Cons(8) a by (auto simp: O_Cons_unfold equivFiltermap_def split: if_splits)
              show "?Fwn\<gamma> (O (trn1' # tro1)) (?swap' trnf trno)" using True .
            next
              fix tro2' trnn' trf
              assume "O_split (Inv na ss) s tr (tr1 @ trl) (trn1' # tro1 @ tro2' ## trnn') (trf @ tr2)"
                 and "never_n\<gamma> (Bw na ss) (O (trn1' # tro1)) tro2'" and "\<gamma> trnn'"
                 and "never_n\<gamma> (Fw na ss) (O (trn1' # tro1)) tro2'"
              then show thesis by (intro Cons(2)[of tro2' trnn' trf]) auto
            qed
          next
            case False
            with Cons(5,4,7,8) a show thesis
              by (intro Cons(2)[of "?swap' trnf trno # tro2''" trnn "[]"])
                 (auto simp: never_n\<gamma>_append O_Cons_unfold equivFiltermap_def split: if_splits)
          qed
        next
          assume a': "?O_split_absorb' (tr1 @ trl) (trn1' # tro1) trnf trno (tro2'' ## trnn) tr2"
          show thesis proof (rule Cons(1))
            show "O_split (Inv na ss) s tr (tr1 @ trl) (trn1' # (tro1 @ trnf # tro2'') ## trnn) tr2" using a' by auto
            show "\<gamma> trnn" using Cons(5) .
            show "never_n\<gamma> (Bw na ss) (O [trn1']) (tro1 @ trnf # tro2'')"
              using Cons(4) a' by (auto simp: never_n\<gamma>_append)
            show "trfs ## trnf = filter_n\<gamma> (Fw na ss) (local.O [trn1']) (tro1 ## trnf)"
              using Cons(6) .
            show "never_n\<gamma> (Bw na ss) (O (trn1' # tro1 ## trnf)) tro2''"
              using Cons(7) a' by (auto simp: never_n\<gamma>_append O_Cons_unfold)
            show "never_n\<gamma> (Fw na ss) (O (trn1' # tro1 ## trnf)) tro2''"
              using Cons(8) a' by (auto simp: never_n\<gamma>_append O_Cons_unfold)
            show "?Fwn\<gamma> (O (trn1' # tro1)) trnf" using Cons(9) .
          next
            fix tro2' trnn' trf
            assume "O_split (Inv na ss) s tr (tr1 @ trl) (trn1' # tro1 @ tro2' ## trnn') (trf @ tr2)"
              and "never_n\<gamma> (Bw na ss) (O (trn1' # tro1)) tro2'" and "\<gamma> trnn'"
              and "never_n\<gamma> (Fw na ss) (O (trn1' # tro1)) tro2'"
            then show thesis using a' by (intro Cons(2)[of tro2' trnn' trf]) auto
          qed
        qed
      qed
      show thesis
      proof (rule snoc(1)[of "tro1 @ tro2'" "trf @ tr2" trnn'])
        show "trfs = filter_n\<gamma> (Fw na ss) (O [trn1']) (tro1 @ tro2')" using tro1 tro2'(2,3) Fwn\<gamma>_trnf
          by (auto simp: filter_n\<gamma>_append O_Cons_unfold never_n\<gamma>_filter_n\<gamma>_Nil)
        show "O_split (Inv na ss) s tr (tr1 @ trl) (trn1' # (tro1 @ tro2') ## trnn') (trf @ tr2)"
          using tro2'(1) by simp
        show "\<gamma> trn1'"  "\<gamma> trnn'" using trn1' trnn' .
        show "never_n\<gamma> (Bw na ss) (O [trn1']) (tro1 @ tro2')" using snoc(5) tro2'(2) unfolding tro'
          by (auto simp: never_n\<gamma>_append O_Cons_unfold)
      next
        fix trm trnn' trr
        assume "O_split (Inv na ss) s tr (tr1 @ trl) (trn1' # trm ## trnn') (trr @ trf @ tr2)"
               "never_n\<gamma> (Bw na ss) [] (trn1' # trm ## trnn')"
               "never_n\<gamma> (Fw na ss) [] (trn1' # trm ## trnn')"
        then show thesis by (intro snoc(3)[of trm trnn' "trr @ trf"]) auto
      qed
    qed

    then show thesis by (elim that) auto
  qed
  then have "O_split (\<lambda>_. True) s tr tr1 (traceOf (?s' s tr1) (\<epsilon>a na (?s' s tr1))) tr2" using `?Oss \<noteq> []` r O
  proof (elim O_split_replace_trm)
    show "validFrom (?s' s tr1) (traceOf (?s' s tr1) (\<epsilon>a na (?s' s tr1)))" by simp
    have s': "reach (?s' s tr1)" using r(2) split unfolding O_split_def by (auto intro: validFrom_reach)
    show "O (traceOf (?s' s tr1) (\<epsilon>a na (?s' s tr1))) = O trm
        \<and> V (traceOf (?s' s tr1) (\<epsilon>a na (?s' s tr1))) = V trm
        \<and> tgtOf (last (traceOf (?s' s tr1) (\<epsilon>a na (?s' s tr1)))) = tgtOf (last trm)"
      using split trm unfolding O_split_def O
      by (intro O_\<epsilon>a_invE[OF O_\<epsilon>a_inv[OF r(1)] s' _ _ \<gamma>_ss]) (auto iff: list_all_iff)
  qed (auto iff: list_all_iff)
  then have "local_right na ss s tr tr1 tr2" using `?Oss \<noteq> []` unfolding O_split_def O local_right_def by auto
  then show "\<exists>tr1 tr2. local_right na ss s tr tr1 tr2" by blast
qed

theorem stepwise_borrow_secure_effective_inv:
assumes (* should be immediate to check: *)
"\<And> na ss. reach ss \<Longrightarrow> \<phi>_\<epsilon>a na ss" 
and  (* much less tedious to check than local_adjust: *) 
\<epsilon>a_inv: "\<And> na ss. reach ss \<Longrightarrow> \<epsilon>a_inv na ss (Inv na ss)" and 
\<epsilon>a_commute_inv: "\<And> na ss. reach ss \<Longrightarrow> \<epsilon>a_commute_inv na ss (Bw na ss) (Fw na ss) (Inv na ss)" 
and O_\<epsilon>a_inv: "\<And> na ss. reach ss \<Longrightarrow> O_\<epsilon>a_inv na ss (Bw na ss) (Fw na ss) (Inv na ss)"
shows "secure \<Longrightarrow> secure'"
using \<epsilon>a_commute_inv_local_adjust[OF \<epsilon>a_inv \<epsilon>a_commute_inv O_\<epsilon>a_inv] assms 
stepwise_borrow_secure by auto

(* BELOW IS ANDREI TERRITORY  *)
(* Further simplification of the conditions: *)


definition "\<epsilon>a_\<gamma>a_inj na ss Inv \<equiv> \<forall> s. let al = filter \<gamma>a (\<epsilon>a na ss) in
  reach s \<and> Inv s \<and> fst (sstep ss al) = fst (sstep s al) \<longrightarrow> al = filter \<gamma>a (\<epsilon>a na s)"
definition "exhaust Fw Bw \<equiv> \<forall> obsl trn. Bw obsl trn \<or> Fw obsl trn"

lemma stepConstant_O_g_traceOf: 
fixes s al assumes "\<forall> a. a \<in>\<in> al \<longrightarrow> snd (step s a) = s"
defines "obsl \<equiv> O (traceOf s al)" 
shows "map g (traceOf s (map fst obsl)) = obsl \<and> O (traceOf s (map fst obsl)) = obsl"
using assms unfolding obsl_def proof(induction al)
  case (Cons a al)
  then obtain ou where st: "step s a = (ou,s)" by (cases "step s a") auto
  show ?case using Cons by (cases "\<gamma> (Trans s a ou s)") (simp_all add: st)
qed auto

lemma O_g_\<epsilon>a: 
assumes e: "\<epsilon>a_noStateChange na s"
defines "obsl \<equiv> O (traceOf s (\<epsilon>a na s))" 
defines "al \<equiv> map fst obsl"
shows "map g (traceOf s al) = obsl \<and> O (traceOf s al) = obsl"
unfolding obsl_def al_def using stepConstant_O_g_traceOf e unfolding \<epsilon>a_noStateChange_def by auto

lemma constantState_O_filter_\<gamma>a: 
assumes g: \<gamma>_act and 0: "\<forall> a. a \<in>\<in> al \<longrightarrow> snd (step s a) = s"
shows "map fst (O (traceOf s al)) = filter \<gamma>a al"
using 0 proof(induction al)
  case Nil thus ?case by auto
next
  case (Cons a al)
  then obtain ou where st: "step s a = (ou,s)" by (cases "step s a") auto
  have [simp]: "\<gamma>a a \<longleftrightarrow> \<gamma> (Trans s a ou s)" using \<gamma>_act_\<gamma>_\<gamma>a[OF g] by auto
  show ?case using Cons by (cases "\<gamma>a a") (auto simp: st split: prod.splits)
qed

lemma \<epsilon>a_O_filter_\<gamma>a: 
assumes g: \<gamma>_act and e: "\<epsilon>a_noStateChange na s"
shows "map fst (O (traceOf s (\<epsilon>a na s))) = filter \<gamma>a (\<epsilon>a na s)"
using constantState_O_filter_\<gamma>a[OF g] e unfolding \<epsilon>a_noStateChange_def by auto

lemma O_g_filter_\<epsilon>a: 
assumes g: \<gamma>_act and e: "\<epsilon>a_noStateChange na s"
defines "obsl \<equiv> O (traceOf s (\<epsilon>a na s))" 
defines "al \<equiv> filter \<gamma>a (\<epsilon>a na s)"
shows "map g (traceOf s al) = obsl \<and> O (traceOf s al) = obsl"
using assms O_g_\<epsilon>a by (metis \<epsilon>a_O_filter_\<gamma>a)

lemma set_traceOf_actOf: 
assumes "trn \<in>\<in> traceOf s al"
shows "actOf trn \<in>\<in> al"
using assms by (induction al arbitrary: s) auto

lemma no_\<phi>_\<epsilon>a_never: "no_\<phi>_\<epsilon>a na ss \<Longrightarrow> never \<phi> (traceOf ss (\<epsilon>a na ss))"
unfolding no_\<phi>_\<epsilon>a_def never_list_ex list_ex_iff 
using set_traceOf_actOf by auto

lemma no_\<phi>_\<epsilon>a_\<phi>_\<epsilon>a: 
assumes "no_\<phi>_\<epsilon>a na ss"
shows "\<phi>_\<epsilon>a na ss"
using no_\<phi>_\<epsilon>a_never[OF assms] unfolding \<phi>_\<epsilon>a_def never_list_ex list_ex_length by auto

lemma set_map_g_traceOf: 
assumes "set (map g (traceOf s al)) \<subseteq> set (map g (traceOf s' al'))" 
shows "set al \<subseteq> set al'"
using set_map_fst[OF assms] by auto

lemma set_O_g: "set (O tr) \<subseteq> set (map g tr)"
apply(induction tr, auto) by (case_tac "\<gamma> a") auto

lemma V_traceOf_no_\<phi>_\<epsilon>a:
assumes n: "no_\<phi>_\<epsilon>a na ss" and s: "set al \<subseteq> set (\<epsilon>a na ss)"
and r: "reach ss"
shows "V (traceOf s al) = []"
proof-
  have "\<forall> trn. actOf trn \<in>\<in> al \<longrightarrow> \<not> \<phi> trn"
  using n s unfolding no_\<phi>_\<epsilon>a_def by auto
  hence "never \<phi> (traceOf s al)"
  unfolding no_\<phi>_\<epsilon>a_def never_list_ex list_ex_iff using set_traceOf_actOf by auto
  thus ?thesis by auto
qed

lemma \<epsilon>a_noStateChange_sstep:
assumes "\<epsilon>a_noStateChange na s"
shows "snd (sstep s (\<epsilon>a na s)) = s"  
using assms snd_sstep_step_constant unfolding \<epsilon>a_noStateChange_def by auto

lemma no_state_change_O_al_step_a:
assumes nc: "\<forall> a. a \<in>\<in> al \<longrightarrow> snd (step ss a) = ss"
assumes O: "(a, ou) \<in>\<in> O (traceOf ss al)"
shows "step ss a = (ou, ss)"
using O nc proof (induction al)
  case (Cons a' al)
  have "snd (step ss a') = ss" using Cons(3) by auto
  moreover have "(a, ou) \<in>\<in> O (traceOf ss al) \<longrightarrow> step ss a = (ou, ss)"
    using Cons(3) by (auto intro: Cons.IH)
  ultimately show ?case using Cons(2)
    by (cases "step ss a'") (auto simp: O_Cons_unfold split: if_splits)
qed auto

lemma fst_sstep_filter_\<gamma>a: 
assumes g: \<gamma>_act and 0: "\<forall> a. a \<in>\<in> al \<longrightarrow> snd (step s a) = s"
shows "fst (sstep s (filter \<gamma>a al)) = map snd (O (traceOf s al))"
using 0 proof(induction al)
  case (Cons a al) 
  then obtain ou where st: "step s a = (ou,s)" by (cases "step s a") auto 
  have [simp]: "\<gamma>a a \<longleftrightarrow> \<gamma> (Trans s a ou s)" using \<gamma>_act_\<gamma>_\<gamma>a[OF g] by auto
  show ?case using Cons apply (cases "\<gamma>a a") by (auto simp: st split: prod.splits)
qed auto
 
lemma fst_sstep_map_actOf: 
assumes "validFrom s tr"
shows "fst (sstep s (map actOf tr)) = map (snd \<circ> g) tr"
using assms by (induction tr arbitrary: s) auto

lemma set_actOf_traceOf_ex:
assumes "a \<in>\<in> al"
shows "\<exists> trn. trn \<in>\<in> traceOf s al \<and> actOf trn = a"
using assms apply(induction al arbitrary: s a)
by (simp_all split: prod.splits) (metis actOf.simps prod.collapse)+

lemma list_ex_\<gamma>a_\<gamma>: 
assumes \<gamma>_act 
shows "list_ex \<gamma>a al = list_ex \<gamma> (traceOf s al)"
using \<gamma>_act_\<gamma>_\<gamma>a[OF assms] set_traceOf_actOf unfolding list_ex_iff
using set_actOf_traceOf_ex by fastforce

lemma list_ex_length_O_\<gamma>a:
assumes g: \<gamma>_act and l: "list_ex \<gamma>a al"
shows "length (O (traceOf s al)) > 0"
using l list_ex_\<gamma>a_\<gamma>[OF g] list_ex_length_O by auto 

lemma length1_O_list_ex_\<gamma>a: 
assumes g: \<gamma>_act and l: "1 < length (O (traceOf s al))"
shows "list_ex \<gamma>a al"
using length1_O_list_ex_iff[OF l] list_ex_\<gamma>a_\<gamma>[OF g] by auto

lemma O_traceOf_noStateCh: 
assumes g: \<gamma>_act and nc: "\<And> a. a \<in>\<in> al \<longrightarrow> snd (step s a) = s"
shows "O (traceOf s al) = map g (traceOf s (filter \<gamma>a al))"
using nc by (induct al) (auto split: prod.splits simp: \<gamma>_act_\<gamma>_\<gamma>a[OF g])

lemma traceOf_srcOf_tgtOf_noStateCh: 
assumes "\<And> a. a \<in>\<in> al \<longrightarrow> snd (step s a) = s"
and "trn \<in>\<in> traceOf s al"
shows "srcOf trn = s \<and> tgtOf trn = s"
using assms apply(induct al arbitrary: trn) 
by (auto split: prod.splits) (metis snd_conv)+

lemma \<epsilon>a_noStateChange_srcOf_tgtOf: 
assumes "\<epsilon>a_noStateChange na ss"
and "trn \<in>\<in> traceOf ss (\<epsilon>a na ss)"
shows "srcOf trn = ss \<and> tgtOf trn = ss"
using assms traceOf_srcOf_tgtOf_noStateCh unfolding \<epsilon>a_noStateChange_def by blast

lemma exhaust_never_n\<gamma>_list_all_\<gamma>:
assumes "exhaust Bw Fw" and "never_n\<gamma> Bw obsl tr" and "never_n\<gamma> Fw obsl tr"
shows "list_all \<gamma> tr"
using assms by (induction tr arbitrary: obsl) (auto split: if_splits simp: exhaust_def)

(* This simple intro rule for O_\<epsilon>a_inv should be sufficient for the (Mini)CoCon family: *)
lemma O_\<epsilon>a_inv_simpleI: 
assumes g: \<gamma>_act and e: "\<And> s. reach s \<Longrightarrow> \<epsilon>a_noStateChange na s" 
and i: "list_ex \<gamma>a (\<epsilon>a na ss) \<Longrightarrow> \<epsilon>a_\<gamma>a_inj na ss Inv"
and ex: "exhaust Bw Fw" and n: "\<And> s. reach s \<Longrightarrow> no_\<phi>_\<epsilon>a na s"
and rs: "reach ss"
shows "O_\<epsilon>a_inv na ss Bw Fw Inv"
unfolding O_\<epsilon>a_inv_def proof safe
  fix s tr
  assume r: "reach s" 
     and O: "O tr = O (traceOf ss (\<epsilon>a na ss))"
     and s_tr: "validFrom s tr"
     and ex_\<gamma>: "list_ex \<gamma> (traceOf ss (\<epsilon>a na ss))"
     and "never_n\<gamma> Bw [] tr" "never_n\<gamma> Fw [] tr"
     and "list_all (\<lambda>trn. Inv (srcOf trn) \<and> Inv (tgtOf trn)) tr"
     (*and "list_all (\<lambda>trn. (\<gamma> trn \<or> \<not> Bw trn \<and> \<not> Fw trn) \<and> Inv (srcOf trn) \<and> Inv (tgtOf trn)) tr"*)
  hence l: "list_all \<gamma> tr" "list_all (\<lambda>trn. Inv (srcOf trn)) tr"
  using exhaust_never_n\<gamma>_list_all_\<gamma>[OF ex] by (auto simp: list_all_iff)
  hence Otr: "O tr = map g tr" by auto 
  have ex_\<gamma>a: "list_ex \<gamma>a (\<epsilon>a na ss)" using ex_\<gamma> list_ex_\<gamma>a_\<gamma>[OF g] by auto
  have trNE: "tr \<noteq> []" using O  ex_\<gamma> unfolding Otr 
  by (metis O_Nil_list_ex O_simps(1) Otr)
  have Inv_s: "Inv s" using l s_tr unfolding list_all_iff validFrom_def by (metis hd_in_set trNE)
  have 1: "V (traceOf s (\<epsilon>a na s)) = []" using no_\<phi>_\<epsilon>a_never[OF n[OF r]] by auto
  def al' \<equiv> "map actOf tr"
  have tr: "tr = traceOf s al'" using s_tr unfolding al'_def
  by (metis IO_Automaton.validFrom_traceOf2)
  have "set (map g tr) \<subseteq> set (map g (traceOf ss (\<epsilon>a na ss)))" 
  using O set_O_g unfolding Otr by auto 
  from set_map_g_traceOf[OF this [unfolded tr]] have 2: "set al' \<subseteq> set (\<epsilon>a na ss)" .
  hence "\<forall> a. a \<in>\<in> al' \<longrightarrow> snd (step s a) = s"
  using e[OF rs] unfolding \<epsilon>a_noStateChange_def by auto
  hence 22: "snd (sstep s al') = s" using snd_sstep_step_constant by auto
  from V_traceOf_no_\<phi>_\<epsilon>a[OF n[OF rs] 2 rs] 
  have 3: "V tr = []" unfolding tr no_\<phi>_\<epsilon>a_def by simp
  show "V (traceOf s (\<epsilon>a na s)) = V tr" using 1 3 by auto
  have al': "al' = filter \<gamma>a (\<epsilon>a na ss)" 
  using e[OF rs] constantState_O_filter_\<gamma>a[OF g, of "\<epsilon>a na ss" ss]  
  unfolding \<epsilon>a_noStateChange_def O[symmetric] Otr al'_def by simp
  have "fst (sstep ss al') = map snd (O (traceOf ss (\<epsilon>a na ss)))"
  unfolding al' using fst_sstep_filter_\<gamma>a[OF g] e[OF rs] unfolding \<epsilon>a_noStateChange_def by auto
  also have "\<dots> = fst (sstep s al')" 
  using fst_sstep_map_actOf[OF s_tr] unfolding O[symmetric] al'_def Otr by auto
  finally have "fst (sstep ss al') = fst (sstep s al')" .
  from r this[unfolded al'] have al's: "al' = filter \<gamma>a (\<epsilon>a na s)" 
  using i ex_\<gamma>a Inv_s unfolding \<epsilon>a_\<gamma>a_inj_def al' by metis
  show Or: "O (traceOf s (\<epsilon>a na s)) = O tr"
  unfolding tr al's by (metis O_g_filter_\<epsilon>a e g r)
  have "O (traceOf ss (\<epsilon>a na ss)) \<noteq> []" using ex_\<gamma> by (metis O_Nil_list_ex)
  hence 4: "O tr \<noteq> []" unfolding O[symmetric] .
  hence eaNE: "\<epsilon>a na s \<noteq> []"  unfolding Or[symmetric] using 4 by auto
  have "snd (sstep s (\<epsilon>a na s)) = s" using \<epsilon>a_noStateChange_sstep[OF e[OF r]] .
  hence 5: "tgtOf (last (traceOf s (\<epsilon>a na s))) = s" 
  using sstep_tgtOf_traceOf[OF eaNE] by (cases "sstep s (\<epsilon>a na s)") auto
  have al'NE: "al' \<noteq> []" using trNE unfolding al'_def by auto
  have 6: "tgtOf (last tr) = s"
  using 22 sstep_tgtOf_traceOf[OF al'NE] unfolding tr by (cases "sstep s al'") auto 
  from 5 6 show "tgtOf (last (traceOf s (\<epsilon>a na s))) = tgtOf (last tr)" by blast
qed 

(*  HERE: *)
theorem stepwise_borrow_secure_EFF:
assumes (* should be immediate to check: *)
\<gamma>_act: \<gamma>_act 
and \<epsilon>a_noStateChange: "\<And> na ss. reach ss \<Longrightarrow> \<epsilon>a_noStateChange na ss" 
and no_\<phi>_\<epsilon>a: "\<And> na ss. reach ss \<Longrightarrow> no_\<phi>_\<epsilon>a na ss"
and exhaust: "\<And> na ss. reach ss \<Longrightarrow> exhaust (Bw na ss) (Fw na ss)" 
(* should be fairly easy to check: *) 
and \<epsilon>a_inv: "\<And> na ss. reach ss \<Longrightarrow> \<epsilon>a_inv na ss (Inv na ss)" 
and \<epsilon>a_commute_inv: "\<And> na ss. reach ss \<Longrightarrow> \<epsilon>a_commute_inv na ss (Bw na ss) (Fw na ss) (Inv na ss)"
and \<epsilon>a_\<gamma>a_inj: "\<And> na ss. \<lbrakk>reach ss; list_ex \<gamma>a (\<epsilon>a na ss)\<rbrakk> \<Longrightarrow> \<epsilon>a_\<gamma>a_inj na ss (Inv na ss)" 
shows "secure \<Longrightarrow> secure'"
using stepwise_borrow_secure_effective_inv
 [OF no_\<phi>_\<epsilon>a_\<phi>_\<epsilon>a[OF no_\<phi>_\<epsilon>a] \<epsilon>a_inv 
     \<epsilon>a_commute_inv O_\<epsilon>a_inv_simpleI[OF \<gamma>_act \<epsilon>a_noStateChange \<epsilon>a_\<gamma>a_inj exhaust no_\<phi>_\<epsilon>a]] by auto

thm O_\<epsilon>a_inv_def
thm absorb_def
thm \<epsilon>a_\<gamma>a_inj_def
thm \<epsilon>a_commute_inv_def
thm \<epsilon>a_commute_goal_def
thm no_\<phi>_\<epsilon>a_def
thm changesState.simps

term sstep

lemma 
assumes "fst (sstep ss (\<epsilon>a na ss)) = fst (sstep s (\<epsilon>a na ss))"
shows "\<epsilon>a na ss = \<epsilon>a na s"
oops

end (* context Extension *)


end

