(* State-event  *)
theory Event_System
imports "../Preliminaries/Trivia"
begin

(* Traces *)

type_synonym 'trans trace = "'trans list"


locale Event_System =
fixes istate :: "'state"
  and validTrans :: "'state \<Rightarrow> 'trans \<Rightarrow> 'state \<Rightarrow> bool"
begin


(* Traces allowed by the system (starting in any given state) *)

(* Two alternative definitions: growing from the left and growing from the right: *)
inductive validTrace :: "'state \<Rightarrow> 'trans trace \<Rightarrow> 'state \<Rightarrow> bool" where
  Nil[simp,intro!]: "validTrace s [] s"
| Cons[intro]:
"\<lbrakk>validTrans s trn s'; validTrace s' tr s''\<rbrakk>
 \<Longrightarrow>
 validTrace s (trn # tr) s''"

inductive_cases validTrace_NilE[elim]: "validTrace s [] s'"
(*inductive_cases valid_SinglE[elim!]: "valid [trn]"*)
inductive_cases validTrace_ConsE[elim]: "validTrace s (trn # tr) s''"

lemma validTrace_append:
"validTrace s tr s' \<Longrightarrow> validTrace s' tr' s'' \<Longrightarrow> validTrace s (tr @ tr') s''"
by (induction rule: validTrace.induct) auto

lemma validTrace_Rcons:
assumes "validTrace s tr s'" and "validTrans s' trn s''"
shows "validTrace s (tr ## trn) s''"
using assms by (intro validTrace_append) auto

(*lemma valid_hd_Rcons[simp]:
assumes "valid tr"
shows "hd (tr ## tran) = hd tr"
by (metis Nil_not_valid assms hd_append)

lemma valid2_hd_Rcons[simp]:
assumes "valid2 tr"
shows "hd (tr ## tran) = hd tr"
by (metis Nil_not_valid2 assms hd_append)

lemma valid2_last_Cons[simp]:
assumes "valid2 tr"
shows "last (tran # tr) = last tr"
by (metis Nil_not_valid2 assms last.simps)

lemma valid2_Cons:
assumes "valid2 tr" and "tgtOf trn = srcOf (hd tr)" and "validTrans trn"
shows "valid2 (trn # tr)"
using assms proof(induct arbitrary: trn)
  case Singl  show ?case
  unfolding two_singl_Rcons apply(rule valid2.Rcons) using Singl
  by (auto intro: valid2.Singl)
next
  case Rcons show ?case
  unfolding append.append_Cons[symmetric] apply(rule valid2.Rcons) using Rcons by auto
qed

lemma valid_valid2: "valid = valid2"
proof(rule ext, safe)
  fix tr assume "valid tr"  thus "valid2 tr"
  by (induct) (auto intro: valid2.Singl valid2_Cons)
next
  fix tr assume "valid2 tr"  thus "valid tr"
  by (induct) (auto intro: valid.Singl valid_Rcons)
qed

lemma valid_Cons_iff: 
"valid (trn # tr) \<longleftrightarrow> validTrans trn \<and> ((tgtOf trn = srcOf (hd tr) \<and> valid tr) \<or> tr = [])"
unfolding valid.simps[of "trn # tr"] by auto

lemma valid_append: 
"tr \<noteq> [] \<Longrightarrow> tr1 \<noteq> [] \<Longrightarrow> 
 valid (tr @ tr1) \<longleftrightarrow> valid tr \<and> valid tr1 \<and> tgtOf (last tr) = srcOf (hd tr1)"
by (induct tr) (auto simp add: valid_Cons_iff)*)

lemma validTrace_appendE:
assumes "validTrace s (tr @ tr') s''"
obtains s' where "validTrace s tr s'" and "validTrace s' tr' s''"
using assms proof (induction tr arbitrary: s)
  case (Cons trn tr s)
    then obtain s1 where trn: "validTrans s trn s1" and s1: "validTrace s1 (tr @ tr') s''" by auto
    show thesis proof (rule Cons.IH[OF _ s1])
      fix s'
      assume "validTrace s1 tr s'" and "validTrace s' tr' s''"
      with trn show thesis using Cons(2) by blast
    qed
qed auto

definition validSystemTrace :: "'trans trace \<Rightarrow> bool"
where "validSystemTrace tr \<equiv> (\<exists>s. validTrace istate tr s)"

lemma validTrace_validSystemTrace: "validTrace istate tr s \<Longrightarrow> validSystemTrace tr"
unfolding validSystemTrace_def by auto

lemma validSystemTrace_validTrace:
assumes "validSystemTrace tr"
obtains s where "validTrace istate tr s"
using assms unfolding validSystemTrace_def by auto

(*lemmas valid2_valid = valid_valid2[symmetric]*)

(* validFrom includes empty traces too (unlike valid): *)
(*definition validFrom :: "'state \<Rightarrow> 'trans trace \<Rightarrow> bool" where
"validFrom s tr \<equiv> tr = [] \<or> (valid tr \<and> srcOf (hd tr) = s)"

lemma validFrom_Nil[simp,intro!]: "validFrom s []"
unfolding validFrom_def by auto

lemma validFrom_valid[simp,intro]: "valid tr \<and> srcOf (hd tr) = s \<Longrightarrow> validFrom s tr"
unfolding validFrom_def by auto

lemma validFrom_append: 
"validFrom s (tr @ tr1) \<longleftrightarrow> (tr = [] \<and> validFrom s tr1) \<or> (tr \<noteq> [] \<and> validFrom s tr \<and> validFrom (tgtOf (last tr)) tr1)"
unfolding validFrom_def using valid_append 
by (cases "tr = [] \<or> tr1 = []") fastforce+

lemma validFrom_Cons: 
"validFrom s (trn # tr) \<longleftrightarrow> validTrans trn \<and> srcOf trn = s \<and> validFrom (tgtOf trn) tr"
unfolding validFrom_def by auto*)


section{* Reachability *}

(* Reachable states: *)
inductive reach :: "'state \<Rightarrow> bool" where
Istate: "reach istate"
|
Step: "reach s \<Longrightarrow> validTrans s trn s' \<Longrightarrow> reach s'"


(* traces versus reachability: *)
lemma valid_reach_src_tgt:
assumes "validTrace s tr s'" and "reach s"
shows "reach s'"
using assms Step by induct auto

lemma valid_init_reach:
assumes "validTrace s tr s'" and "s = istate"
shows "reach s'"
using valid_reach_src_tgt assms reach.Istate by metis

lemma reach_validTrace:
assumes "reach s"
obtains tr where "validTrace istate tr s"
using assms proof (induction rule: reach.induct)
  case Istate show thesis using Istate.prems[of "[]"] by auto
next
  case (Step s trn s')
    show thesis proof (rule Step.IH)
      fix tr
      assume "validTrace istate tr s"
      with Step.hyps have "validTrace istate (tr ## trn) s'" by (intro validTrace_Rcons)
      then show thesis using Step.prems by blast
    qed
qed

(*inductive reachFrom :: "'state \<Rightarrow> 'state \<Rightarrow> bool"
  for s :: "'state"
where
  Refl[intro]: "reachFrom s s"
| Step: "\<lbrakk>reachFrom s s'; validTrans trn; srcOf trn = s'; tgtOf trn = s''\<rbrakk> \<Longrightarrow> reachFrom s s''"

lemma reachFrom_Step1: 
"\<lbrakk>validTrans trn; srcOf trn = s; tgtOf trn = s'\<rbrakk> \<Longrightarrow> reachFrom s s'"
by (auto intro: reachFrom.Step)

lemma reachFrom_Step_Left: 
"reachFrom s' s'' \<Longrightarrow> validTrans trn \<Longrightarrow> srcOf trn = s \<Longrightarrow> tgtOf trn = s' \<Longrightarrow> reachFrom s s''"
by (induction s'' rule: reachFrom.induct) (auto intro: reachFrom.Step)

lemma reachFrom_trans: "reachFrom s0 s1 \<Longrightarrow> reachFrom s1 s2 \<Longrightarrow> reachFrom s0 s2"
by (induction s1 arbitrary: s2 rule: reachFrom.induct) (auto intro: reachFrom_Step_Left)

lemma reachFrom_reach: "reachFrom s s' \<Longrightarrow> reach s \<Longrightarrow> reach s'"
by (induction rule: reachFrom.induct) (auto intro: reach.Step)


lemma valid_validTrans_set:
assumes "valid tr" and "trn \<in>\<in> tr"
shows "validTrans trn"
using assms by (induct tr arbitrary: trn) auto

lemma validFrom_validTrans_set:
assumes "validFrom s tr" and "trn \<in>\<in> tr"
shows "validTrans trn"
by (metis assms validFrom_def empty_iff list.set valid_validTrans_set)

lemma valid_validTrans_nth:
assumes v: "valid tr" and i: "i < length tr"
shows "validTrans (tr!i)"
using valid_validTrans_set[OF v] i by auto

lemma valid_validTrans_nth_srcOf_tgtOf:
assumes v: "valid tr" and i: "Suc i < length tr"
shows "srcOf (tr!(Suc i)) = tgtOf (tr!i)"
by (metis Cons_nth_drop_Suc valid_append Suc_lessD append_self_conv2 hd_drop_conv_nth i id_take_nth_drop list.distinct(1) v valid_ConsE)

lemma validFrom_reach: "validFrom s tr \<Longrightarrow> reach s \<Longrightarrow> tr \<noteq> [] \<Longrightarrow> reach (tgtOf (last tr))"
by (intro valid_reach_src_tgt) (auto simp add: validFrom_def)*)


end (* locale Nondet_IO_Automaton *)






end
