theory Strict_BD_Security
imports BD_Security_IO
begin

context BD_Security_IO_Aut
begin

text \<open>Backwards-strict BD security, inspired by the backwards-strict MAKS BSPs:\<close>

definition bw_secure :: bool where
"bw_secure \<equiv>
 \<forall> tr0 tr1 vl1.
   validFrom istate (tr0 @ tr1) \<and> never T (tr0 @ tr1) \<and> B (V tr0 @ V tr1) (V tr0 @ vl1) \<longrightarrow>
   (\<exists> tr1'. validFrom istate (tr0 @ tr1') \<and> O tr1' = O tr1 \<and> V tr1' = vl1)"


text \<open>Helper lemmas\<close>

lemma validFrom_snoc:
"validFrom s (tr ## trn) \<longleftrightarrow> validTrans trn \<and> ((tr = [] \<and> srcOf trn = s) \<or> (tr \<noteq> [] \<and> validFrom s tr \<and> tgtOf (last tr) = srcOf trn))"
unfolding validFrom_append by (cases trn) (auto simp add: validFrom_Cons)

lemma validFrom_reach:
  "validFrom istate tr \<Longrightarrow> tr \<noteq> [] \<Longrightarrow> reach (tgtOf (last tr))"
unfolding validFrom_def by (intro valid_init_reach) auto

lemma validFrom_reachNT:
  "validFrom istate tr \<Longrightarrow> tr \<noteq> [] \<Longrightarrow> never T tr \<Longrightarrow> reachNT (tgtOf (last tr))"
proof (induction tr rule: rev_induct)
  case (snoc trn tr)
    show ?case
    proof (cases "tr = []")
      case True with snoc reachNT.intros show ?thesis by (auto simp add: validFrom_snoc validFrom_Cons validTrans)
    next
      case False with snoc show ?thesis by (auto simp add: validFrom_snoc validTrans intro: reachNT.Step)
    qed
qed (auto)

text \<open>Unwinding\<close>

theorem bw_unwind_secure:
assumes init: "\<And> vl vl1. B vl vl1 \<Longrightarrow> \<Delta> istate vl istate vl1"
      (*init': "\<And> s tr vl vl1. validFrom istate tr \<Longrightarrow> never T tr \<Longrightarrow> (tr = [] \<and> s = istate) \<or> (tr \<noteq> [] \<and> s = tgtOf (last tr)) \<Longrightarrow> B (V tr @ vl) (V tr @ vl1) \<Longrightarrow> \<Delta> s vl s vl1"*)
    and step: "\<And> s vl vl1 vl' vl1' a ou s' trn. \<Delta> s vl s vl1 \<Longrightarrow> reachNT s \<Longrightarrow> trn = Trans s a ou s' \<Longrightarrow> validTrans trn \<Longrightarrow> \<not>T trn \<Longrightarrow> consume trn vl vl' \<Longrightarrow> consume trn vl1 vl1' \<Longrightarrow> \<Delta> s' vl' s' vl1'"
and unwind: "unwind \<Delta>"
shows bw_secure
unfolding bw_secure_def
proof safe
  fix tr0 tr1 vl1
  assume prems: "validFrom istate (tr0 @ tr1)"
       "never T (tr0 @ tr1)"
       "B (V tr0 @ V tr1) (V tr0 @ vl1)"
  let ?s' = "\<lambda>s tr. if tr = [] then s else tgtOf (last tr)"
  have "\<Delta> istate (V tr0 @ V tr1) istate (V tr0 @ vl1)" using prems(3) init by auto
  { fix s
    assume "validFrom s (tr0 @ tr1)"
       and "reachNT s"
       and "\<Delta> s (V tr0 @ V tr1) s (V tr0 @ vl1)"
    then have \<Delta>: "\<Delta> (?s' s tr0) (V tr1) (?s' s tr0) vl1" using prems(2)
    proof (induction tr0 arbitrary: s)
      case (Cons trn tr0 s)
        obtain a ou s' where trn: "trn = Trans s a ou s'" "validTrans trn" using Cons(2) by (cases trn) (auto simp add: validFrom_Cons)
        then have "consume trn (V (trn # tr0) @ V tr1) (V tr0 @ V tr1)"
              and "consume trn (V (trn # tr0) @ vl1) (V tr0 @ vl1)"
          unfolding consume_def by auto
        moreover have s': "reachNT (tgtOf trn)" using Cons(3,5) trn by (intro reachNT.Step) auto
        ultimately have "\<Delta> s' (V tr0 @ V tr1) s' (V tr0 @ vl1)" using trn Cons(3,5) by (intro step[OF Cons(4)]) auto
        then have "\<Delta> (?s' s' tr0) (V tr1) (?s' s' tr0) vl1" using Cons(2,5) s' unfolding trn by (intro Cons.IH) (auto simp add: validFrom_Cons)
        then show ?case unfolding trn by auto
    qed auto
  } 
  from this[OF prems(1) reachNT.Istate]
  have \<Delta>: "\<Delta> (?s' istate tr0) (V tr1) (?s' istate tr0) vl1" using prems(3) init by auto
  from prems(1-3)
  have "\<exists>tr1'. validFrom (?s' istate tr0) tr1' \<and> O tr1' = O tr1 \<and> V tr1' = vl1" by (intro unwind_trace[OF unwind _ _ \<Delta>, of tr1]) (auto simp add: validFrom_append intro: reachNT.Istate validFrom_reachNT reach.Istate validFrom_reach)
  then show "\<exists>tr1'. validFrom istate (tr0 @ tr1') \<and> O tr1' = O tr1 \<and> V tr1' = vl1" using prems(1) unfolding validFrom_append by auto
qed

lemma bw_secureE:
  assumes "bw_secure"
      and "validFrom istate (tr0 @ tr1)" "never T (tr0 @ tr1)" "B (V tr0 @ V tr1) (V tr0 @ vl1)"
  obtains tr1' where "validFrom istate (tr0 @ tr1')" "O tr1' = O tr1" "V tr1' = vl1"
using assms unfolding bw_secure_def by blast



text \<open>Alternative formulation with strictness relation\<close>


definition strictly_secure :: "('state, 'act, 'out) trace rel \<Rightarrow> bool" where
"strictly_secure S \<equiv>
 \<forall> tr vl vl1.
   validFrom istate tr \<and> never T tr \<and> B vl vl1 \<and> V tr = vl \<longrightarrow>
   (\<exists> tr1. validFrom istate tr1 \<and> O tr1 = O tr \<and> V tr1 = vl1 \<and> (tr, tr1) \<in> S)"

definition S_bw :: "('state, 'act, 'out) trace rel" where
"S_bw \<equiv> {(t, t') | t t'. \<forall>\<alpha>. prefixeq \<alpha> t \<and> prefixeq (V \<alpha>) (V t') \<longrightarrow> prefixeq \<alpha> t'}"


text \<open>Helper lemmas\<close>

lemma strictly_secureE:
  assumes "strictly_secure S"
      and "validFrom istate tr" "never T tr" "B vl vl1" "V tr = vl"
  obtains tr1 where "validFrom istate tr1" "O tr1 = O tr" "V tr1 = vl1" "(tr, tr1) \<in> S"
using assms unfolding strictly_secure_def by blast

lemma S_bwI:
  assumes "\<And>\<alpha>. prefixeq \<alpha> t \<Longrightarrow> prefixeq (V \<alpha>) (V t') \<Longrightarrow> prefixeq \<alpha> t'"
  shows "(t, t') \<in> S_bw"
using assms unfolding S_bw_def by auto

lemma filtermap_ConsE:
  assumes "filtermap pred func xs = v # vl"
  obtains xs0 x xs1 where "xs = xs0 @ x # xs1"
                      and "never pred xs0"
                      and "pred x" "func x = v"
                      and "filtermap pred func xs1 = vl"
using assms by (auto dest: filtermap_eq_Cons)

lemma filtermap_appendE:
  assumes "filtermap pred func xs = vl1 @ vl2"
  obtains xs1 xs2 where "xs = xs1 @ xs2"
                    and "filtermap pred func xs1 = vl1"
                    and "filtermap pred func xs2 = vl2"
using assms by (auto dest: filtermap_eq_append)

lemma filtermap_append_ConsE:
  assumes "filtermap pred func xs = vl0 @ v # vl1"
  obtains xs0 x xs1 where "xs = xs0 @ x # xs1"
                      and "filtermap pred func xs0 = vl0"
                      and "pred x" and "func x = v"
                      and "filtermap pred func xs1 = vl1"
proof -
  from assms obtain xs0 xs' where "xs = xs0 @ xs'"
                              and "filtermap pred func xs0 = vl0"
                              and xs': "filtermap pred func xs' = v # vl1"
    by (auto elim: filtermap_appendE)
  moreover obtain xs0' x xs1 where "xs' = xs0' @ x # xs1"
                               and "never pred xs0'"
                               and "pred x" "func x = v"
                               and "filtermap pred func xs1 = vl1"
    using xs' by (auto elim: filtermap_ConsE)
  ultimately show thesis
    by (intro that[of "xs0 @ xs0'" x xs1]) (auto simp add: filtermap_append filtermap_Nil_never)
qed

lemma prefixeq_filtermap:
  assumes "prefixeq xs ys"
  shows "prefixeq (filtermap pred func xs) (filtermap pred func ys)"
using assms by (auto simp add: prefixeq_def filtermap_append)

lemma prefixeq_append_Cons_filtermap:
  assumes "prefixeq xs (ys @ y # zs)"
          "prefixeq (filtermap pred func xs) (filtermap pred func ys)"
          "pred y"
  shows "prefixeq xs ys"
proof -
  have "\<not> prefixeq (ys ## y) xs"
  proof
    assume "prefixeq (ys ## y) xs"
    then have "prefixeq (filtermap pred func (ys ## y)) (filtermap pred func xs)"
      by (intro prefixeq_filtermap)
    then have "prefixeq (filtermap pred func ys ## func y) (filtermap pred func xs)"
      using assms(3) by auto
    with assms(2) show "False" by (auto simp add: prefixeq_def)
  qed
  then show "prefixeq xs ys" using assms(1) by (auto simp add: prefixeq_append prefixeq_Cons)
qed

lemma common_prefixeq:
  assumes "prefixeq xs (ys @ z # zs)"
      and "prefixeq xs (ys @ z' # zs')"
      and "z \<noteq> z'"
  shows "prefixeq xs ys"
using assms
proof (induction ys arbitrary: xs)
  case (Nil xs) then show ?case by (cases xs) (auto simp add: prefixeq_def)
next
  case (Cons y ys xs)
    obtain xs' where "xs = [] \<or> xs = y # xs'" using Cons(2) by (cases xs) (auto simp add: prefixeq_def)
    then show "prefixeq xs (y # ys)"
    proof
      assume "xs = y # xs'"
      moreover then have "prefixeq xs' ys" using Cons by auto
      ultimately show ?thesis by auto
    qed auto
qed


text \<open>Both formulations are equivalent:\<close>

lemma bw_secure_S_bw:
  assumes "bw_secure"
  shows "strictly_secure S_bw"
using assms unfolding strictly_secure_def
proof safe
  fix tr vl vl1
  assume "bw_secure"
     and tr: "validFrom istate tr"
     and T: "never T tr"
     and B: "B (V tr) vl1"
  show "\<exists>tr1. validFrom istate tr1 \<and> O tr1 = O tr \<and> V tr1 = vl1 \<and> (tr, tr1) \<in> S_bw" (is "\<exists>tr1. ?goal tr1")
  proof (cases "V tr" "vl1" rule: prefixeq_cases)
    assume "prefixeq (V tr) vl1"
    then obtain vl1' where vl1: "vl1 = V tr @ vl1'" by (elim prefixeqE)
    with tr T B obtain tr1' where "validFrom istate (tr @ tr1')" "O tr1' = O []" "V tr1' = vl1'"
      by (auto elim: bw_secureE[OF `bw_secure`, of tr "[]", simplified])
    moreover have "(tr, tr @ tr1') \<in> S_bw" unfolding S_bw_def by auto
    ultimately have "?goal (tr @ tr1')" by (auto simp add: O_append V_append vl1)
    then show ?thesis by blast
  next
    assume "prefix vl1 (V tr)"
    then obtain v vl1' where "V tr = vl1 @ v # vl1'" by (elim prefixE')
    then obtain tr0 trn tr1 where tr': "tr = tr0 @ trn # tr1"
                              and tr0: "V tr0 = vl1"
                              and trn: "\<phi> trn" and "f trn = v"
                              and tr1: "V tr1 = vl1'"
      unfolding V_def by (auto elim: filtermap_append_ConsE)
    with tr T B obtain tr1' where tr1': "validFrom istate (tr0 @ tr1')" "O tr1' = O (trn # tr1)" "V tr1' = []"
      by (auto elim!: bw_secureE[OF `bw_secure`, of tr0 "trn # tr1" "[]"] simp add: V_append)
    moreover have "(tr0 @ trn # tr1, tr0 @ tr1') \<in> S_bw"
    proof (intro S_bwI)
      fix \<alpha>
      assume "prefixeq \<alpha> (tr0 @ trn # tr1)" and "prefixeq (V \<alpha>) (V (tr0 @ tr1'))"
      then have "prefixeq \<alpha> tr0" using tr1' trn by (auto simp add: filtermap_append V_def intro: prefixeq_append_Cons_filtermap)
      then show "prefixeq \<alpha> (tr0 @ tr1')" by auto
    qed
    ultimately have "?goal (tr0 @ tr1')" using tr0 trn tr1 unfolding tr' by (auto simp add: O_append V_append)
    then show ?thesis by blast
  next
    assume "parallel (V tr) vl1"
    then obtain vl0 v0 v1 vl0' vl1' where v0: "v0 \<noteq> v1" "V tr = vl0 @ v0 # vl0'" and vl1: "vl1 = vl0 @ v1 # vl1'" by (auto dest: parallel_decomp)
    then obtain tr0 trn tr0' where tr': "tr = tr0 @ trn # tr0'"
                              and tr0: "V tr0 = vl0"
                              and trn: "\<phi> trn" "f trn = v0"
                              and tr0': "V tr0' = vl0'"
      unfolding V_def by (auto elim: filtermap_append_ConsE)
    with tr T B obtain tr1' where tr1': "validFrom istate (tr0 @ tr1')" "O tr1' = O (trn # tr0')" "V tr1' = v1 # vl1'"
      using vl1 by (auto elim!: bw_secureE[OF `bw_secure`, of tr0 "trn # tr0'" "v1 # vl1'"] simp add: V_append)
    moreover have "(tr0 @ trn # tr0', tr0 @ tr1') \<in> S_bw"
    proof (intro S_bwI)
      fix \<alpha>
      assume \<alpha>: "prefixeq \<alpha> (tr0 @ trn # tr0')" and V_\<alpha>: "prefixeq (V \<alpha>) (V (tr0 @ tr1'))"
      from \<alpha> have "prefixeq (V \<alpha>) (V (tr0 @ trn # tr0'))" unfolding V_def by (intro prefixeq_filtermap)
      with V_\<alpha> have "prefixeq (V \<alpha>) (V tr0)" using tr0 tr1'(3) v0 trn by (auto intro: common_prefixeq simp add: V_append)
      then have "prefixeq \<alpha> tr0" using \<alpha> trn by (auto simp add: filtermap_append V_def intro: prefixeq_append_Cons_filtermap)
      then show "prefixeq \<alpha> (tr0 @ tr1')" by auto
    qed
    ultimately have "?goal (tr0 @ tr1')" using tr0 unfolding tr' vl1 by (auto simp add: O_append V_append)
    then show ?thesis by blast
  qed
qed

lemma S_bw_bw_secure:
  assumes "strictly_secure S_bw"
  shows "bw_secure"
unfolding bw_secure_def
proof safe
  fix tr0 tr1 vl1
  assume "validFrom istate (tr0 @ tr1)"
         "never T (tr0 @ tr1)"
         "B (V tr0 @ V tr1) (V tr0 @ vl1)"
  then obtain tr' where tr': "validFrom istate tr'" "O tr' = O (tr0 @ tr1)" "V tr' = (V tr0 @ vl1)" "(tr0 @ tr1, tr') \<in> S_bw"
    using assms by (auto elim: strictly_secureE simp add: V_append)
  then have "prefixeq tr0 tr'" unfolding S_bw_def by auto
  then obtain tr1' where "tr' = tr0 @ tr1'" and "O tr1' = O tr1" and "V tr1' = vl1" using tr'(2,3) by (auto simp add: prefixeq_def O_append V_append)
  then show "\<exists>tr1'. validFrom istate (tr0 @ tr1') \<and> O tr1' = O tr1 \<and> V tr1' = vl1" using tr'(1) by blast
qed

theorem S_bw_iff_bw_secure:
  "strictly_secure S_bw \<longleftrightarrow> bw_secure"
using S_bw_bw_secure bw_secure_S_bw by blast

end

end

