object HOL {

trait equal[A] {
  val `HOL.equal`: (A, A) => Boolean
}
def equal[A](a: A, b: A)(implicit A: equal[A]): Boolean = A.`HOL.equal`(a, b)
object equal {
  implicit def `Prelim.equal_paperID`: equal[Prelim.paperID] = new
    equal[Prelim.paperID] {
    val `HOL.equal` = (a: Prelim.paperID, b: Prelim.paperID) =>
      Prelim.equal_paperIDa(a, b)
  }
  implicit def `Prelim.equal_userID`: equal[Prelim.userID] = new
    equal[Prelim.userID] {
    val `HOL.equal` = (a: Prelim.userID, b: Prelim.userID) =>
      Prelim.equal_userIDa(a, b)
  }
  implicit def `Prelim.equal_confID`: equal[Prelim.confID] = new
    equal[Prelim.confID] {
    val `HOL.equal` = (a: Prelim.confID, b: Prelim.confID) =>
      Prelim.equal_confIDa(a, b)
  }
  implicit def `String.equal_char`: equal[Char] = new equal[Char] {
    val `HOL.equal` = (a: Char, b: Char) => a == b
  }
  implicit def `Prelim.equal_role`: equal[Prelim.role] = new equal[Prelim.role]
    {
    val `HOL.equal` = (a: Prelim.role, b: Prelim.role) =>
      Prelim.equal_rolea(a, b)
  }
  implicit def `Lista.equal_list`[A : equal]: equal[List[A]] = new
    equal[List[A]] {
    val `HOL.equal` = (a: List[A], b: List[A]) => Lista.equal_lista[A](a, b)
  }
}

def eq[A : equal](a: A, b: A): Boolean = equal[A](a, b)

} /* object HOL */

object Fun {

def comp[A, B, C](f: A => B, g: C => A): C => B = ((x: C) => f(g(x)))

def fun_upd[A : HOL.equal, B](f: A => B, a: A, b: B): A => B =
  ((x: A) => (if (HOL.eq[A](x, a)) b else f(x)))

} /* object Fun */

object Power {

trait power[A] extends Groups.one[A] with Groups.times[A] {
}
object power {
  implicit def `Code_Numeral.power_integer`: power[BigInt] = new power[BigInt] {
    val `Groups.times` = (a: BigInt, b: BigInt) => a * b
    val `Groups.one` = Code_Numeral.one_integera
  }
}

} /* object Power */

object Groups {

trait plus[A] {
  val `Groups.plus`: (A, A) => A
}
def plus[A](a: A, b: A)(implicit A: plus[A]): A = A.`Groups.plus`(a, b)
object plus {
  implicit def `Code_Numeral.plus_integer`: plus[BigInt] = new plus[BigInt] {
    val `Groups.plus` = (a: BigInt, b: BigInt) => a + b
  }
}

trait semigroup_add[A] extends plus[A] {
}
object semigroup_add {
  implicit def `Code_Numeral.semigroup_add_integer`: semigroup_add[BigInt] = new
    semigroup_add[BigInt] {
    val `Groups.plus` = (a: BigInt, b: BigInt) => a + b
  }
}

trait zero[A] {
  val `Groups.zero`: A
}
def zero[A](implicit A: zero[A]): A = A.`Groups.zero`
object zero {
  implicit def `Code_Numeral.zero_integer`: zero[BigInt] = new zero[BigInt] {
    val `Groups.zero` = BigInt(0)
  }
}

trait monoid_add[A] extends semigroup_add[A] with zero[A] {
}
object monoid_add {
  implicit def `Code_Numeral.monoid_add_integer`: monoid_add[BigInt] = new
    monoid_add[BigInt] {
    val `Groups.zero` = BigInt(0)
    val `Groups.plus` = (a: BigInt, b: BigInt) => a + b
  }
}

trait times[A] {
  val `Groups.times`: (A, A) => A
}
def times[A](a: A, b: A)(implicit A: times[A]): A = A.`Groups.times`(a, b)
object times {
  implicit def `Code_Numeral.times_integer`: times[BigInt] = new times[BigInt] {
    val `Groups.times` = (a: BigInt, b: BigInt) => a * b
  }
}

trait semigroup_mult[A] extends times[A] {
}
object semigroup_mult {
  implicit def `Code_Numeral.semigroup_mult_integer`: semigroup_mult[BigInt] =
    new semigroup_mult[BigInt] {
    val `Groups.times` = (a: BigInt, b: BigInt) => a * b
  }
}

trait monoid_mult[A] extends semigroup_mult[A] with Power.power[A] {
}
object monoid_mult {
  implicit def `Code_Numeral.monoid_mult_integer`: monoid_mult[BigInt] = new
    monoid_mult[BigInt] {
    val `Groups.one` = Code_Numeral.one_integera
    val `Groups.times` = (a: BigInt, b: BigInt) => a * b
  }
}

trait ab_semigroup_add[A] extends semigroup_add[A] {
}
object ab_semigroup_add {
  implicit def `Code_Numeral.ab_semigroup_add_integer`: ab_semigroup_add[BigInt]
    = new ab_semigroup_add[BigInt] {
    val `Groups.plus` = (a: BigInt, b: BigInt) => a + b
  }
}

trait comm_monoid_add[A] extends ab_semigroup_add[A] with monoid_add[A] {
}
object comm_monoid_add {
  implicit def `Code_Numeral.comm_monoid_add_integer`: comm_monoid_add[BigInt] =
    new comm_monoid_add[BigInt] {
    val `Groups.zero` = BigInt(0)
    val `Groups.plus` = (a: BigInt, b: BigInt) => a + b
  }
}

trait one[A] {
  val `Groups.one`: A
}
def one[A](implicit A: one[A]): A = A.`Groups.one`
object one {
  implicit def `Code_Numeral.one_integer`: one[BigInt] = new one[BigInt] {
    val `Groups.one` = Code_Numeral.one_integera
  }
}

} /* object Groups */

object Rings {

trait semiring[A]
  extends Groups.ab_semigroup_add[A] with Groups.semigroup_mult[A] {
}
object semiring {
  implicit def `Code_Numeral.semiring_integer`: semiring[BigInt] = new
    semiring[BigInt] {
    val `Groups.times` = (a: BigInt, b: BigInt) => a * b
    val `Groups.plus` = (a: BigInt, b: BigInt) => a + b
  }
}

trait mult_zero[A] extends Groups.times[A] with Groups.zero[A] {
}
object mult_zero {
  implicit def `Code_Numeral.mult_zero_integer`: mult_zero[BigInt] = new
    mult_zero[BigInt] {
    val `Groups.zero` = BigInt(0)
    val `Groups.times` = (a: BigInt, b: BigInt) => a * b
  }
}

trait semiring_0[A]
  extends Groups.comm_monoid_add[A] with mult_zero[A] with semiring[A] {
}
object semiring_0 {
  implicit def `Code_Numeral.semiring_0_integer`: semiring_0[BigInt] = new
    semiring_0[BigInt] {
    val `Groups.times` = (a: BigInt, b: BigInt) => a * b
    val `Groups.zero` = BigInt(0)
    val `Groups.plus` = (a: BigInt, b: BigInt) => a + b
  }
}

trait zero_neq_one[A] extends Groups.one[A] with Groups.zero[A] {
}
object zero_neq_one {
  implicit def `Code_Numeral.zero_neq_one_integer`: zero_neq_one[BigInt] = new
    zero_neq_one[BigInt] {
    val `Groups.zero` = BigInt(0)
    val `Groups.one` = Code_Numeral.one_integera
  }
}

trait semiring_1[A]
  extends Num.semiring_numeral[A] with semiring_0[A] with zero_neq_one[A] {
}
object semiring_1 {
  implicit def `Code_Numeral.semiring_1_integer`: semiring_1[BigInt] = new
    semiring_1[BigInt] {
    val `Groups.zero` = BigInt(0)
    val `Groups.one` = Code_Numeral.one_integera
    val `Groups.times` = (a: BigInt, b: BigInt) => a * b
    val `Groups.plus` = (a: BigInt, b: BigInt) => a + b
  }
}

} /* object Rings */

object Num {

trait numeral[A] extends Groups.one[A] with Groups.semigroup_add[A] {
}
object numeral {
  implicit def `Code_Numeral.numeral_integer`: numeral[BigInt] = new
    numeral[BigInt] {
    val `Groups.plus` = (a: BigInt, b: BigInt) => a + b
    val `Groups.one` = Code_Numeral.one_integera
  }
}

trait semiring_numeral[A]
  extends Groups.monoid_mult[A] with numeral[A] with Rings.semiring[A] {
}
object semiring_numeral {
  implicit def `Code_Numeral.semiring_numeral_integer`: semiring_numeral[BigInt]
    = new semiring_numeral[BigInt] {
    val `Groups.plus` = (a: BigInt, b: BigInt) => a + b
    val `Groups.one` = Code_Numeral.one_integera
    val `Groups.times` = (a: BigInt, b: BigInt) => a * b
  }
}

abstract sealed class num
final case class One() extends num
final case class Bit0(a: num) extends num
final case class Bit1(a: num) extends num

def nat_of_num(x0: num): Nat.nat = x0 match {
  case Bit1(n) => {
                    val m: Nat.nat = nat_of_num(n);
                    Nat.Suc(Nat.plus_nat(m, m))
                  }
  case Bit0(n) => {
                    val m: Nat.nat = nat_of_num(n);
                    Nat.plus_nat(m, m)
                  }
  case One() => Nat.one_nat
}

} /* object Num */

object Nat {

abstract sealed class nat
final case class zero_nat() extends nat
final case class Suc(a: nat) extends nat

def one_nat: nat = Suc(zero_nat())

def less_eq_nat(x0: nat, n: nat): Boolean = (x0, n) match {
  case (Suc(m), n) => less_nat(m, n)
  case (zero_nat(), n) => true
}

def less_nat(m: nat, x1: nat): Boolean = (m, x1) match {
  case (m, Suc(n)) => less_eq_nat(m, n)
  case (n, zero_nat()) => false
}

def plus_nat(x0: nat, n: nat): nat = (x0, n) match {
  case (Suc(m), n) => plus_nat(m, Suc(n))
  case (zero_nat(), n) => n
}

def of_nat_aux[A : Rings.semiring_1](inc: A => A, x1: nat, i: A): A =
  (inc, x1, i) match {
  case (inc, zero_nat(), i) => i
  case (inc, Suc(n), i) => of_nat_aux[A](inc, n, inc(i))
}

def of_nat[A : Rings.semiring_1](n: nat): A =
  of_nat_aux[A](((i: A) => Groups.plus[A](i, Groups.one[A])), n, Groups.zero[A])

def equal_nat(x0: nat, x1: nat): Boolean = (x0, x1) match {
  case (zero_nat(), Suc(x2)) => false
  case (Suc(x2), zero_nat()) => false
  case (Suc(x2), Suc(y2)) => equal_nat(x2, y2)
  case (zero_nat(), zero_nat()) => true
}

} /* object Nat */

object Product_Type {

def apsnd[A, B, C](f: A => B, x1: (C, A)): (C, B) = (f, x1) match {
  case (f, (x, y)) => (x, f(y))
}

} /* object Product_Type */

object Code_Numeral {

def one_integera: BigInt = BigInt(1)

def sgn_integer(k: BigInt): BigInt =
  (if (k == BigInt(0)) BigInt(0)
    else (if (k < BigInt(0)) BigInt(-1) else BigInt(1)))

def divmod_integer(k: BigInt, l: BigInt): (BigInt, BigInt) =
  (if (k == BigInt(0)) (BigInt(0), BigInt(0))
    else (if (l == BigInt(0)) (BigInt(0), k)
           else (Fun.comp[BigInt, ((BigInt, BigInt)) => (BigInt, BigInt),
                           BigInt](Fun.comp[BigInt => BigInt,
     ((BigInt, BigInt)) => (BigInt, BigInt),
     BigInt](((a: BigInt => BigInt) => (b: (BigInt, BigInt)) =>
               Product_Type.apsnd[BigInt, BigInt, BigInt](a, b)),
              ((a: BigInt) => (b: BigInt) => a * b)),
                                    ((a: BigInt) =>
                                      sgn_integer(a)))).apply(l).apply((if (sgn_integer(k) ==
                                      sgn_integer(l))
                                 ((k: BigInt) => (l: BigInt) => if (l == 0)
                                   (BigInt(0), k) else
                                   (k.abs /% l.abs)).apply(k).apply(l)
                                 else {
val (r, s): (BigInt, BigInt) =
  ((k: BigInt) => (l: BigInt) => if (l == 0) (BigInt(0), k) else
    (k.abs /% l.abs)).apply(k).apply(l);
(if (s == BigInt(0)) ((- r), BigInt(0)) else ((- r) - BigInt(1), l.abs - s))
                                      }))))

def integer_of_nat: Nat.nat => BigInt = ((a: Nat.nat) => Nat.of_nat[BigInt](a))

def nat_of_integer(k: BigInt): Nat.nat =
  (if (k <= BigInt(0)) Nat.zero_nat()
    else {
           val (l, j): (BigInt, BigInt) = divmod_integer(k, BigInt(2))
           val la: Nat.nat = nat_of_integer(l)
           val lb: Nat.nat = Nat.plus_nat(la, la);
           (if (j == BigInt(0)) lb else Nat.plus_nat(lb, Nat.one_nat))
         })

} /* object Code_Numeral */

object String {

def char_of_nat: Nat.nat => Char =
  Fun.comp[BigInt, Char,
            Nat.nat](((k: BigInt) => if (BigInt(0) <= k && k < BigInt(256)) k.charValue else error("character value out of range")),
                      Code_Numeral.integer_of_nat)

def nat_of_char: Char => Nat.nat =
  Fun.comp[BigInt, Nat.nat,
            Char](((a: BigInt) => Code_Numeral.nat_of_integer(a)),
                   ((a: Char) => BigInt(a.toInt)))

} /* object String */

object Lista {

def equal_lista[A : HOL.equal](x0: List[A], x1: List[A]): Boolean = (x0, x1)
  match {
  case (Nil, x21 :: x22) => false
  case (x21 :: x22, Nil) => false
  case (x21 :: x22, y21 :: y22) =>
    HOL.eq[A](x21, y21) && equal_lista[A](x22, y22)
  case (Nil, Nil) => true
}

def nth[A](x0: List[A], x1: Nat.nat): A = (x0, x1) match {
  case (x :: xs, Nat.Suc(n)) => nth[A](xs, n)
  case (x :: xs, Nat.zero_nat()) => x
}

def find[A](uu: A => Boolean, x1: List[A]): Option[A] = (uu, x1) match {
  case (uu, Nil) => None
  case (p, x :: xs) => (if (p(x)) Some[A](x) else find[A](p, xs))
}

def nulla[A](x0: List[A]): Boolean = x0 match {
  case Nil => true
  case x :: xs => false
}

def last[A](x0: List[A]): A = x0 match {
  case x :: xs => (if (nulla[A](xs)) x else last[A](xs))
}

def maps[A, B](f: A => List[B], x1: List[A]): List[B] = (f, x1) match {
  case (f, Nil) => Nil
  case (f, x :: xs) => f(x) ++ maps[A, B](f, xs)
}

def filter[A](p: A => Boolean, x1: List[A]): List[A] = (p, x1) match {
  case (p, Nil) => Nil
  case (p, x :: xs) => (if (p(x)) x :: filter[A](p, xs) else filter[A](p, xs))
}

def member[A : HOL.equal](x0: List[A], y: A): Boolean = (x0, y) match {
  case (Nil, y) => false
  case (x :: xs, y) => HOL.eq[A](x, y) || member[A](xs, y)
}

def insert[A : HOL.equal](x: A, xs: List[A]): List[A] =
  (if (member[A](xs, x)) xs else x :: xs)

def butlast[A](x0: List[A]): List[A] = x0 match {
  case Nil => Nil
  case x :: xs => (if (nulla[A](xs)) Nil else x :: butlast[A](xs))
}

def hd[A](x0: List[A]): A = x0 match {
  case x21 :: x22 => x21
}

def list_ex[A](p: A => Boolean, x1: List[A]): Boolean = (p, x1) match {
  case (p, Nil) => false
  case (p, x :: xs) => p(x) || list_ex[A](p, xs)
}

def remove1[A : HOL.equal](x: A, xa1: List[A]): List[A] = (x, xa1) match {
  case (x, Nil) => Nil
  case (x, y :: xs) => (if (HOL.eq[A](x, y)) xs else y :: remove1[A](x, xs))
}

def map[A, B](f: A => B, x1: List[A]): List[B] = (f, x1) match {
  case (f, Nil) => Nil
  case (f, x21 :: x22) => f(x21) :: map[A, B](f, x22)
}

def gen_length[A](n: Nat.nat, x1: List[A]): Nat.nat = (n, x1) match {
  case (n, x :: xs) => gen_length[A](Nat.Suc(n), xs)
  case (n, Nil) => n
}

def list_update[A](x0: List[A], i: Nat.nat, y: A): List[A] = (x0, i, y) match {
  case (x :: xs, Nat.Suc(i), y) => x :: list_update[A](xs, i, y)
  case (x :: xs, Nat.zero_nat(), y) => y :: xs
  case (Nil, i, y) => Nil
}

def list_all[A](p: A => Boolean, x1: List[A]): Boolean = (p, x1) match {
  case (p, Nil) => true
  case (p, x :: xs) => p(x) && list_all[A](p, xs)
}

def size_list[A]: (List[A]) => Nat.nat =
  ((a: List[A]) => gen_length[A](Nat.zero_nat(), a))

} /* object Lista */

object Fresh {

def upChar(y: List[Char]): List[Char] =
  (if (! (Lista.nulla[Char](y)) &&
         (Nat.less_eq_nat(Num.nat_of_num(Num.Bit1(Num.Bit0(Num.Bit0(Num.Bit0(Num.Bit0(Num.Bit1(Num.One()))))))),
                           String.nat_of_char.apply(Lista.last[Char](y))) &&
           Nat.less_nat(String.nat_of_char.apply(Lista.last[Char](y)),
                         Num.nat_of_num(Num.Bit0(Num.Bit1(Num.Bit0(Num.Bit1(Num.Bit1(Num.Bit1(Num.One()))))))))))
    Lista.butlast[Char](y) ++
      List(String.char_of_nat.apply(Nat.plus_nat(String.nat_of_char.apply(Lista.last[Char](y)),
          Nat.one_nat)))
    else y ++ List('a'))

def fresh(xs: List[List[Char]], y: List[Char]): List[Char] =
  (if (Lista.member[List[Char]](xs, y))
    fresh(Lista.remove1[List[Char]](y, xs), upChar(y)) else y)

} /* object Fresh */

object Prelim {

abstract sealed class paperID
final case class PaperID(a: String) extends paperID

def equal_paperIDa(x0: paperID, x1: paperID): Boolean = (x0, x1) match {
  case (PaperID(x), PaperID(ya)) => x == ya
}

abstract sealed class role
final case class Aut(a: paperID) extends role
final case class Rev(a: paperID, b: Nat.nat) extends role
final case class PC() extends role
final case class Chair() extends role

def equal_rolea(x0: role, x1: role): Boolean = (x0, x1) match {
  case (PC(), Chair()) => false
  case (Chair(), PC()) => false
  case (Rev(x21, x22), Chair()) => false
  case (Chair(), Rev(x21, x22)) => false
  case (Rev(x21, x22), PC()) => false
  case (PC(), Rev(x21, x22)) => false
  case (Aut(x1), Chair()) => false
  case (Chair(), Aut(x1)) => false
  case (Aut(x1), PC()) => false
  case (PC(), Aut(x1)) => false
  case (Aut(x1), Rev(x21, x22)) => false
  case (Rev(x21, x22), Aut(x1)) => false
  case (Rev(x21, x22), Rev(y21, y22)) =>
    equal_paperIDa(x21, y21) && Nat.equal_nat(x22, y22)
  case (Aut(x1), Aut(y1)) => equal_paperIDa(x1, y1)
  case (Chair(), Chair()) => true
  case (PC(), PC()) => true
}

abstract sealed class confID
final case class ConfID(a: String) extends confID

def equal_confIDa(x0: confID, x1: confID): Boolean = (x0, x1) match {
  case (ConfID(x), ConfID(ya)) => x == ya
}

abstract sealed class userID
final case class UserID(a: String) extends userID

def equal_userIDa(x0: userID, x1: userID): Boolean = (x0, x1) match {
  case (UserID(x), UserID(ya)) => x == ya
}

abstract sealed class exp
final case class NoExp() extends exp
final case class Zero() extends exp
final case class One() extends exp
final case class Two() extends exp
final case class Three() extends exp
final case class Four() extends exp

abstract sealed class conf
final case class Conf(a: String, b: String) extends conf

abstract sealed class user
final case class User(a: String, b: String, c: String) extends user

abstract sealed class discussion
final case class Dis(a: List[String]) extends discussion

abstract sealed class fpcontent
final case class NoFPContent() extends fpcontent
final case class FPContent(a: java.io.File) extends fpcontent

abstract sealed class pcontent
final case class NoPContent() extends pcontent
final case class PContent(a: java.io.File) extends pcontent

abstract sealed class decision
final case class NoDecision() extends decision
final case class Accept() extends decision
final case class Reject() extends decision

abstract sealed class score
final case class NoScore() extends score
final case class MinusThree() extends score
final case class MinusTwo() extends score
final case class MinusOne() extends score
final case class Zeroa() extends score
final case class Onea() extends score
final case class Twoa() extends score
final case class Threea() extends score

abstract sealed class paper
final case class
  Paper(a: String, b: String, c: pcontent,
         d: List[List[(exp, (score, String))]], e: discussion,
         f: List[decision], g: fpcontent)
  extends paper

abstract sealed class password
final case class Password(a: String) extends password

abstract sealed class preference
final case class NoPref() extends preference
final case class Want() extends preference
final case class Would() extends preference
final case class WouldNot() extends preference
final case class Conflict() extends preference

def SucPH(ph: Nat.nat): Nat.nat =
  (if (Nat.equal_nat(ph, Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.zero_nat())))))))))
    Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.zero_nat())))))))
    else Nat.Suc(ph))

def decsPaper(x0: paper): List[decision] = x0 match {
  case Paper(title, abstracta, content, reviews, dis, decs, fcontent) => decs
}

def decPaper(pap: paper): decision = Lista.hd[decision](decsPaper(pap))

def emptyDis: discussion = Dis(Nil)

def emptyStr: String = ""

def getFresh(strs: List[String], str: String): String =
  ("" ++
    (Fresh.fresh(Lista.map[String,
                            List[Char]](((a: String) => (a.toList)), strs),
                  (str.toList))))

def infoConf(x0: conf): String = x0 match {
  case Conf(name, info) => info
}

def nameConf(x0: conf): String = x0 match {
  case Conf(name, info) => name
}

def emptyConf: conf = Conf(emptyStr, emptyStr)

def emptyPass: password = Password(emptyStr)

def emptyUser: user = User(emptyStr, emptyStr, emptyStr)

def isRevRole(x0: role): Boolean = x0 match {
  case Rev(uu, uv) => true
  case Aut(v) => false
  case PC() => false
  case Chair() => false
}

def emptyPaper: paper =
  Paper(emptyStr, emptyStr, NoPContent(), Nil, emptyDis, Nil, NoFPContent())

def isRevRoleFor(papIDa: paperID, x1: role): Boolean = (papIDa, x1) match {
  case (papIDa, Rev(papID, n)) => equal_paperIDa(papIDa, papID)
  case (papID, Aut(v)) => false
  case (papID, PC()) => false
  case (papID, Chair()) => false
}

def paperIDAsStr(x0: paperID): String = x0 match {
  case PaperID(str) => str
}

def reviewsPaper(x0: paper): List[List[(exp, (score, String))]] = x0 match {
  case Paper(title, abstracta, content, reviews, dis, decs, fcontent) => reviews
}

def voronkovUserID: userID = UserID("voronkov")

def getFreshPaperID(paperIDs: List[paperID]): paperID =
  PaperID(getFresh(Lista.map[paperID,
                              String](((a: paperID) => paperIDAsStr(a)),
                                       paperIDs),
                    "2"))

def equal_decision(x0: decision, x1: decision): Boolean = (x0, x1) match {
  case (Accept(), Reject()) => false
  case (Reject(), Accept()) => false
  case (NoDecision(), Reject()) => false
  case (Reject(), NoDecision()) => false
  case (NoDecision(), Accept()) => false
  case (Accept(), NoDecision()) => false
  case (Reject(), Reject()) => true
  case (Accept(), Accept()) => true
  case (NoDecision(), NoDecision()) => true
}

def equal_password(x0: password, x1: password): Boolean = (x0, x1) match {
  case (Password(x), Password(ya)) => x == ya
}

def equal_preference(x0: preference, x1: preference): Boolean = (x0, x1) match {
  case (WouldNot(), Conflict()) => false
  case (Conflict(), WouldNot()) => false
  case (Would(), Conflict()) => false
  case (Conflict(), Would()) => false
  case (Would(), WouldNot()) => false
  case (WouldNot(), Would()) => false
  case (Want(), Conflict()) => false
  case (Conflict(), Want()) => false
  case (Want(), WouldNot()) => false
  case (WouldNot(), Want()) => false
  case (Want(), Would()) => false
  case (Would(), Want()) => false
  case (NoPref(), Conflict()) => false
  case (Conflict(), NoPref()) => false
  case (NoPref(), WouldNot()) => false
  case (WouldNot(), NoPref()) => false
  case (NoPref(), Would()) => false
  case (Would(), NoPref()) => false
  case (NoPref(), Want()) => false
  case (Want(), NoPref()) => false
  case (Conflict(), Conflict()) => true
  case (WouldNot(), WouldNot()) => true
  case (Would(), Would()) => true
  case (Want(), Want()) => true
  case (NoPref(), NoPref()) => true
}

} /* object Prelim */

object Trivia {

def fun_upd2[A : HOL.equal, B : HOL.equal, C](f: A => B => C, a: A, b: B, c: C):
      A => B => C
  =
  ((aa: A) => (ba: B) =>
    (if (HOL.eq[A](a, aa) && HOL.eq[B](b, ba)) c else (f(aa))(ba)))

} /* object Trivia */

object System_Specification {

abstract sealed class uuAct
final case class
  uuNews(a: Prelim.confID, b: Prelim.userID, c: Prelim.password, d: String)
  extends uuAct
final case class
  uuDis(a: Prelim.confID, b: Prelim.userID, c: Prelim.password,
         d: Prelim.paperID, e: String)
  extends uuAct
final case class
  uuReview(a: Prelim.confID, b: Prelim.userID, c: Prelim.password,
            d: Prelim.paperID, e: Nat.nat,
            f: (Prelim.exp, (Prelim.score, String)))
  extends uuAct
final case class
  uuDec(a: Prelim.confID, b: Prelim.userID, c: Prelim.password,
         d: Prelim.paperID, e: Prelim.decision)
  extends uuAct

abstract sealed class uAct
final case class
  uUser(a: Prelim.userID, b: Prelim.password, c: Prelim.password, d: String,
         e: String, f: String)
  extends uAct
final case class uConfA(a: Prelim.confID, b: Prelim.userID, c: Prelim.password)
  extends uAct
final case class
  uPhase(a: Prelim.confID, b: Prelim.userID, c: Prelim.password, d: Nat.nat)
  extends uAct
final case class
  uPaperTA(a: Prelim.confID, b: Prelim.userID, c: Prelim.password,
            d: Prelim.paperID, e: String, f: String)
  extends uAct
final case class
  uPaperC(a: Prelim.confID, b: Prelim.userID, c: Prelim.password,
           d: Prelim.paperID, e: Prelim.pcontent)
  extends uAct
final case class
  uPref(a: Prelim.confID, b: Prelim.userID, c: Prelim.password,
         d: Prelim.paperID, e: Prelim.preference)
  extends uAct
final case class
  uReview(a: Prelim.confID, b: Prelim.userID, c: Prelim.password,
           d: Prelim.paperID, e: Nat.nat,
           f: (Prelim.exp, (Prelim.score, String)))
  extends uAct
final case class
  uFPaperC(a: Prelim.confID, b: Prelim.userID, c: Prelim.password,
            d: Prelim.paperID, e: Prelim.fpcontent)
  extends uAct

abstract sealed class rAct
final case class rAmIVoronkov(a: Prelim.userID, b: Prelim.password) extends rAct
final case class rUser(a: Prelim.userID, b: Prelim.password, c: Prelim.userID)
  extends rAct
final case class rConf(a: Prelim.confID, b: Prelim.userID, c: Prelim.password)
  extends rAct
final case class rNews(a: Prelim.confID, b: Prelim.userID, c: Prelim.password)
  extends rAct
final case class
  rPaperTAA(a: Prelim.confID, b: Prelim.userID, c: Prelim.password,
             d: Prelim.paperID)
  extends rAct
final case class
  rPaperC(a: Prelim.confID, b: Prelim.userID, c: Prelim.password,
           d: Prelim.paperID)
  extends rAct
final case class
  rPref(a: Prelim.confID, b: Prelim.userID, c: Prelim.password,
         d: Prelim.paperID)
  extends rAct
final case class
  rMyReview(a: Prelim.confID, b: Prelim.userID, c: Prelim.password,
             d: Prelim.paperID)
  extends rAct
final case class
  rReviews(a: Prelim.confID, b: Prelim.userID, c: Prelim.password,
            d: Prelim.paperID)
  extends rAct
final case class
  rDecs(a: Prelim.confID, b: Prelim.userID, c: Prelim.password,
         d: Prelim.paperID)
  extends rAct
final case class
  rDis(a: Prelim.confID, b: Prelim.userID, c: Prelim.password,
        d: Prelim.paperID)
  extends rAct
final case class
  rFinalReviews(a: Prelim.confID, b: Prelim.userID, c: Prelim.password,
                 d: Prelim.paperID)
  extends rAct
final case class
  rFinalDec(a: Prelim.confID, b: Prelim.userID, c: Prelim.password,
             d: Prelim.paperID)
  extends rAct
final case class
  rPrefOfPC(a: Prelim.confID, b: Prelim.userID, c: Prelim.password,
             d: Prelim.paperID, e: Prelim.userID)
  extends rAct
final case class
  rFPaperC(a: Prelim.confID, b: Prelim.userID, c: Prelim.password,
            d: Prelim.paperID)
  extends rAct

abstract sealed class lAct
final case class lConfs(a: Prelim.userID, b: Prelim.password) extends lAct
final case class lAConfs(a: Prelim.userID, b: Prelim.password) extends lAct
final case class lSConfs(a: Prelim.userID, b: Prelim.password) extends lAct
final case class lMyConfs(a: Prelim.userID, b: Prelim.password) extends lAct
final case class lAllUsers(a: Prelim.userID, b: Prelim.password) extends lAct
final case class lAllPapers(a: Prelim.userID, b: Prelim.password) extends lAct
final case class lPC(a: Prelim.confID, b: Prelim.userID, c: Prelim.password)
  extends lAct
final case class lChair(a: Prelim.confID, b: Prelim.userID, c: Prelim.password)
  extends lAct
final case class lPapers(a: Prelim.confID, b: Prelim.userID, c: Prelim.password)
  extends lAct
final case class
  lMyPapers(a: Prelim.confID, b: Prelim.userID, c: Prelim.password)
  extends lAct
final case class
  lMyAssignedPapers(a: Prelim.confID, b: Prelim.userID, c: Prelim.password)
  extends lAct
final case class
  lAssignedReviewers(a: Prelim.confID, b: Prelim.userID, c: Prelim.password,
                      d: Prelim.paperID)
  extends lAct

abstract sealed class cAct
final case class
  cUser(a: Prelim.userID, b: Prelim.password, c: String, d: String, e: String)
  extends cAct
final case class
  cConf(a: Prelim.confID, b: Prelim.userID, c: Prelim.password, d: String,
         e: String)
  extends cAct
final case class
  cPC(a: Prelim.confID, b: Prelim.userID, c: Prelim.password, d: Prelim.userID)
  extends cAct
final case class
  cChair(a: Prelim.confID, b: Prelim.userID, c: Prelim.password,
          d: Prelim.userID)
  extends cAct
final case class
  cPaper(a: Prelim.confID, b: Prelim.userID, c: Prelim.password,
          d: Prelim.paperID, e: String, f: String)
  extends cAct
final case class
  cAuthor(a: Prelim.confID, b: Prelim.userID, c: Prelim.password,
           d: Prelim.paperID, e: Prelim.userID)
  extends cAct
final case class
  cConflict(a: Prelim.confID, b: Prelim.userID, c: Prelim.password,
             d: Prelim.paperID, e: Prelim.userID)
  extends cAct
final case class
  cReview(a: Prelim.confID, b: Prelim.userID, c: Prelim.password,
           d: Prelim.paperID, e: Prelim.userID)
  extends cAct

abstract sealed class act
final case class Cact(a: cAct) extends act
final case class Uact(a: uAct) extends act
final case class UUact(a: uuAct) extends act
final case class Ract(a: rAct) extends act
final case class Lact(a: lAct) extends act

abstract sealed class out
final case class outOK() extends out
final case class outErr() extends out
final case class outBool(a: Boolean) extends out
final case class outSTRT(a: (String, (String, String))) extends out
final case class outSTRL(a: List[String]) extends out
final case class outCONF(a: (String, (String, (List[Prelim.role], Nat.nat))))
  extends out
final case class outPREF(a: Prelim.preference) extends out
final case class outCON(a: Prelim.pcontent) extends out
final case class outFCON(a: Prelim.fpcontent) extends out
final case class
  outNREV(a: (Nat.nat, List[(Prelim.exp, (Prelim.score, String))]))
  extends out
final case class outREVL(a: List[List[(Prelim.exp, (Prelim.score, String))]])
  extends out
final case class
  outRREVL(a: List[(Prelim.userID, List[(Prelim.exp, (Prelim.score, String))])])
  extends out
final case class outDEC(a: Prelim.decision) extends out
final case class outDECL(a: List[Prelim.decision]) extends out
final case class outCIDL(a: List[Prelim.confID]) extends out
final case class outUIDL(a: List[Prelim.userID]) extends out
final case class outPIDL(a: List[Prelim.paperID]) extends out
final case class outSTRPAL(a: (String, (String, List[Prelim.userID]))) extends
  out

abstract sealed class state_ext[A]
final case class
  state_exta[A](a: List[Prelim.confID], b: Prelim.confID => Prelim.conf,
                 c: List[Prelim.userID], d: Prelim.userID => Prelim.password,
                 e: Prelim.userID => Prelim.user,
                 f: Prelim.confID => Prelim.userID => List[Prelim.role],
                 g: Prelim.confID => List[Prelim.paperID],
                 h: Prelim.paperID => Prelim.paper,
                 i: Prelim.userID => Prelim.paperID => Prelim.preference,
                 j: Prelim.userID, k: Prelim.confID => List[String],
                 l: Prelim.confID => Nat.nat, m: A)
  extends state_ext[A]

def roles[A](x0: state_ext[A]):
      Prelim.confID => Prelim.userID => List[Prelim.role]
  =
  x0 match {
  case state_exta(confIDs, conf, userIDs, pass, user, roles, paperIDs, paper,
                   pref, voronkov, news, phase, more)
    => roles
}

def phase[A](x0: state_ext[A]): Prelim.confID => Nat.nat = x0 match {
  case state_exta(confIDs, conf, userIDs, pass, user, roles, paperIDs, paper,
                   pref, voronkov, news, phase, more)
    => phase
}

def pref[A](x0: state_ext[A]):
      Prelim.userID => Prelim.paperID => Prelim.preference
  =
  x0 match {
  case state_exta(confIDs, conf, userIDs, pass, user, roles, paperIDs, paper,
                   pref, voronkov, news, phase, more)
    => pref
}

def pass[A](x0: state_ext[A]): Prelim.userID => Prelim.password = x0 match {
  case state_exta(confIDs, conf, userIDs, pass, user, roles, paperIDs, paper,
                   pref, voronkov, news, phase, more)
    => pass
}

def paperIDs[A](x0: state_ext[A]): Prelim.confID => List[Prelim.paperID] = x0
  match {
  case state_exta(confIDs, conf, userIDs, pass, user, roles, paperIDs, paper,
                   pref, voronkov, news, phase, more)
    => paperIDs
}

def userIDs[A](x0: state_ext[A]): List[Prelim.userID] = x0 match {
  case state_exta(confIDs, conf, userIDs, pass, user, roles, paperIDs, paper,
                   pref, voronkov, news, phase, more)
    => userIDs
}

def confIDs[A](x0: state_ext[A]): List[Prelim.confID] = x0 match {
  case state_exta(confIDs, conf, userIDs, pass, user, roles, paperIDs, paper,
                   pref, voronkov, news, phase, more)
    => confIDs
}

def IDsOK(s: state_ext[Unit], cIDs: List[Prelim.confID],
           uIDs: List[Prelim.userID], papIDs: List[Prelim.paperID]):
      Boolean
  =
  Lista.list_all[Prelim.confID](((a: Prelim.confID) =>
                                  Lista.member[Prelim.confID](confIDs[Unit](s),
                       a)),
                                 cIDs) &&
    (Lista.list_all[Prelim.userID](((a: Prelim.userID) =>
                                     Lista.member[Prelim.userID](userIDs[Unit](s),
                          a)),
                                    uIDs) &&
      Lista.list_all[Prelim.paperID](((a: Prelim.paperID) =>
                                       Lista.member[Prelim.paperID]((paperIDs[Unit](s)).apply(Lista.hd[Prelim.confID](cIDs)),
                             a)),
                                      papIDs))

def e_listAssignedReviewers(s: state_ext[Unit], confID: Prelim.confID,
                             uID: Prelim.userID, p: Prelim.password,
                             papID: Prelim.paperID):
      Boolean
  =
  IDsOK(s, List(confID), List(uID), List(papID)) &&
    (Prelim.equal_password((pass[Unit](s)).apply(uID), p) &&
      (Nat.less_eq_nat(Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.zero_nat())))),
                        (phase[Unit](s)).apply(confID)) &&
        (Lista.member[Prelim.role]((roles[Unit](s)).apply(confID).apply(uID),
                                    Prelim.Chair()) &&
          ! (Prelim.equal_preference((pref[Unit](s)).apply(uID).apply(papID),
                                      Prelim.Conflict())))))

def e_listMyAssignedPapers(s: state_ext[Unit], confID: Prelim.confID,
                            uID: Prelim.userID, p: Prelim.password):
      Boolean
  =
  IDsOK(s, List(confID), List(uID), Nil) &&
    (Prelim.equal_password((pass[Unit](s)).apply(uID), p) &&
      (Nat.less_eq_nat(Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.zero_nat())))),
                        (phase[Unit](s)).apply(confID)) &&
        Lista.member[Prelim.role]((roles[Unit](s)).apply(confID).apply(uID),
                                   Prelim.PC())))

def isRev(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
           papID: Prelim.paperID):
      Boolean
  =
  Lista.list_ex[Prelim.role](((a: Prelim.role) =>
                               Prelim.isRevRoleFor(papID, a)),
                              (roles[Unit](s)).apply(confID).apply(uID))

def listAssignedReviewers(s: state_ext[Unit], confID: Prelim.confID,
                           uID: Prelim.userID, p: Prelim.password,
                           papID: Prelim.paperID):
      List[Prelim.userID]
  =
  Lista.filter[Prelim.userID](((uIDa: Prelim.userID) =>
                                isRev(s, confID, uIDa, papID)),
                               userIDs[Unit](s))

def paper[A](x0: state_ext[A]): Prelim.paperID => Prelim.paper = x0 match {
  case state_exta(confIDs, conf, userIDs, pass, user, roles, paperIDs, paper,
                   pref, voronkov, news, phase, more)
    => paper
}

def listMyAssignedPapers(s: state_ext[Unit], confID: Prelim.confID,
                          uID: Prelim.userID, p: Prelim.password):
      List[Prelim.paperID]
  =
  {
    paper[Unit](s);
    Lista.maps[Prelim.paperID,
                Prelim.paperID](((papID: Prelim.paperID) =>
                                  (if (isRev(s, confID, uID, papID)) List(papID)
                                    else Nil)),
                                 (paperIDs[Unit](s)).apply(confID))
  }

def e_listAllPapers(s: state_ext[Unit], uID: Prelim.userID, p: Prelim.password):
      Boolean
  =
  IDsOK(s, Nil, List(uID), Nil) &&
    Prelim.equal_password((pass[Unit](s)).apply(uID), p)

def e_listMyPapers(s: state_ext[Unit], confID: Prelim.confID,
                    uID: Prelim.userID, p: Prelim.password):
      Boolean
  =
  IDsOK(s, List(confID), List(uID), Nil) &&
    (Prelim.equal_password((pass[Unit](s)).apply(uID), p) &&
      Nat.less_eq_nat(Nat.Suc(Nat.Suc(Nat.zero_nat())),
                       (phase[Unit](s)).apply(confID)))

def e_listAllUsers(s: state_ext[Unit], uID: Prelim.userID, p: Prelim.password):
      Boolean
  =
  IDsOK(s, Nil, List(uID), Nil) &&
    Prelim.equal_password((pass[Unit](s)).apply(uID), p)

def getAllPaperIDs(s: state_ext[Unit]): List[Prelim.paperID] =
  Lista.maps[Prelim.confID, Prelim.paperID](paperIDs[Unit](s), confIDs[Unit](s))

def listAllPapers(s: state_ext[Unit], uID: Prelim.userID, p: Prelim.password):
      List[Prelim.paperID]
  =
  getAllPaperIDs(s)

def e_listMyConfs(s: state_ext[Unit], uID: Prelim.userID, p: Prelim.password):
      Boolean
  =
  IDsOK(s, Nil, List(uID), Nil) &&
    Prelim.equal_password((pass[Unit](s)).apply(uID), p)

def listMyPapers(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
                  p: Prelim.password):
      List[Prelim.paperID]
  =
  {
    paper[Unit](s);
    Lista.maps[Prelim.paperID,
                Prelim.paperID](((papID: Prelim.paperID) =>
                                  (if (Lista.member[Prelim.role]((roles[Unit](s)).apply(confID).apply(uID),
                          Prelim.Aut(papID)))
                                    List(papID) else Nil)),
                                 (paperIDs[Unit](s)).apply(confID))
  }

def listAllUsers(s: state_ext[Unit], uID: Prelim.userID, p: Prelim.password):
      List[Prelim.userID]
  =
  userIDs[Unit](s)

def e_listSConfs(s: state_ext[Unit], uID: Prelim.userID, p: Prelim.password):
      Boolean
  =
  IDsOK(s, Nil, List(uID), Nil) &&
    Prelim.equal_password((pass[Unit](s)).apply(uID), p)

def e_listPapers(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
                  p: Prelim.password):
      Boolean
  =
  IDsOK(s, List(confID), List(uID), Nil) &&
    (Prelim.equal_password((pass[Unit](s)).apply(uID), p) &&
      (Nat.less_eq_nat(Nat.Suc(Nat.Suc(Nat.zero_nat())),
                        (phase[Unit](s)).apply(confID)) &&
        Lista.member[Prelim.role]((roles[Unit](s)).apply(confID).apply(uID),
                                   Prelim.PC())))

def voronkov[A](x0: state_ext[A]): Prelim.userID = x0 match {
  case state_exta(confIDs, conf, userIDs, pass, user, roles, paperIDs, paper,
                   pref, voronkov, news, phase, more)
    => voronkov
}

def e_listAConfs(s: state_ext[Unit], uID: Prelim.userID, p: Prelim.password):
      Boolean
  =
  IDsOK(s, Nil, List(uID), Nil) &&
    (Prelim.equal_password((pass[Unit](s)).apply(uID), p) &&
      Prelim.equal_userIDa(uID, voronkov[Unit](s)))

def listMyConfs(s: state_ext[Unit], uID: Prelim.userID, p: Prelim.password):
      List[Prelim.confID]
  =
  Lista.maps[Prelim.confID,
              Prelim.confID](((confID: Prelim.confID) =>
                               (if (! (Lista.nulla[Prelim.role]((roles[Unit](s)).apply(confID).apply(uID))))
                                 List(confID) else Nil)),
                              confIDs[Unit](s))

def e_listConfs(s: state_ext[Unit], uID: Prelim.userID, p: Prelim.password):
      Boolean
  =
  IDsOK(s, Nil, List(uID), Nil) &&
    (Prelim.equal_password((pass[Unit](s)).apply(uID), p) &&
      Prelim.equal_userIDa(uID, voronkov[Unit](s)))

def e_listChair(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
                 p: Prelim.password):
      Boolean
  =
  IDsOK(s, List(confID), List(uID), Nil) &&
    (Prelim.equal_password((pass[Unit](s)).apply(uID), p) &&
      (Nat.less_eq_nat(Nat.Suc(Nat.Suc(Nat.zero_nat())),
                        (phase[Unit](s)).apply(confID)) ||
        Nat.less_eq_nat(Nat.Suc(Nat.zero_nat()),
                         (phase[Unit](s)).apply(confID)) &&
          Lista.member[Prelim.role]((roles[Unit](s)).apply(confID).apply(uID),
                                     Prelim.Chair())))

def listSConfs(s: state_ext[Unit], uID: Prelim.userID, p: Prelim.password):
      List[Prelim.confID]
  =
  Lista.maps[Prelim.confID,
              Prelim.confID](((confID: Prelim.confID) =>
                               (if (Nat.equal_nat((phase[Unit](s)).apply(confID),
           Nat.Suc(Nat.Suc(Nat.zero_nat()))))
                                 List(confID) else Nil)),
                              confIDs[Unit](s))

def listPapers(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
                p: Prelim.password):
      List[Prelim.paperID]
  =
  {
    paper[Unit](s);
    Lista.map[Prelim.paperID,
               Prelim.paperID](((papID: Prelim.paperID) => papID),
                                (paperIDs[Unit](s)).apply(confID))
  }

def listAConfs(s: state_ext[Unit], uID: Prelim.userID, p: Prelim.password):
      List[Prelim.confID]
  =
  Lista.maps[Prelim.confID,
              Prelim.confID](((confID: Prelim.confID) =>
                               (if (Nat.equal_nat((phase[Unit](s)).apply(confID),
           Nat.zero_nat()))
                                 List(confID) else Nil)),
                              confIDs[Unit](s))

def listConfs(s: state_ext[Unit], uID: Prelim.userID, p: Prelim.password):
      List[Prelim.confID]
  =
  confIDs[Unit](s)

def listChair(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
               p: Prelim.password):
      List[Prelim.userID]
  =
  Lista.maps[Prelim.userID,
              Prelim.userID](((uIDa: Prelim.userID) =>
                               (if (Lista.member[Prelim.role]((roles[Unit](s)).apply(confID).apply(uIDa),
                       Prelim.Chair()))
                                 List(uIDa) else Nil)),
                              userIDs[Unit](s))

def e_listPC(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
              p: Prelim.password):
      Boolean
  =
  IDsOK(s, List(confID), List(uID), Nil) &&
    (Prelim.equal_password((pass[Unit](s)).apply(uID), p) &&
      (Nat.less_eq_nat(Nat.Suc(Nat.Suc(Nat.zero_nat())),
                        (phase[Unit](s)).apply(confID)) ||
        Nat.less_eq_nat(Nat.Suc(Nat.zero_nat()),
                         (phase[Unit](s)).apply(confID)) &&
          Lista.member[Prelim.role]((roles[Unit](s)).apply(confID).apply(uID),
                                     Prelim.Chair())))

def listPC(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
            p: Prelim.password):
      List[Prelim.userID]
  =
  Lista.maps[Prelim.userID,
              Prelim.userID](((uIDa: Prelim.userID) =>
                               (if (Lista.member[Prelim.role]((roles[Unit](s)).apply(confID).apply(uIDa),
                       Prelim.PC()))
                                 List(uIDa) else Nil)),
                              userIDs[Unit](s))

def lObs(s: state_ext[Unit], x1: lAct): out = (s, x1) match {
  case (s, lConfs(uID, p)) =>
    (if (e_listConfs(s, uID, p)) outCIDL(listConfs(s, uID, p)) else outErr())
  case (s, lAConfs(uID, p)) =>
    (if (e_listAConfs(s, uID, p)) outCIDL(listAConfs(s, uID, p)) else outErr())
  case (s, lSConfs(uID, p)) =>
    (if (e_listSConfs(s, uID, p)) outCIDL(listSConfs(s, uID, p)) else outErr())
  case (s, lMyConfs(uID, p)) =>
    (if (e_listMyConfs(s, uID, p)) outCIDL(listMyConfs(s, uID, p))
      else outErr())
  case (s, lAllUsers(uID, p)) =>
    (if (e_listAllUsers(s, uID, p)) outUIDL(listAllUsers(s, uID, p))
      else outErr())
  case (s, lAllPapers(uID, p)) =>
    (if (e_listAllPapers(s, uID, p)) outPIDL(listAllPapers(s, uID, p))
      else outErr())
  case (s, lPC(confID, uID, p)) =>
    (if (e_listPC(s, confID, uID, p)) outUIDL(listPC(s, confID, uID, p))
      else outErr())
  case (s, lChair(confID, uID, p)) =>
    (if (e_listChair(s, confID, uID, p)) outUIDL(listChair(s, confID, uID, p))
      else outErr())
  case (s, lPapers(confID, uID, p)) =>
    (if (e_listPapers(s, confID, uID, p)) outPIDL(listPapers(s, confID, uID, p))
      else outErr())
  case (s, lMyPapers(confID, uID, p)) =>
    (if (e_listMyPapers(s, confID, uID, p))
      outPIDL(listMyPapers(s, confID, uID, p)) else outErr())
  case (s, lMyAssignedPapers(confID, uID, p)) =>
    (if (e_listMyAssignedPapers(s, confID, uID, p))
      outPIDL(listMyAssignedPapers(s, confID, uID, p)) else outErr())
  case (s, lAssignedReviewers(confID, uID, p, papID)) =>
    (if (e_listAssignedReviewers(s, confID, uID, p, papID))
      outUIDL(listAssignedReviewers(s, confID, uID, p, papID)) else outErr())
}

def e_readFinalReviews(s: state_ext[Unit], confID: Prelim.confID,
                        uID: Prelim.userID, p: Prelim.password,
                        papID: Prelim.paperID):
      Boolean
  =
  IDsOK(s, List(confID), List(uID), List(papID)) &&
    (Prelim.equal_password((pass[Unit](s)).apply(uID), p) &&
      (Nat.less_eq_nat(Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.zero_nat())))))),
                        (phase[Unit](s)).apply(confID)) &&
        (Lista.member[Prelim.role]((roles[Unit](s)).apply(confID).apply(uID),
                                    Prelim.Aut(papID)) ||
          Lista.member[Prelim.role]((roles[Unit](s)).apply(confID).apply(uID),
                                     Prelim.PC()) &&
            ! (Prelim.equal_preference((pref[Unit](s)).apply(uID).apply(papID),
Prelim.Conflict())))))

def e_readAmIVoronkov(s: state_ext[Unit], uID: Prelim.userID,
                       p: Prelim.password):
      Boolean
  =
  IDsOK(s, Nil, List(uID), Nil) &&
    Prelim.equal_password((pass[Unit](s)).apply(uID), p)

def readFinalReviews(s: state_ext[Unit], confID: Prelim.confID,
                      uID: Prelim.userID, p: Prelim.password,
                      papID: Prelim.paperID):
      List[List[(Prelim.exp, (Prelim.score, String))]]
  =
  Lista.map[List[(Prelim.exp, (Prelim.score, String))],
             List[(Prelim.exp,
                    (Prelim.score,
                      String))]](((a: List[(Prelim.exp,
     (Prelim.score, String))])
                                    =>
                                   (a match {
                                      case Nil =>
List((Prelim.NoExp(), (Prelim.NoScore(), Prelim.emptyStr)))
                                      case (exp, (score, comm)) :: _ =>
List((exp, (score, comm)))
                                    })),
                                  Prelim.reviewsPaper((paper[Unit](s)).apply(papID)))

def readAmIVoronkov(s: state_ext[Unit], uID: Prelim.userID, p: Prelim.password):
      Boolean
  =
  Prelim.equal_userIDa(uID, voronkov[Unit](s))

def e_readPrefOfPC(s: state_ext[Unit], confID: Prelim.confID,
                    uIDa: Prelim.userID, p: Prelim.password,
                    papID: Prelim.paperID, uID: Prelim.userID):
      Boolean
  =
  IDsOK(s, List(confID), List(uIDa, uID), List(papID)) &&
    (Prelim.equal_password((pass[Unit](s)).apply(uIDa), p) &&
      (Nat.less_eq_nat(Nat.Suc(Nat.Suc(Nat.Suc(Nat.zero_nat()))),
                        (phase[Unit](s)).apply(confID)) &&
         (Lista.member[Prelim.role]((roles[Unit](s)).apply(confID).apply(uIDa),
                                     Prelim.Chair()) &&
           Lista.member[Prelim.role]((roles[Unit](s)).apply(confID).apply(uID),
                                      Prelim.PC())) ||
        Nat.equal_nat((phase[Unit](s)).apply(confID),
                       Nat.Suc(Nat.Suc(Nat.zero_nat()))) &&
          Lista.member[Prelim.role]((roles[Unit](s)).apply(confID).apply(uIDa),
                                     Prelim.Aut(papID))))

def e_readPaperTAA(s: state_ext[Unit], confID: Prelim.confID,
                    uID: Prelim.userID, p: Prelim.password,
                    papID: Prelim.paperID):
      Boolean
  =
  IDsOK(s, List(confID), List(uID), List(papID)) &&
    (Prelim.equal_password((pass[Unit](s)).apply(uID), p) &&
      (Nat.less_eq_nat(Nat.Suc(Nat.Suc(Nat.zero_nat())),
                        (phase[Unit](s)).apply(confID)) &&
        (Lista.member[Prelim.role]((roles[Unit](s)).apply(confID).apply(uID),
                                    Prelim.Aut(papID)) ||
          Lista.member[Prelim.role]((roles[Unit](s)).apply(confID).apply(uID),
                                     Prelim.PC()))))

def e_readMyReview(s: state_ext[Unit], confID: Prelim.confID,
                    uID: Prelim.userID, p: Prelim.password,
                    papID: Prelim.paperID):
      Boolean
  =
  IDsOK(s, List(confID), List(uID), List(papID)) &&
    (Prelim.equal_password((pass[Unit](s)).apply(uID), p) &&
      (Nat.less_eq_nat(Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.zero_nat())))),
                        (phase[Unit](s)).apply(confID)) &&
        isRev(s, confID, uID, papID)))

def e_readFinalDec(s: state_ext[Unit], confID: Prelim.confID,
                    uID: Prelim.userID, p: Prelim.password,
                    papID: Prelim.paperID):
      Boolean
  =
  IDsOK(s, List(confID), List(uID), List(papID)) &&
    (Prelim.equal_password((pass[Unit](s)).apply(uID), p) &&
      (Nat.less_eq_nat(Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.zero_nat())))))),
                        (phase[Unit](s)).apply(confID)) &&
        (Lista.member[Prelim.role]((roles[Unit](s)).apply(confID).apply(uID),
                                    Prelim.Aut(papID)) ||
          Lista.member[Prelim.role]((roles[Unit](s)).apply(confID).apply(uID),
                                     Prelim.PC()))))

def e_readReviews(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
                   p: Prelim.password, papID: Prelim.paperID):
      Boolean
  =
  IDsOK(s, List(confID), List(uID), List(papID)) &&
    (Prelim.equal_password((pass[Unit](s)).apply(uID), p) &&
      (Nat.less_eq_nat(Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.zero_nat()))))),
                        (phase[Unit](s)).apply(confID)) &&
        (Lista.member[Prelim.role]((roles[Unit](s)).apply(confID).apply(uID),
                                    Prelim.PC()) &&
          ! (Prelim.equal_preference((pref[Unit](s)).apply(uID).apply(papID),
                                      Prelim.Conflict())))))

def e_readFPaperC(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
                   p: Prelim.password, papID: Prelim.paperID):
      Boolean
  =
  IDsOK(s, List(confID), List(uID), List(papID)) &&
    (Prelim.equal_password((pass[Unit](s)).apply(uID), p) &&
      (Nat.less_eq_nat(Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.zero_nat())))))),
                        (phase[Unit](s)).apply(confID)) &&
        ((Lista.member[Prelim.role]((roles[Unit](s)).apply(confID).apply(uID),
                                     Prelim.Aut(papID)) ||
           Lista.member[Prelim.role]((roles[Unit](s)).apply(confID).apply(uID),
                                      Prelim.Chair())) &&
          Prelim.equal_decision(Prelim.decPaper((paper[Unit](s)).apply(papID)),
                                 Prelim.Accept()))))

def readPrefOfPC(s: state_ext[Unit], confID: Prelim.confID, uIDa: Prelim.userID,
                  p: Prelim.password, papID: Prelim.paperID,
                  uID: Prelim.userID):
      Prelim.preference
  =
  (pref[Unit](s)).apply(uID).apply(papID)

def readPaperTAA(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
                  p: Prelim.password, papID: Prelim.paperID):
      (String, (String, List[Prelim.userID]))
  =
  {
    val (Prelim.Paper(title, abstracta, _, _, _, _, _)): Prelim.paper =
      (paper[Unit](s)).apply(papID);
    (title,
      (abstracta,
        Lista.maps[Prelim.userID,
                    Prelim.userID](((uIDa: Prelim.userID) =>
                                     (if (Lista.member[Prelim.role]((roles[Unit](s)).apply(confID).apply(uIDa),
                             Prelim.Aut(papID)))
                                       List(uIDa) else Nil)),
                                    userIDs[Unit](s))))
  }

def getNthReview(s: state_ext[Unit], papID: Prelim.paperID, n: Nat.nat):
      List[(Prelim.exp, (Prelim.score, String))]
  =
  Lista.nth[List[(Prelim.exp,
                   (Prelim.score,
                     String))]](Prelim.reviewsPaper((paper[Unit](s)).apply(papID)),
                                 n)

def getRevRole(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
                papID: Prelim.paperID):
      Option[Prelim.role]
  =
  Lista.find[Prelim.role](((a: Prelim.role) => Prelim.isRevRoleFor(papID, a)),
                           (roles[Unit](s)).apply(confID).apply(uID))

def readMyReview(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
                  p: Prelim.password, papID: Prelim.paperID):
      (Nat.nat, List[(Prelim.exp, (Prelim.score, String))])
  =
  {
    val (Some(Prelim.Rev(_, n))): Option[Prelim.role] =
      getRevRole(s, confID, uID, papID);
    (n, getNthReview(s, papID, n))
  }

def readFinalDec(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
                  p: Prelim.password, papID: Prelim.paperID):
      Prelim.decision
  =
  ((paper[Unit](s)).apply(papID) match {
     case Prelim.Paper(_, _, _, _, _, Nil, _) => Prelim.NoDecision()
     case Prelim.Paper(_, _, _, _, _, dec :: _, _) => dec
   })

def e_readPaperC(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
                  p: Prelim.password, papID: Prelim.paperID):
      Boolean
  =
  IDsOK(s, List(confID), List(uID), List(papID)) &&
    (Prelim.equal_password((pass[Unit](s)).apply(uID), p) &&
      (Nat.less_eq_nat(Nat.Suc(Nat.Suc(Nat.zero_nat())),
                        (phase[Unit](s)).apply(confID)) &&
         Lista.member[Prelim.role]((roles[Unit](s)).apply(confID).apply(uID),
                                    Prelim.Aut(papID)) ||
        Nat.less_eq_nat(Nat.Suc(Nat.Suc(Nat.Suc(Nat.zero_nat()))),
                         (phase[Unit](s)).apply(confID)) &&
          Lista.member[Prelim.role]((roles[Unit](s)).apply(confID).apply(uID),
                                     Prelim.PC())))

def getReviewIndex(s: state_ext[Unit], confID: Prelim.confID,
                    uID: Prelim.userID, papID: Prelim.paperID):
      Nat.nat
  =
  {
    val (Some(Prelim.Rev(_, n))): Option[Prelim.role] =
      getRevRole(s, confID, uID, papID);
    n
  }

def getReviewersReviews(s: state_ext[Unit], confID: Prelim.confID,
                         papID: Prelim.paperID):
      List[(Prelim.userID, List[(Prelim.exp, (Prelim.score, String))])]
  =
  Lista.maps[Prelim.userID,
              (Prelim.userID,
                List[(Prelim.exp,
                       (Prelim.score,
                         String))])](((uID: Prelim.userID) =>
                                       (if (isRev(s, confID, uID, papID))
 List((uID, getNthReview(s, papID, getReviewIndex(s, confID, uID, papID))))
 else Nil)),
                                      userIDs[Unit](s))

def readReviews(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
                 p: Prelim.password, papID: Prelim.paperID):
      List[(Prelim.userID, List[(Prelim.exp, (Prelim.score, String))])]
  =
  getReviewersReviews(s, confID, papID)

def readFPaperC(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
                 p: Prelim.password, papID: Prelim.paperID):
      Prelim.fpcontent
  =
  {
    val (Prelim.Paper(_, _, _, _, _, _, fpc)): Prelim.paper =
      (paper[Unit](s)).apply(papID);
    fpc
  }

def readPaperC(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
                p: Prelim.password, papID: Prelim.paperID):
      Prelim.pcontent
  =
  {
    val (Prelim.Paper(_, _, pc, _, _, _, _)): Prelim.paper =
      (paper[Unit](s)).apply(papID);
    pc
  }

def e_readUser(s: state_ext[Unit], uIDa: Prelim.userID, p: Prelim.password,
                uID: Prelim.userID):
      Boolean
  =
  IDsOK(s, Nil, List(uIDa, uID), Nil) &&
    Prelim.equal_password((pass[Unit](s)).apply(uIDa), p)

def e_readPref(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
                p: Prelim.password, papID: Prelim.paperID):
      Boolean
  =
  IDsOK(s, List(confID), List(uID), List(papID)) &&
    (Prelim.equal_password((pass[Unit](s)).apply(uID), p) &&
      (Nat.less_eq_nat(Nat.Suc(Nat.Suc(Nat.Suc(Nat.zero_nat()))),
                        (phase[Unit](s)).apply(confID)) &&
        Lista.member[Prelim.role]((roles[Unit](s)).apply(confID).apply(uID),
                                   Prelim.PC())))

def e_readNews(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
                p: Prelim.password):
      Boolean
  =
  IDsOK(s, List(confID), List(uID), Nil) &&
    (Prelim.equal_password((pass[Unit](s)).apply(uID), p) &&
      (Nat.less_eq_nat(Nat.Suc(Nat.zero_nat()),
                        (phase[Unit](s)).apply(confID)) &&
        Lista.member[Prelim.role]((roles[Unit](s)).apply(confID).apply(uID),
                                   Prelim.PC())))

def e_readDecs(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
                p: Prelim.password, papID: Prelim.paperID):
      Boolean
  =
  IDsOK(s, List(confID), List(uID), List(papID)) &&
    (Prelim.equal_password((pass[Unit](s)).apply(uID), p) &&
      (Nat.less_eq_nat(Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.zero_nat()))))),
                        (phase[Unit](s)).apply(confID)) &&
        (Lista.member[Prelim.role]((roles[Unit](s)).apply(confID).apply(uID),
                                    Prelim.PC()) &&
          ! (Prelim.equal_preference((pref[Unit](s)).apply(uID).apply(papID),
                                      Prelim.Conflict())))))

def e_readConf(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
                p: Prelim.password):
      Boolean
  =
  IDsOK(s, List(confID), List(uID), Nil) &&
    Prelim.equal_password((pass[Unit](s)).apply(uID), p)

def e_readDis(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
               p: Prelim.password, papID: Prelim.paperID):
      Boolean
  =
  IDsOK(s, List(confID), List(uID), List(papID)) &&
    (Prelim.equal_password((pass[Unit](s)).apply(uID), p) &&
      (Nat.less_eq_nat(Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.zero_nat()))))),
                        (phase[Unit](s)).apply(confID)) &&
        (Lista.member[Prelim.role]((roles[Unit](s)).apply(confID).apply(uID),
                                    Prelim.PC()) &&
          ! (Prelim.equal_preference((pref[Unit](s)).apply(uID).apply(papID),
                                      Prelim.Conflict())))))

def user[A](x0: state_ext[A]): Prelim.userID => Prelim.user = x0 match {
  case state_exta(confIDs, conf, userIDs, pass, user, roles, paperIDs, paper,
                   pref, voronkov, news, phase, more)
    => user
}

def readUser(s: state_ext[Unit], uIDa: Prelim.userID, p: Prelim.password,
              uID: Prelim.userID):
      (String, (String, String))
  =
  {
    val (Prelim.User(name, info, email)): Prelim.user =
      (user[Unit](s)).apply(uID);
    (name, (info, email))
  }

def readPref(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
              p: Prelim.password, papID: Prelim.paperID):
      Prelim.preference
  =
  (pref[Unit](s)).apply(uID).apply(papID)

def news[A](x0: state_ext[A]): Prelim.confID => List[String] = x0 match {
  case state_exta(confIDs, conf, userIDs, pass, user, roles, paperIDs, paper,
                   pref, voronkov, news, phase, more)
    => news
}

def readNews(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
              p: Prelim.password):
      List[String]
  =
  (news[Unit](s)).apply(confID)

def readDecs(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
              p: Prelim.password, papID: Prelim.paperID):
      List[Prelim.decision]
  =
  {
    val (Prelim.Paper(_, _, _, _, _, decs, _)): Prelim.paper =
      (paper[Unit](s)).apply(papID);
    decs
  }

def conf[A](x0: state_ext[A]): Prelim.confID => Prelim.conf = x0 match {
  case state_exta(confIDs, conf, userIDs, pass, user, roles, paperIDs, paper,
                   pref, voronkov, news, phase, more)
    => conf
}

def readConf(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
              p: Prelim.password):
      (String, (String, (List[Prelim.role], Nat.nat)))
  =
  (Prelim.nameConf((conf[Unit](s)).apply(confID)),
    (Prelim.infoConf((conf[Unit](s)).apply(confID)),
      (Lista.filter[Prelim.role](((rl: Prelim.role) =>
                                   ! (Prelim.isRevRole(rl))),
                                  (roles[Unit](s)).apply(confID).apply(uID)),
        (phase[Unit](s)).apply(confID))))

def readDis(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
             p: Prelim.password, papID: Prelim.paperID):
      List[String]
  =
  {
    val (Prelim.Paper(_, _, _, _, Prelim.Dis(comments), _, _)): Prelim.paper =
      (paper[Unit](s)).apply(papID);
    comments
  }

def rObs(s: state_ext[Unit], x1: rAct): out = (s, x1) match {
  case (s, rAmIVoronkov(uID, p)) =>
    (if (e_readAmIVoronkov(s, uID, p)) outBool(readAmIVoronkov(s, uID, p))
      else outErr())
  case (s, rUser(uIDa, p, uID)) =>
    (if (e_readUser(s, uIDa, p, uID)) outSTRT(readUser(s, uIDa, p, uID))
      else outErr())
  case (s, rConf(confID, uID, p)) =>
    (if (e_readConf(s, confID, uID, p)) outCONF(readConf(s, confID, uID, p))
      else outErr())
  case (s, rNews(confID, uID, p)) =>
    (if (e_readNews(s, confID, uID, p)) outSTRL(readNews(s, confID, uID, p))
      else outErr())
  case (s, rPaperTAA(confID, uID, p, papID)) =>
    (if (e_readPaperTAA(s, confID, uID, p, papID))
      outSTRPAL(readPaperTAA(s, confID, uID, p, papID)) else outErr())
  case (s, rPaperC(confID, uID, p, papID)) =>
    (if (e_readPaperC(s, confID, uID, p, papID))
      outCON(readPaperC(s, confID, uID, p, papID)) else outErr())
  case (s, rPref(confID, uID, p, papID)) =>
    (if (e_readPref(s, confID, uID, p, papID))
      outPREF(readPref(s, confID, uID, p, papID)) else outErr())
  case (s, rMyReview(confID, uID, p, papID)) =>
    (if (e_readMyReview(s, confID, uID, p, papID))
      outNREV(readMyReview(s, confID, uID, p, papID)) else outErr())
  case (s, rReviews(confID, uID, p, papID)) =>
    (if (e_readReviews(s, confID, uID, p, papID))
      outRREVL(readReviews(s, confID, uID, p, papID)) else outErr())
  case (s, rDecs(confID, uID, p, papID)) =>
    (if (e_readDecs(s, confID, uID, p, papID))
      outDECL(readDecs(s, confID, uID, p, papID)) else outErr())
  case (s, rDis(confID, uID, p, papID)) =>
    (if (e_readDis(s, confID, uID, p, papID))
      outSTRL(readDis(s, confID, uID, p, papID)) else outErr())
  case (s, rFinalReviews(confID, uID, p, papID)) =>
    (if (e_readFinalReviews(s, confID, uID, p, papID))
      outREVL(readFinalReviews(s, confID, uID, p, papID)) else outErr())
  case (s, rFinalDec(confID, uID, p, papID)) =>
    (if (e_readFinalDec(s, confID, uID, p, papID))
      outDEC(readFinalDec(s, confID, uID, p, papID)) else outErr())
  case (s, rPrefOfPC(confID, uIDa, p, papID, uID)) =>
    (if (e_readPrefOfPC(s, confID, uIDa, p, papID, uID))
      outPREF(readPrefOfPC(s, confID, uIDa, p, papID, uID)) else outErr())
  case (s, rFPaperC(confID, uID, p, papID)) =>
    (if (e_readFPaperC(s, confID, uID, p, papID))
      outFCON(readFPaperC(s, confID, uID, p, papID)) else outErr())
}

def e_uupdateReview(s: state_ext[Unit], confID: Prelim.confID,
                     uID: Prelim.userID, p: Prelim.password,
                     papID: Prelim.paperID, n: Nat.nat,
                     rc: (Prelim.exp, (Prelim.score, String))):
      Boolean
  =
  IDsOK(s, List(confID), List(uID), List(papID)) &&
    (Prelim.equal_password((pass[Unit](s)).apply(uID), p) &&
      (Nat.equal_nat((phase[Unit](s)).apply(confID),
                      Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.zero_nat())))))) &&
        (isRev(s, confID, uID, papID) &&
          Nat.equal_nat(getReviewIndex(s, confID, uID, papID), n))))

def paper_update[A](papera:
                      (Prelim.paperID => Prelim.paper) =>
                        Prelim.paperID => Prelim.paper,
                     x1: state_ext[A]):
      state_ext[A]
  =
  (papera, x1) match {
  case (papera,
         state_exta(confIDs, conf, userIDs, pass, user, roles, paperIDs, paper,
                     pref, voronkov, news, phase, more))
    => state_exta[A](confIDs, conf, userIDs, pass, user, roles, paperIDs,
                      papera(paper), pref, voronkov, news, phase, more)
}

def uupdateReview(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
                   p: Prelim.password, papID: Prelim.paperID, n: Nat.nat,
                   rc: (Prelim.exp, (Prelim.score, String))):
      state_ext[Unit]
  =
  {
    val (Prelim.Paper(title, abstracta, pc, reviews, dis, decs, fpc)):
          Prelim.paper
      = (paper[Unit](s)).apply(papID)
    val review: List[(Prelim.exp, (Prelim.score, String))] =
      rc :: Lista.nth[List[(Prelim.exp, (Prelim.score, String))]](reviews, n)
    val reviewsa: List[List[(Prelim.exp, (Prelim.score, String))]] =
      Lista.list_update[List[(Prelim.exp,
                               (Prelim.score, String))]](reviews, n, review);
    paper_update[Unit](((_: Prelim.paperID => Prelim.paper) =>
                         Fun.fun_upd[Prelim.paperID,
                                      Prelim.paper](paper[Unit](s), papID,
             Prelim.Paper(title, abstracta, pc, reviewsa, dis, decs, fpc))),
                        s)
  }

def e_uupdateNews(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
                   p: Prelim.password, ev: String):
      Boolean
  =
  IDsOK(s, List(confID), List(uID), Nil) &&
    (Prelim.equal_password((pass[Unit](s)).apply(uID), p) &&
      (Nat.less_eq_nat(Nat.Suc(Nat.zero_nat()),
                        (phase[Unit](s)).apply(confID)) &&
        (Nat.less_nat((phase[Unit](s)).apply(confID),
                       Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.zero_nat())))))))) &&
          Lista.member[Prelim.role]((roles[Unit](s)).apply(confID).apply(uID),
                                     Prelim.Chair()))))

def e_uupdateDis(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
                  p: Prelim.password, papID: Prelim.paperID, comm: String):
      Boolean
  =
  IDsOK(s, List(confID), List(uID), List(papID)) &&
    (Prelim.equal_password((pass[Unit](s)).apply(uID), p) &&
      (Nat.equal_nat((phase[Unit](s)).apply(confID),
                      Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.zero_nat())))))) &&
        (Lista.member[Prelim.role]((roles[Unit](s)).apply(confID).apply(uID),
                                    Prelim.PC()) &&
          ! (Prelim.equal_preference((pref[Unit](s)).apply(uID).apply(papID),
                                      Prelim.Conflict())))))

def e_uupdateDec(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
                  p: Prelim.password, papID: Prelim.paperID,
                  dec: Prelim.decision):
      Boolean
  =
  IDsOK(s, List(confID), List(uID), List(papID)) &&
    (Prelim.equal_password((pass[Unit](s)).apply(uID), p) &&
      (Nat.equal_nat((phase[Unit](s)).apply(confID),
                      Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.zero_nat())))))) &&
        (Lista.member[Prelim.role]((roles[Unit](s)).apply(confID).apply(uID),
                                    Prelim.Chair()) &&
          ! (Prelim.equal_preference((pref[Unit](s)).apply(uID).apply(papID),
                                      Prelim.Conflict())))))

def news_update[A](newsa:
                     (Prelim.confID => List[String]) =>
                       Prelim.confID => List[String],
                    x1: state_ext[A]):
      state_ext[A]
  =
  (newsa, x1) match {
  case (newsa,
         state_exta(confIDs, conf, userIDs, pass, user, roles, paperIDs, paper,
                     pref, voronkov, news, phase, more))
    => state_exta[A](confIDs, conf, userIDs, pass, user, roles, paperIDs, paper,
                      pref, voronkov, newsa(news), phase, more)
}

def uupdateNews(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
                 p: Prelim.password, ev: String):
      state_ext[Unit]
  =
  {
    val evs: List[String] = (news[Unit](s)).apply(confID);
    news_update[Unit](((_: Prelim.confID => List[String]) =>
                        Fun.fun_upd[Prelim.confID,
                                     List[String]](news[Unit](s), confID,
            ev :: evs)),
                       s)
  }

def uupdateDis(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
                p: Prelim.password, papID: Prelim.paperID, comm: String):
      state_ext[Unit]
  =
  {
    val (Prelim.Paper(title, abstracta, pc, reviews, Prelim.Dis(comments), decs,
                       fpc)):
          Prelim.paper
      = (paper[Unit](s)).apply(papID);
    paper_update[Unit](((_: Prelim.paperID => Prelim.paper) =>
                         Fun.fun_upd[Prelim.paperID,
                                      Prelim.paper](paper[Unit](s), papID,
             Prelim.Paper(title, abstracta, pc, reviews,
                           Prelim.Dis(comm :: comments), decs, fpc))),
                        s)
  }

def uupdateDec(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
                p: Prelim.password, papID: Prelim.paperID,
                dec: Prelim.decision):
      state_ext[Unit]
  =
  {
    val (Prelim.Paper(title, abstracta, pc, reviews, dis, decs, fpc)):
          Prelim.paper
      = (paper[Unit](s)).apply(papID);
    paper_update[Unit](((_: Prelim.paperID => Prelim.paper) =>
                         Fun.fun_upd[Prelim.paperID,
                                      Prelim.paper](paper[Unit](s), papID,
             Prelim.Paper(title, abstracta, pc, reviews, dis, dec :: decs,
                           fpc))),
                        s)
  }

def uuStep(s: state_ext[Unit], x1: uuAct): (out, state_ext[Unit]) = (s, x1)
  match {
  case (s, uuNews(confID, uID, p, ev)) =>
    (if (e_uupdateNews(s, confID, uID, p, ev))
      (outOK(), uupdateNews(s, confID, uID, p, ev)) else (outErr(), s))
  case (s, uuDis(confID, uID, p, papID, comm)) =>
    (if (e_uupdateDis(s, confID, uID, p, papID, comm))
      (outOK(), uupdateDis(s, confID, uID, p, papID, comm)) else (outErr(), s))
  case (s, uuReview(confID, uID, p, papID, n, rc)) =>
    (if (e_uupdateReview(s, confID, uID, p, papID, n, rc))
      (outOK(), uupdateReview(s, confID, uID, p, papID, n, rc))
      else (outErr(), s))
  case (s, uuDec(confID, uID, p, papID, dec)) =>
    (if (e_uupdateDec(s, confID, uID, p, papID, dec))
      (outOK(), uupdateDec(s, confID, uID, p, papID, dec)) else (outErr(), s))
}

def e_updatePaperTA(s: state_ext[Unit], confID: Prelim.confID,
                     uID: Prelim.userID, p: Prelim.password,
                     papID: Prelim.paperID, name: String, info: String):
      Boolean
  =
  IDsOK(s, List(confID), List(uID), List(papID)) &&
    (Prelim.equal_password((pass[Unit](s)).apply(uID), p) &&
      (Nat.equal_nat((phase[Unit](s)).apply(confID),
                      Nat.Suc(Nat.Suc(Nat.zero_nat()))) &&
        Lista.member[Prelim.role]((roles[Unit](s)).apply(confID).apply(uID),
                                   Prelim.Aut(papID))))

def e_updateFPaperC(s: state_ext[Unit], confID: Prelim.confID,
                     uID: Prelim.userID, p: Prelim.password,
                     papID: Prelim.paperID, fpc: Prelim.fpcontent):
      Boolean
  =
  IDsOK(s, List(confID), List(uID), List(papID)) &&
    (Prelim.equal_password((pass[Unit](s)).apply(uID), p) &&
      (Nat.equal_nat((phase[Unit](s)).apply(confID),
                      Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.zero_nat()))))))) &&
        (Lista.member[Prelim.role]((roles[Unit](s)).apply(confID).apply(uID),
                                    Prelim.Aut(papID)) &&
          Prelim.equal_decision(Prelim.decPaper((paper[Unit](s)).apply(papID)),
                                 Prelim.Accept()))))

def e_updateReview(s: state_ext[Unit], confID: Prelim.confID,
                    uID: Prelim.userID, p: Prelim.password,
                    papID: Prelim.paperID, n: Nat.nat,
                    rc: (Prelim.exp, (Prelim.score, String))):
      Boolean
  =
  IDsOK(s, List(confID), List(uID), List(papID)) &&
    (Prelim.equal_password((pass[Unit](s)).apply(uID), p) &&
      (Nat.equal_nat((phase[Unit](s)).apply(confID),
                      Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.zero_nat()))))) &&
        (isRev(s, confID, uID, papID) &&
          Nat.equal_nat(getReviewIndex(s, confID, uID, papID), n))))

def e_updatePaperC(s: state_ext[Unit], confID: Prelim.confID,
                    uID: Prelim.userID, p: Prelim.password,
                    papID: Prelim.paperID, pc: Prelim.pcontent):
      Boolean
  =
  IDsOK(s, List(confID), List(uID), List(papID)) &&
    (Prelim.equal_password((pass[Unit](s)).apply(uID), p) &&
      (Nat.equal_nat((phase[Unit](s)).apply(confID),
                      Nat.Suc(Nat.Suc(Nat.zero_nat()))) &&
        Lista.member[Prelim.role]((roles[Unit](s)).apply(confID).apply(uID),
                                   Prelim.Aut(papID))))

def updatePaperTA(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
                   p: Prelim.password, papID: Prelim.paperID, title: String,
                   abstracta: String):
      state_ext[Unit]
  =
  {
    val (Prelim.Paper(_, _, pc, reviews, dis, decs, fpc)): Prelim.paper =
      (paper[Unit](s)).apply(papID);
    paper_update[Unit](((_: Prelim.paperID => Prelim.paper) =>
                         Fun.fun_upd[Prelim.paperID,
                                      Prelim.paper](paper[Unit](s), papID,
             Prelim.Paper(title, abstracta, pc, reviews, dis, decs, fpc))),
                        s)
  }

def updateFPaperC(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
                   p: Prelim.password, papID: Prelim.paperID,
                   fpc: Prelim.fpcontent):
      state_ext[Unit]
  =
  {
    val (Prelim.Paper(title, abstracta, pc, reviews, dis, decs, _)):
          Prelim.paper
      = (paper[Unit](s)).apply(papID);
    paper_update[Unit](((_: Prelim.paperID => Prelim.paper) =>
                         Fun.fun_upd[Prelim.paperID,
                                      Prelim.paper](paper[Unit](s), papID,
             Prelim.Paper(title, abstracta, pc, reviews, dis, decs, fpc))),
                        s)
  }

def e_updatePhase(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
                   p: Prelim.password, ph: Nat.nat):
      Boolean
  =
  IDsOK(s, List(confID), List(uID), Nil) &&
    (Prelim.equal_password((pass[Unit](s)).apply(uID), p) &&
      (Nat.less_eq_nat(Nat.Suc(Nat.zero_nat()),
                        (phase[Unit](s)).apply(confID)) &&
        (Nat.less_nat((phase[Unit](s)).apply(confID),
                       Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.zero_nat())))))))) &&
          (Lista.member[Prelim.role]((roles[Unit](s)).apply(confID).apply(uID),
                                      Prelim.Chair()) &&
            Nat.equal_nat(ph, Prelim.SucPH((phase[Unit](s)).apply(confID)))))))

def e_updateConfA(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
                   p: Prelim.password):
      Boolean
  =
  IDsOK(s, List(confID), List(uID), Nil) &&
    (Prelim.equal_password((pass[Unit](s)).apply(uID), p) &&
      (Prelim.equal_userIDa(uID, voronkov[Unit](s)) &&
        Nat.equal_nat((phase[Unit](s)).apply(confID), Nat.zero_nat())))

def updateReview(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
                  p: Prelim.password, papID: Prelim.paperID, n: Nat.nat,
                  rc: (Prelim.exp, (Prelim.score, String))):
      state_ext[Unit]
  =
  {
    val (Prelim.Paper(title, abstracta, pc, reviews, dis, decs, fpc)):
          Prelim.paper
      = (paper[Unit](s)).apply(papID)
    val review: List[(Prelim.exp, (Prelim.score, String))] = List(rc)
    val reviewsa: List[List[(Prelim.exp, (Prelim.score, String))]] =
      Lista.list_update[List[(Prelim.exp,
                               (Prelim.score, String))]](reviews, n, review);
    paper_update[Unit](((_: Prelim.paperID => Prelim.paper) =>
                         Fun.fun_upd[Prelim.paperID,
                                      Prelim.paper](paper[Unit](s), papID,
             Prelim.Paper(title, abstracta, pc, reviewsa, dis, decs, fpc))),
                        s)
  }

def updatePaperC(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
                  p: Prelim.password, papID: Prelim.paperID,
                  pc: Prelim.pcontent):
      state_ext[Unit]
  =
  {
    val (Prelim.Paper(title, abstracta, _, reviews, dis, decs, fpc)):
          Prelim.paper
      = (paper[Unit](s)).apply(papID);
    paper_update[Unit](((_: Prelim.paperID => Prelim.paper) =>
                         Fun.fun_upd[Prelim.paperID,
                                      Prelim.paper](paper[Unit](s), papID,
             Prelim.Paper(title, abstracta, pc, reviews, dis, decs, fpc))),
                        s)
  }

def e_updateUser(s: state_ext[Unit], uID: Prelim.userID, pa: Prelim.password,
                  p: Prelim.password, name: String, info: String,
                  email: String):
      Boolean
  =
  IDsOK(s, Nil, List(uID), Nil) &&
    Prelim.equal_password((pass[Unit](s)).apply(uID), pa)

def e_updatePref(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
                  p: Prelim.password, papID: Prelim.paperID,
                  pr: Prelim.preference):
      Boolean
  =
  IDsOK(s, List(confID), List(uID), List(papID)) &&
    (Prelim.equal_password((pass[Unit](s)).apply(uID), p) &&
      (Nat.equal_nat((phase[Unit](s)).apply(confID),
                      Nat.Suc(Nat.Suc(Nat.Suc(Nat.zero_nat())))) &&
        (Lista.member[Prelim.role]((roles[Unit](s)).apply(confID).apply(uID),
                                    Prelim.PC()) &&
          ! (Lista.member[Prelim.role]((roles[Unit](s)).apply(confID).apply(uID),
Prelim.Aut(papID))))))

def phase_update[A](phasea:
                      (Prelim.confID => Nat.nat) => Prelim.confID => Nat.nat,
                     x1: state_ext[A]):
      state_ext[A]
  =
  (phasea, x1) match {
  case (phasea,
         state_exta(confIDs, conf, userIDs, pass, user, roles, paperIDs, paper,
                     pref, voronkov, news, phase, more))
    => state_exta[A](confIDs, conf, userIDs, pass, user, roles, paperIDs, paper,
                      pref, voronkov, news, phasea(phase), more)
}

def updatePhase(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
                 p: Prelim.password, ph: Nat.nat):
      state_ext[Unit]
  =
  phase_update[Unit](((_: Prelim.confID => Nat.nat) =>
                       Fun.fun_upd[Prelim.confID,
                                    Nat.nat](phase[Unit](s), confID, ph)),
                      s)

def updateConfA(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
                 p: Prelim.password):
      state_ext[Unit]
  =
  phase_update[Unit](((_: Prelim.confID => Nat.nat) =>
                       Fun.fun_upd[Prelim.confID,
                                    Nat.nat](phase[Unit](s), confID,
      Nat.Suc(Nat.zero_nat()))),
                      s)

def user_update[A](usera:
                     (Prelim.userID => Prelim.user) =>
                       Prelim.userID => Prelim.user,
                    x1: state_ext[A]):
      state_ext[A]
  =
  (usera, x1) match {
  case (usera,
         state_exta(confIDs, conf, userIDs, pass, user, roles, paperIDs, paper,
                     pref, voronkov, news, phase, more))
    => state_exta[A](confIDs, conf, userIDs, pass, usera(user), roles, paperIDs,
                      paper, pref, voronkov, news, phase, more)
}

def pass_update[A](passa:
                     (Prelim.userID => Prelim.password) =>
                       Prelim.userID => Prelim.password,
                    x1: state_ext[A]):
      state_ext[A]
  =
  (passa, x1) match {
  case (passa,
         state_exta(confIDs, conf, userIDs, pass, user, roles, paperIDs, paper,
                     pref, voronkov, news, phase, more))
    => state_exta[A](confIDs, conf, userIDs, passa(pass), user, roles, paperIDs,
                      paper, pref, voronkov, news, phase, more)
}

def updateUser(s: state_ext[Unit], uID: Prelim.userID, pa: Prelim.password,
                p: Prelim.password, name: String, info: String, email: String):
      state_ext[Unit]
  =
  pass_update[Unit](((_: Prelim.userID => Prelim.password) =>
                      Fun.fun_upd[Prelim.userID,
                                   Prelim.password](pass[Unit](s), uID, p)),
                     user_update[Unit](((_: Prelim.userID => Prelim.user) =>
 Fun.fun_upd[Prelim.userID,
              Prelim.user](user[Unit](s), uID, Prelim.User(name, info, email))),
s))

def pref_update[A](prefa:
                     (Prelim.userID => Prelim.paperID => Prelim.preference) =>
                       Prelim.userID => Prelim.paperID => Prelim.preference,
                    x1: state_ext[A]):
      state_ext[A]
  =
  (prefa, x1) match {
  case (prefa,
         state_exta(confIDs, conf, userIDs, pass, user, roles, paperIDs, paper,
                     pref, voronkov, news, phase, more))
    => state_exta[A](confIDs, conf, userIDs, pass, user, roles, paperIDs, paper,
                      prefa(pref), voronkov, news, phase, more)
}

def updatePref(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
                p: Prelim.password, papID: Prelim.paperID,
                pr: Prelim.preference):
      state_ext[Unit]
  =
  pref_update[Unit](((_: Prelim.userID => Prelim.paperID => Prelim.preference)
                       =>
                      Trivia.fun_upd2[Prelim.userID, Prelim.paperID,
                                       Prelim.preference](pref[Unit](s), uID,
                   papID, pr)),
                     s)

def uStep(s: state_ext[Unit], x1: uAct): (out, state_ext[Unit]) = (s, x1) match
  {
  case (s, uUser(uID, pa, p, name, info, email)) =>
    (if (e_updateUser(s, uID, pa, p, name, info, email))
      (outOK(), updateUser(s, uID, pa, p, name, info, email))
      else (outErr(), s))
  case (s, uConfA(confID, uID, p)) =>
    (if (e_updateConfA(s, confID, uID, p))
      (outOK(), updateConfA(s, confID, uID, p)) else (outErr(), s))
  case (s, uPhase(confID, uID, p, ph)) =>
    (if (e_updatePhase(s, confID, uID, p, ph))
      (outOK(), updatePhase(s, confID, uID, p, ph)) else (outErr(), s))
  case (s, uPaperTA(confID, uID, p, papID, name, info)) =>
    (if (e_updatePaperTA(s, confID, uID, p, papID, name, info))
      (outOK(), updatePaperTA(s, confID, uID, p, papID, name, info))
      else (outErr(), s))
  case (s, uPaperC(confID, uID, p, papID, pc)) =>
    (if (e_updatePaperC(s, confID, uID, p, papID, pc))
      (outOK(), updatePaperC(s, confID, uID, p, papID, pc)) else (outErr(), s))
  case (s, uPref(confID, uID, p, papID, pr)) =>
    (if (e_updatePref(s, confID, uID, p, papID, pr))
      (outOK(), updatePref(s, confID, uID, p, papID, pr)) else (outErr(), s))
  case (s, uReview(confID, uID, p, papID, n, rc)) =>
    (if (e_updateReview(s, confID, uID, p, papID, n, rc))
      (outOK(), updateReview(s, confID, uID, p, papID, n, rc))
      else (outErr(), s))
  case (s, uFPaperC(confID, uID, p, papID, fpc)) =>
    (if (e_updateFPaperC(s, confID, uID, p, papID, fpc))
      (outOK(), updateFPaperC(s, confID, uID, p, papID, fpc))
      else (outErr(), s))
}

def e_createConflict(s: state_ext[Unit], confID: Prelim.confID,
                      uIDa: Prelim.userID, p: Prelim.password,
                      papID: Prelim.paperID, uID: Prelim.userID):
      Boolean
  =
  IDsOK(s, List(confID), List(uIDa, uID), List(papID)) &&
    (Prelim.equal_password((pass[Unit](s)).apply(uIDa), p) &&
      (Nat.equal_nat((phase[Unit](s)).apply(confID),
                      Nat.Suc(Nat.Suc(Nat.zero_nat()))) &&
        (Lista.member[Prelim.role]((roles[Unit](s)).apply(confID).apply(uIDa),
                                    Prelim.Aut(papID)) &&
          Lista.member[Prelim.role]((roles[Unit](s)).apply(confID).apply(uID),
                                     Prelim.PC()))))

def e_createReview(s: state_ext[Unit], confID: Prelim.confID,
                    uIDa: Prelim.userID, p: Prelim.password,
                    papID: Prelim.paperID, uID: Prelim.userID):
      Boolean
  =
  IDsOK(s, List(confID), List(uIDa, uID), List(papID)) &&
    (Prelim.equal_password((pass[Unit](s)).apply(uIDa), p) &&
      (Nat.equal_nat((phase[Unit](s)).apply(confID),
                      Nat.Suc(Nat.Suc(Nat.Suc(Nat.Suc(Nat.zero_nat()))))) &&
        (Lista.member[Prelim.role]((roles[Unit](s)).apply(confID).apply(uIDa),
                                    Prelim.Chair()) &&
          (! (Prelim.equal_preference((pref[Unit](s)).apply(uIDa).apply(papID),
                                       Prelim.Conflict())) &&
            (Lista.member[Prelim.role]((roles[Unit](s)).apply(confID).apply(uID),
Prelim.PC()) &&
              (! (isRev(s, confID, uID, papID)) &&
                ! (Prelim.equal_preference((pref[Unit](s)).apply(uID).apply(papID),
    Prelim.Conflict()))))))))

def e_createAuthor(s: state_ext[Unit], confID: Prelim.confID,
                    uIDa: Prelim.userID, p: Prelim.password,
                    papID: Prelim.paperID, uID: Prelim.userID):
      Boolean
  =
  IDsOK(s, List(confID), List(uIDa, uID), List(papID)) &&
    (Prelim.equal_password((pass[Unit](s)).apply(uIDa), p) &&
      (Nat.equal_nat((phase[Unit](s)).apply(confID),
                      Nat.Suc(Nat.Suc(Nat.zero_nat()))) &&
        (Lista.member[Prelim.role]((roles[Unit](s)).apply(confID).apply(uIDa),
                                    Prelim.Aut(papID)) &&
          ! (Prelim.equal_userIDa(uIDa, uID)))))

def createConflict(s: state_ext[Unit], confID: Prelim.confID,
                    uIDa: Prelim.userID, p: Prelim.password,
                    papID: Prelim.paperID, uID: Prelim.userID):
      state_ext[Unit]
  =
  pref_update[Unit](((_: Prelim.userID => Prelim.paperID => Prelim.preference)
                       =>
                      Trivia.fun_upd2[Prelim.userID, Prelim.paperID,
                                       Prelim.preference](pref[Unit](s), uID,
                   papID, Prelim.Conflict())),
                     s)

def e_createPaper(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
                   p: Prelim.password, papID: Prelim.paperID, name: String,
                   info: String):
      Boolean
  =
  IDsOK(s, List(confID), List(uID), Nil) &&
    (Prelim.equal_password((pass[Unit](s)).apply(uID), p) &&
      (Nat.equal_nat((phase[Unit](s)).apply(confID),
                      Nat.Suc(Nat.Suc(Nat.zero_nat()))) &&
        ! (Lista.member[Prelim.paperID](getAllPaperIDs(s), papID))))

def e_createChair(s: state_ext[Unit], confID: Prelim.confID,
                   uIDa: Prelim.userID, p: Prelim.password, uID: Prelim.userID):
      Boolean
  =
  {
    userIDs[Unit](s);
    IDsOK(s, List(confID), List(uIDa, uID), Nil) &&
      (Prelim.equal_password((pass[Unit](s)).apply(uIDa), p) &&
        (Nat.equal_nat((phase[Unit](s)).apply(confID),
                        Nat.Suc(Nat.zero_nat())) &&
          Lista.member[Prelim.role]((roles[Unit](s)).apply(confID).apply(uIDa),
                                     Prelim.Chair())))
  }

def e_createUser(s: state_ext[Unit], uID: Prelim.userID, p: Prelim.password,
                  name: String, info: String, email: String):
      Boolean
  =
  ! (Lista.member[Prelim.userID](userIDs[Unit](s), uID))

def e_createConf(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
                  p: Prelim.password, name: String, info: String):
      Boolean
  =
  IDsOK(s, Nil, List(uID), Nil) &&
    (Prelim.equal_password((pass[Unit](s)).apply(uID), p) &&
      ! (Lista.member[Prelim.confID](confIDs[Unit](s), confID)))

def roles_update[A](rolesa:
                      (Prelim.confID => Prelim.userID => List[Prelim.role]) =>
                        Prelim.confID => Prelim.userID => List[Prelim.role],
                     x1: state_ext[A]):
      state_ext[A]
  =
  (rolesa, x1) match {
  case (rolesa,
         state_exta(confIDs, conf, userIDs, pass, user, roles, paperIDs, paper,
                     pref, voronkov, news, phase, more))
    => state_exta[A](confIDs, conf, userIDs, pass, user, rolesa(roles),
                      paperIDs, paper, pref, voronkov, news, phase, more)
}

def createReview(s: state_ext[Unit], confID: Prelim.confID, uIDa: Prelim.userID,
                  p: Prelim.password, papID: Prelim.paperID,
                  uID: Prelim.userID):
      state_ext[Unit]
  =
  {
    val (Prelim.Paper(title, abstracta, pc, reviews, dis, decs, fpc)):
          Prelim.paper
      = (paper[Unit](s)).apply(papID)
    val rls: List[Prelim.role] = (roles[Unit](s)).apply(confID).apply(uID)
    val n: Nat.nat =
      Lista.size_list[List[(Prelim.exp,
                             (Prelim.score,
                               String))]].apply(Prelim.reviewsPaper((paper[Unit](s)).apply(papID)))
    val reviewsa: List[List[(Prelim.exp, (Prelim.score, String))]] =
      reviews ++ List(Nil);
    paper_update[Unit](((_: Prelim.paperID => Prelim.paper) =>
                         Fun.fun_upd[Prelim.paperID,
                                      Prelim.paper](paper[Unit](s), papID,
             Prelim.Paper(title, abstracta, pc, reviewsa, dis, decs, fpc))),
                        roles_update[Unit](((_:
       Prelim.confID => Prelim.userID => List[Prelim.role])
      =>
     Trivia.fun_upd2[Prelim.confID, Prelim.userID,
                      List[Prelim.role]](roles[Unit](s), confID, uID,
  Lista.insert[Prelim.role](Prelim.Rev(papID, n), rls))),
    s))
  }

def createAuthor(s: state_ext[Unit], confID: Prelim.confID, uIDa: Prelim.userID,
                  p: Prelim.password, papID: Prelim.paperID,
                  uID: Prelim.userID):
      state_ext[Unit]
  =
  {
    val rls: List[Prelim.role] = (roles[Unit](s)).apply(confID).apply(uID);
    pref_update[Unit](((_: Prelim.userID => Prelim.paperID => Prelim.preference)
                         =>
                        Trivia.fun_upd2[Prelim.userID, Prelim.paperID,
 Prelim.preference](pref[Unit](s), uID, papID, Prelim.Conflict())),
                       roles_update[Unit](((_:
      Prelim.confID => Prelim.userID => List[Prelim.role])
     =>
    Trivia.fun_upd2[Prelim.confID, Prelim.userID,
                     List[Prelim.role]](roles[Unit](s), confID, uID,
 Lista.insert[Prelim.role](Prelim.Aut(papID), rls))),
   s))
  }

def paperIDs_update[A](paperIDsa:
                         (Prelim.confID => List[Prelim.paperID]) =>
                           Prelim.confID => List[Prelim.paperID],
                        x1: state_ext[A]):
      state_ext[A]
  =
  (paperIDsa, x1) match {
  case (paperIDsa,
         state_exta(confIDs, conf, userIDs, pass, user, roles, paperIDs, paper,
                     pref, voronkov, news, phase, more))
    => state_exta[A](confIDs, conf, userIDs, pass, user, roles,
                      paperIDsa(paperIDs), paper, pref, voronkov, news, phase,
                      more)
}

def createPaper(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
                 p: Prelim.password, papID: Prelim.paperID, title: String,
                 abstracta: String):
      state_ext[Unit]
  =
  {
    val papIDs: List[Prelim.paperID] = (paperIDs[Unit](s)).apply(confID)
    val rls: List[Prelim.role] = (roles[Unit](s)).apply(confID).apply(uID);
    pref_update[Unit](((_: Prelim.userID => Prelim.paperID => Prelim.preference)
                         =>
                        Trivia.fun_upd2[Prelim.userID, Prelim.paperID,
 Prelim.preference](pref[Unit](s), uID, papID, Prelim.Conflict())),
                       roles_update[Unit](((_:
      Prelim.confID => Prelim.userID => List[Prelim.role])
     =>
    Trivia.fun_upd2[Prelim.confID, Prelim.userID,
                     List[Prelim.role]](roles[Unit](s), confID, uID,
 Lista.insert[Prelim.role](Prelim.Aut(papID), rls))),
   paper_update[Unit](((_: Prelim.paperID => Prelim.paper) =>
                        Fun.fun_upd[Prelim.paperID,
                                     Prelim.paper](paper[Unit](s), papID,
            Prelim.Paper(title, abstracta, Prelim.NoPContent(), Nil,
                          Prelim.Dis(Nil), Nil, Prelim.NoFPContent()))),
                       paperIDs_update[Unit](((_:
         Prelim.confID => List[Prelim.paperID])
        =>
       Fun.fun_upd[Prelim.confID,
                    List[Prelim.paperID]](paperIDs[Unit](s), confID,
   papID :: papIDs)),
      s))))
  }

def createChair(s: state_ext[Unit], confID: Prelim.confID, uIDa: Prelim.userID,
                 p: Prelim.password, uID: Prelim.userID):
      state_ext[Unit]
  =
  {
    val rls: List[Prelim.role] = (roles[Unit](s)).apply(confID).apply(uID);
    roles_update[Unit](((_: Prelim.confID => Prelim.userID => List[Prelim.role])
                          =>
                         Trivia.fun_upd2[Prelim.confID, Prelim.userID,
  List[Prelim.role]](roles[Unit](s), confID, uID,
                      Lista.insert[Prelim.role](Prelim.PC(),
         Lista.insert[Prelim.role](Prelim.Chair(), rls)))),
                        s)
  }

def e_createPC(s: state_ext[Unit], confID: Prelim.confID, uIDa: Prelim.userID,
                p: Prelim.password, uID: Prelim.userID):
      Boolean
  =
  {
    userIDs[Unit](s);
    IDsOK(s, List(confID), List(uIDa, uID), Nil) &&
      (Prelim.equal_password((pass[Unit](s)).apply(uIDa), p) &&
        (Nat.equal_nat((phase[Unit](s)).apply(confID),
                        Nat.Suc(Nat.zero_nat())) &&
          Lista.member[Prelim.role]((roles[Unit](s)).apply(confID).apply(uIDa),
                                     Prelim.Chair())))
  }

def userIDs_update[A](userIDsa: (List[Prelim.userID]) => List[Prelim.userID],
                       x1: state_ext[A]):
      state_ext[A]
  =
  (userIDsa, x1) match {
  case (userIDsa,
         state_exta(confIDs, conf, userIDs, pass, user, roles, paperIDs, paper,
                     pref, voronkov, news, phase, more))
    => state_exta[A](confIDs, conf, userIDsa(userIDs), pass, user, roles,
                      paperIDs, paper, pref, voronkov, news, phase, more)
}

def createUser(s: state_ext[Unit], uID: Prelim.userID, p: Prelim.password,
                name: String, info: String, email: String):
      state_ext[Unit]
  =
  {
    val uIDs: List[Prelim.userID] = userIDs[Unit](s);
    pass_update[Unit](((_: Prelim.userID => Prelim.password) =>
                        Fun.fun_upd[Prelim.userID,
                                     Prelim.password](pass[Unit](s), uID, p)),
                       user_update[Unit](((_: Prelim.userID => Prelim.user) =>
   Fun.fun_upd[Prelim.userID,
                Prelim.user](user[Unit](s), uID,
                              Prelim.User(name, info, email))),
  userIDs_update[Unit](((_: List[Prelim.userID]) => uID :: uIDs), s)))
  }

def confIDs_update[A](confIDsa: (List[Prelim.confID]) => List[Prelim.confID],
                       x1: state_ext[A]):
      state_ext[A]
  =
  (confIDsa, x1) match {
  case (confIDsa,
         state_exta(confIDs, conf, userIDs, pass, user, roles, paperIDs, paper,
                     pref, voronkov, news, phase, more))
    => state_exta[A](confIDsa(confIDs), conf, userIDs, pass, user, roles,
                      paperIDs, paper, pref, voronkov, news, phase, more)
}

def conf_update[A](confa:
                     (Prelim.confID => Prelim.conf) =>
                       Prelim.confID => Prelim.conf,
                    x1: state_ext[A]):
      state_ext[A]
  =
  (confa, x1) match {
  case (confa,
         state_exta(confIDs, conf, userIDs, pass, user, roles, paperIDs, paper,
                     pref, voronkov, news, phase, more))
    => state_exta[A](confIDs, confa(conf), userIDs, pass, user, roles, paperIDs,
                      paper, pref, voronkov, news, phase, more)
}

def createConf(s: state_ext[Unit], confID: Prelim.confID, uID: Prelim.userID,
                p: Prelim.password, name: String, info: String):
      state_ext[Unit]
  =
  {
    val confIDsa: List[Prelim.confID] = confIDs[Unit](s);
    roles_update[Unit](((_: Prelim.confID => Prelim.userID => List[Prelim.role])
                          =>
                         Trivia.fun_upd2[Prelim.confID, Prelim.userID,
  List[Prelim.role]](roles[Unit](s), confID, uID,
                      List(Prelim.PC(), Prelim.Chair()))),
                        conf_update[Unit](((_: Prelim.confID => Prelim.conf) =>
    Fun.fun_upd[Prelim.confID,
                 Prelim.conf](conf[Unit](s), confID, Prelim.Conf(name, info))),
   confIDs_update[Unit](((_: List[Prelim.confID]) => confID :: confIDsa), s)))
  }

def createPC(s: state_ext[Unit], confID: Prelim.confID, uIDa: Prelim.userID,
              p: Prelim.password, uID: Prelim.userID):
      state_ext[Unit]
  =
  {
    val rls: List[Prelim.role] = (roles[Unit](s)).apply(confID).apply(uID);
    roles_update[Unit](((_: Prelim.confID => Prelim.userID => List[Prelim.role])
                          =>
                         Trivia.fun_upd2[Prelim.confID, Prelim.userID,
  List[Prelim.role]](roles[Unit](s), confID, uID,
                      Lista.insert[Prelim.role](Prelim.PC(), rls))),
                        s)
  }

def cStep(s: state_ext[Unit], x1: cAct): (out, state_ext[Unit]) = (s, x1) match
  {
  case (s, cUser(uID, p, name, info, email)) =>
    (if (e_createUser(s, uID, p, name, info, email))
      (outOK(), createUser(s, uID, p, name, info, email)) else (outErr(), s))
  case (s, cConf(confID, uID, p, name, info)) =>
    (if (e_createConf(s, confID, uID, p, name, info))
      (outOK(), createConf(s, confID, uID, p, name, info)) else (outErr(), s))
  case (s, cPC(confID, uIDa, p, uID)) =>
    (if (e_createPC(s, confID, uIDa, p, uID))
      (outOK(), createPC(s, confID, uIDa, p, uID)) else (outErr(), s))
  case (s, cChair(confID, uIDa, p, uID)) =>
    (if (e_createChair(s, confID, uIDa, p, uID))
      (outOK(), createChair(s, confID, uIDa, p, uID)) else (outErr(), s))
  case (s, cPaper(confID, uID, p, papID, name, info)) =>
    (if (e_createPaper(s, confID, uID, p, papID, name, info))
      (outOK(), createPaper(s, confID, uID, p, papID, name, info))
      else (outErr(), s))
  case (s, cAuthor(confID, uIDa, p, papID, uID)) =>
    (if (e_createAuthor(s, confID, uIDa, p, papID, uID))
      (outOK(), createAuthor(s, confID, uIDa, p, papID, uID))
      else (outErr(), s))
  case (s, cConflict(confID, uIDa, p, papID, uID)) =>
    (if (e_createConflict(s, confID, uIDa, p, papID, uID))
      (outOK(), createConflict(s, confID, uIDa, p, papID, uID))
      else (outErr(), s))
  case (s, cReview(confID, uIDa, p, papID, uID)) =>
    (if (e_createReview(s, confID, uIDa, p, papID, uID))
      (outOK(), createReview(s, confID, uIDa, p, papID, uID))
      else (outErr(), s))
}

def step(s: state_ext[Unit], x1: act): (out, state_ext[Unit]) = (s, x1) match {
  case (s, Cact(ca)) => cStep(s, ca)
  case (s, Uact(ua)) => uStep(s, ua)
  case (s, UUact(uua)) => uuStep(s, uua)
  case (s, Ract(ra)) => (rObs(s, ra), s)
  case (s, Lact(la)) => (lObs(s, la), s)
}

def istate: state_ext[Unit] =
  state_exta[Unit](Nil, ((_: Prelim.confID) => Prelim.emptyConf),
                    List(Prelim.voronkovUserID),
                    ((_: Prelim.userID) => Prelim.emptyPass),
                    ((_: Prelim.userID) => Prelim.emptyUser),
                    ((_: Prelim.confID) => (_: Prelim.userID) => Nil),
                    ((_: Prelim.confID) => Nil),
                    ((_: Prelim.paperID) => Prelim.emptyPaper),
                    ((_: Prelim.userID) => (_: Prelim.paperID) =>
                      Prelim.NoPref()),
                    Prelim.voronkovUserID, ((_: Prelim.confID) => Nil),
                    ((_: Prelim.confID) => Nat.zero_nat()), ())

} /* object System_Specification */
