(* Confidentiality of failed updates *)
theory FailedUUpdate
imports "../Observation_Setup" "~~/src/HOL/Library/Infinite_Set"
        "../../BD_Security_Framework/Compositional_Reasoning"
begin

text {* The content of failed undestructive update actions is not released. *}

datatype uuContent =
  uucNews confID string
 |uucDis confID paperID string
 |uucReview confID paperID nat rcontent
 |uucDec confID  paperID decision

primrec uuContentOfA :: "uuAct \<Rightarrow> userID \<times> uuContent" where
  "uuContentOfA (uuNews confID userID _ string) = (userID, uucNews confID string)"
|  "uuContentOfA (uuDis confID userID _ paperID string) = (userID, uucDis confID paperID string)"
|  "uuContentOfA (uuReview confID userID _ paperID n rc) =
    (userID, uucReview confID paperID n rc)"
|  "uuContentOfA (uuDec confID userID _ paperID decision) = (userID, uucDec confID  paperID decision)"

section{*  Value Setup *}

type_synonym "value" = "userID \<times> uuContent"

fun \<phi> :: "(state,act,out) trans \<Rightarrow> bool" where
  "\<phi> (Trans _ (UUact uact) outErr _) \<longleftrightarrow> uuUserOfA uact \<notin> UIDs"
| "\<phi> _ = False"

fun f :: "(state,act,out) trans \<Rightarrow> value" where
  "f (Trans _ (UUact uact) outErr _) = uuContentOfA uact"

definition T :: "(state,act,out) trans \<Rightarrow> bool" where [simp]: "T tr \<equiv> False"

definition B :: "value list \<Rightarrow> value list \<Rightarrow> bool" where
  [simp]: "B vl vl1 \<equiv> (UIDs=UNIV \<longrightarrow> vl\<noteq>[]) \<and> never (\<lambda> v. fst v \<in> UIDs) vl1"

interpretation BD_Security_IO_Aut where
istate = istate and step = step and
\<phi> = \<phi> and f = f and \<gamma> = \<gamma> and g = g and T = T and B = B
done

definition \<Delta> :: "state \<Rightarrow> value list \<Rightarrow> state \<Rightarrow> value list \<Rightarrow> bool"
where "\<Delta> s vl s1 vl1 \<equiv> s1=s \<and> (UIDs=UNIV \<longrightarrow> vl=[] \<longrightarrow> vl1=[]) \<and> never (\<lambda> v. fst v \<in> UIDs) vl1"

theorem init_\<Delta>:
assumes B: "B vl vl1"
shows "\<Delta> istate vl istate vl1"
using B unfolding \<Delta>_def B_def by simp


theorem unwind_\<Delta>: "unwind \<Delta>"
proof(rule unwindI)
  fix s s1 :: state and vl vl1 :: "value list"
  assume rsP: "reachNT s" and rs1: "reach s1" and \<Delta>: "\<Delta> s vl s1 vl1"
  hence rs: "reach s" by (metis reachNT_reach)
  from \<Delta> have [simp]: "s1=s" unfolding \<Delta>_def by simp

  let ?ph = "phase s CID"
  show "(vl \<noteq> [] \<and> exit s (hd vl)) \<or>
        iaction \<Delta> s vl s1 vl1 \<or>
        ((vl \<noteq> [] \<or> vl1 = []) \<and> reaction \<Delta> s vl s1 vl1)" (is "?exit \<or> ?iact \<or> (_ \<and> ?react)")
  proof (cases vl1)
    case (Cons v vl1') note this[simp]

    obtain uid c where [simp]: "v=(uid,c)" by (cases v)

    have uid[simp]: "uid \<notin> UIDs" using \<Delta> unfolding \<Delta>_def by auto

    show ?thesis
    proof (cases "UIDs=UNIV")
      case False note this[simp]
      have ?iact
      proof -

        obtain inv_pwd where [simp]: "pass s uid \<noteq> inv_pwd"
          apply (cases "pass s uid")
          apply (rule_tac inv_pwd = "Password (getFresh [x] x)" in that)
          using getFresh[of "[a]" for a]
          apply simp
          apply metis
          done

        obtain a where step: "step s (UUact a) = (outErr,s)"
          and [simp]: "uuUserOfA a = uid"
          and [simp]: "uuContentOfA a = (uid,c)"
          apply (cases c, simp_all)

          apply (rule_tac a="uuNews x11 uid inv_pwd x12" in that, auto simp: uu_defs) []
          apply (rule_tac a="uuDis x21 uid inv_pwd x22 x23" in that, auto simp: uu_defs) []
          apply (rule_tac a="uuReview x31 uid inv_pwd x32 x33 x34" in that, auto simp: uu_defs) []
          apply (rule_tac a="uuDec x41 uid inv_pwd x42 x43" in that, auto simp: uu_defs) []
          done

        show ?thesis
        proof
          show "step s1 (UUact a) = (outErr,s)" using step by simp
          show "\<phi> (Trans s1 (UUact a) outErr s)" by simp

          show "consume (Trans s1 (UUact a) outErr s) vl1 vl1'"
            by (simp add: consume_def)
          show "\<not> \<gamma> (Trans s1 (UUact a) outErr s)"
            by simp
          show "\<Delta> s vl s vl1'" using \<Delta> unfolding \<Delta>_def by auto
        qed
      qed
      thus ?thesis by simp
    next
      case True note this[simp]
      from \<Delta> have [simp]: "vl\<noteq>[]" by (simp add: \<Delta>_def)

      have [simp]: "\<And>trn. \<not>\<phi> trn"
        by (case_tac trn rule: \<phi>.cases) simp_all

      have ?exit
        apply simp
        apply (rule exit_coind[where K="\<lambda>_. True"])
        apply simp_all
        done
      thus ?thesis by simp
    qed
  next
    case Nil note this[simp]
    have "?react" proof
      fix a ou s' vl'
      assume step: "step s a = (ou, s')"
        and "consume (Trans s a ou s') vl vl'"

      show "match \<Delta> s s1 vl1 a ou s' vl' \<or> ignore \<Delta> s s1 vl1 a ou s' vl'" (is "?match \<or> ?ignore")
      proof (cases "ou=outErr")
        case True note this[simp]

        from step have [simp]: "s'=s" using step_outErr_eq by simp

        show ?thesis proof (cases "userOfA a \<in> UIDs")
          case False note this[simp]
          have "?ignore"
            unfolding ignore_def
            by (simp add: \<Delta>_def)
          thus ?thesis ..
        next
          case True
          hence [simp]: "\<not> \<phi> (Trans s a outErr s)"
            by (cases "(Trans s a outErr s)" rule: \<phi>.cases) simp_all


          have ?match apply standard using True step
            apply (auto simp add: \<Delta>_def consume_def)
            done
          thus ?thesis ..
        qed
      next
        case False note this[simp]

        have [simp]: "\<not> \<phi> (Trans s a ou s')"
          by (cases "(Trans s a ou s')" rule: \<phi>.cases) simp_all

        have ?match apply standard using False step
            apply (auto simp add: \<Delta>_def consume_def)
            done
        thus ?thesis ..
      qed
    qed
    thus ?thesis by auto
  qed
qed

theorem secure: secure
  by (rule unwind_secure[OF init_\<Delta> unwind_\<Delta>]) simp_all

end
