(* Confidentiality of failed updates *)
theory FailedUpdate
imports "../Observation_Setup" "~~/src/HOL/Library/Infinite_Set"
        "../../BD_Security_Framework/Compositional_Reasoning"
begin

text {* The content of failed update actions is not released. *}

datatype uContent =
  ucUser password string string string
 |ucConfA confID
 |ucPhase confID  phase
 |ucPaperTA confID  paperID string string
 |ucPaperC confID   paperID pcontent
 |ucPref confID   paperID preference
 |ucReview confID   paperID nat rcontent

primrec uContentOfA :: "uAct \<Rightarrow> userID \<times> uContent" where
  "uContentOfA (uUser uID p p' name info email) = (uID, ucUser p' name info email)"
|  "uContentOfA (uConfA confID uID p) = (uID, ucConfA confID) "
|  "uContentOfA (uPhase confID uID p ph) = (uID, ucPhase confID ph)"
|  "uContentOfA (uPaperTA confID uID p papID name info) = (uID, ucPaperTA confID papID name info)"
|  "uContentOfA (uPaperC confID uID p papID pc) = (uID, ucPaperC confID papID pc)"
|  "uContentOfA (uPref confID uID p papID pr) = (uID, ucPref confID papID pr)"
|  "uContentOfA (uReview confID uID p papID n rc) = (uID, ucReview confID papID n rc)"


section{*  Value Setup *}


type_synonym "value" = "userID \<times> uContent"

fun \<phi> :: "(state,act,out) trans \<Rightarrow> bool" where
  "\<phi> (Trans _ (Uact uact) outErr _) \<longleftrightarrow> uUserOfA uact \<notin> UIDs"
| "\<phi> _ = False"

fun f :: "(state,act,out) trans \<Rightarrow> value" where
  "f (Trans _ (Uact uact) outErr _) = uContentOfA uact"

definition T :: "(state,act,out) trans \<Rightarrow> bool" where [simp]: "T tr \<equiv> False"

definition B :: "value list \<Rightarrow> value list \<Rightarrow> bool" where
  [simp]: "B vl vl1 \<equiv> (UIDs=UNIV \<longrightarrow> vl\<noteq>[]) \<and> never (\<lambda> v. fst v \<in> UIDs) vl1"

interpretation BD_Security_IO_Aut where
istate = istate and step = step and
\<phi> = \<phi> and f = f and \<gamma> = \<gamma> and g = g and T = T and B = B
done

definition \<Delta> :: "state \<Rightarrow> value list \<Rightarrow> state \<Rightarrow> value list \<Rightarrow> bool"
where "\<Delta> s vl s1 vl1 \<equiv> s1=s \<and> (UIDs=UNIV \<longrightarrow> vl=[] \<longrightarrow> vl1=[]) \<and> never (\<lambda> v. fst v \<in> UIDs) vl1"

theorem init_\<Delta>:
assumes B: "B vl vl1"
shows "\<Delta> istate vl istate vl1"
using B unfolding \<Delta>_def B_def by simp


theorem unwind_\<Delta>: "unwind \<Delta>"
proof
  fix s s1 :: state and vl vl1 :: "value list"
  assume rsT: "reachNT s" and rs1: "reach s1" and \<Delta>: "\<Delta> s vl s1 vl1"
  hence rs: "reach s" by (metis reachNT_reach)
  from \<Delta> have [simp]: "s1=s" unfolding \<Delta>_def by simp

  show "(vl \<noteq> [] \<and> exit s (hd vl)) \<or>
        iaction \<Delta> s vl s1 vl1 \<or>
        ((vl \<noteq> [] \<or> vl1 = []) \<and> reaction \<Delta> s vl s1 vl1)" (is "?exit \<or> ?iact \<or> (_ \<and> ?react)")
  proof (cases vl1)
    case (Cons v vl1') note Cons[simp]

    obtain uid c where [simp]: "v=(uid,c)" by (cases v)

    have uid[simp]: "uid \<notin> UIDs" using \<Delta> unfolding \<Delta>_def by auto

    show ?thesis
    proof (cases "UIDs=UNIV")
      case False note False[simp]
      have ?iact
      proof -

        obtain inv_pwd where [simp]: "pass s uid \<noteq> inv_pwd"
          apply (cases "pass s uid")
          apply (rule_tac inv_pwd = "Password (getFresh [x] x)" in that)
          using getFresh[of "[a]" for a]
          apply simp
          apply metis
          done

        obtain a where step: "step s (Uact a) = (outErr,s)"
          and [simp]: "uUserOfA a = uid"
          and [simp]: "uContentOfA a = (uid,c)"
          apply (cases c, simp_all)

          apply (rule_tac a="uUser uid inv_pwd x11 x12 x13 x14" in that, auto simp: u_defs) []
          apply (rule_tac a="uConfA x2 uid inv_pwd" in that, auto simp: u_defs) []
          apply (rule_tac a="uPhase x31 uid inv_pwd x32" in that, auto simp: u_defs) []
          apply (rule_tac a="uPaperTA x41 uid inv_pwd x42 x43 x44" in that, auto simp: u_defs) []
          apply (rule_tac a="uPaperC x51 uid inv_pwd x52 x53" in that, auto simp: u_defs) []
          apply (rule_tac a="uPref x61 uid inv_pwd x62 x63" in that, auto simp: u_defs) []
          apply (rule_tac a="uReview x71 uid inv_pwd x72 x73 x74" in that, auto simp: u_defs) []
          done

        show ?thesis
        proof
          show "step s1 (Uact a) = (outErr,s)" using step by simp
          show "\<phi> (Trans s1 (Uact a) outErr s)" by simp

          show "consume (Trans s1 (Uact a) outErr s) vl1 vl1'"
            by (simp add: consume_def)
          show "\<not> \<gamma> (Trans s1 (Uact a) outErr s)"
            by simp
          show "\<Delta> s vl s vl1'" using \<Delta> unfolding \<Delta>_def by auto
        qed
      qed
      thus ?thesis by simp
    next
      case True note True[simp]
      from \<Delta> have [simp]: "vl\<noteq>[]" by (simp add: \<Delta>_def)

      have [simp]: "\<And>trn. \<not>\<phi> trn"
        by (case_tac trn rule: \<phi>.cases) simp_all

      have ?exit
        apply simp
        apply (rule exit_coind[where K="\<lambda>_. True"])
        apply simp_all
        done
      thus ?thesis by simp
    qed
  next
    case Nil note Nil[simp]
    have "?react" proof
      fix a ou s' vl'
      assume step: "step s a = (ou, s')"
        and "consume (Trans s a ou s') vl vl'"

      show "match \<Delta> s s1 vl1 a ou s' vl' \<or> ignore \<Delta> s s1 vl1 a ou s' vl'" (is "?match \<or> ?ignore")
      proof (cases "ou=outErr")
        case True note True[simp]

        from step have [simp]: "s'=s" using step_outErr_eq by simp

        show ?thesis proof (cases "userOfA a \<in> UIDs")
          case False note this[simp]
          have "?ignore"
            unfolding ignore_def
            by (simp add: \<Delta>_def)
          thus ?thesis ..
        next
          case True
          hence [simp]: "\<not> \<phi> (Trans s a outErr s)"
            by (cases "(Trans s a outErr s)" rule: \<phi>.cases) simp_all


          have ?match 
            apply standard using True step
            apply (auto simp add: \<Delta>_def consume_def)
            done
          thus ?thesis ..
        qed
      next
        case False note this[simp]

        have [simp]: "\<not> \<phi> (Trans s a ou s')"
          by (cases "(Trans s a ou s')" rule: \<phi>.cases) simp_all

        have ?match
          apply standard using False step
            apply (auto simp add: \<Delta>_def consume_def)
            done
        thus ?thesis ..
      qed
    qed
    thus ?thesis by auto
  qed
qed

theorem secure: secure
  by (rule unwind_secure[OF init_\<Delta> unwind_\<Delta>]) simp_all

end
